/**/

   /*
   to do:
         Tokenizer first, then start working on tasks.

          fix: rethink EOC-- if the only reason is so the Toker can find commands, might
               as well just change type of word instead of inserting EOC. Should there
               be an EOC =everywhere= a command might appear? Including inside %@exec[],
               for instance? Or should EOC mean something different, like have EOCs only
               on the left side of a pipe and not inside @exec[].

          fix: the 4OS2 docs do not explicitly say so, but right-side comments seem to be
               legal in some places. Or at least they are ignored by 4OS2. BTMTool.cmd
               recognizes right-side comments after a label and after a few keywords.
               Where else is a comment legal?
                  YES: then, enddo, endiff, endswitch
                  NO:  echo, echoerr, etc, if..., do..., default

   */
   Main: /*fold00*/
   /*=================================================================================================*/
   /*  There are a lot of global variables and other things to be set up before the work starts.      */
   /*  The Setup process is divided into two parts, Init, at the start of the program, and            */
   /*  Setup, at the start of a script.                                                               */
   /*  The beginning of the work flow looks like this:                                                */
   /*        Main                                                                                     */
   /*          |__Init                  sets default values for variables                             */
   /*          |__Script_xxx            can change some variables                                     */
   /*               |__Setup            sets more variables based on Init and Script_xxx              */
   /*               |__Task_yyy         tasks do the work                                             */
   /*               |__Task_zzz                                                                       */
   /*               |__etc.                                                                           */
   /*                                                                                                 */
   /*  This gives a script a chance to customize most of the variables to suit itself.                */
   /*  See Style_akm for an example.                                                                  */
   /* =============================================================================================== */
   parse arg MainArgs
   signal on Syntax
   signal on NoValue

   call Init
   /* =============================================================================================== */
   /*  decide which script to call, call it, and quit                                                 */
   /* =============================================================================================== */
   Command = translate(G.0Command)
   /*  the scripts do not return, they call Egress themselves                                         */
   if Command == 'FORMAT' then call Script_Format
   if Command == 'BARE' then call Script_Bare
   if Command == 'DEPENDS'  then call Script_Depends
   if Command == 'CHECKONLY' then call Script_CheckOnly
   if Command == 'TEST' then call Script_Test
   if Command == 'HELP' then call Script_Help

   /*  this lets the user type "BTMTool.cmd -config" without a command                                */
   if Opt.0config then call Script_Format

   call Msg 'unrecognized command "'G.0Command'"'
   call Egress 'usage'
   /* === End of Main =============================================================================== */

   Init: /*fold00*/
   /* =============================================================================================== */
   /*  initializes a lot of variables                                                                 */
   /* =============================================================================================== */
   /*  the rest of Init has been split off into subroutines to make things understandable             */
   call Init_Global
   /*  Init_Config has the descriptions and default values of all the command line options            */
   call Init_Config
   call Init_CommandLine
   return 0
   /* === End of Init =============================================================================== */

   Init_Global: /*fold00*/
   /* =============================================================================================== */
   /*  define constants and shared variables                                                          */
   /* =============================================================================================== */
   /*  the G stem is a group of global variables used for various things                              */
   G. = ''

   parse source G.0OpSystem . G.0me
   parse upper version G.0Interpreter .
   if abbrev(G.0Interpreter, 'REXX-REGINA') then
      G.0Interpreter = 'REGINA'
   select
      when G.0Interpreter == 'REXXSAA' then do
            if RxFuncQuery('SysLoadFuncs') then do
                  call RxFuncAdd 'SysLoadFuncs', 'REXXUTIL', 'SysLoadFuncs'
                  call SysLoadFuncs
               end
            /*  the DumpVars routine will use RXU.DLL if it is available.                             */
            /*  If you do not have RXU.DLL installed, you could just comment out these lines.         */
            if RxFuncQuery('RxVList') then do
                  call RxFuncAdd 'RxuInit', 'RXU', 'RxuInit'
                  call RxuInit
               end
         end
      when G.0Interpreter == 'REGINA' then do
            call RxFuncAdd 'SysLoadFuncs', 'regutil', 'SysLoadFuncs'
            call SysLoadFuncs
         end
      otherwise
         call Msg 'unknown Rexx interpreter "'G.0Interpreter'"'
         call Msg 'I expected REXXSAA or REGINA'
         call Egress 'Usage'
   end

   /* ----------------------------------------------------------------------------------------------- */
   /*  These can be used for showing messages. 'G.0me' is the name of this file. (When                */
   /*  a Rexx script is invoked from the command line, the name is in ALL CAPS, which                 */
   /*  is hard to read.  To get the correct capitalization of it, the only way I have                 */
   /*  found is a cumbersome call to SysFileTree(). And even that only corrects the                   */
   /*  filename, not the path.)                                                                       */
   /* ----------------------------------------------------------------------------------------------- */
   parse source . . G.0me
   call SysFileTree G.0me, 'tmp.', 'FO'
   G.0me_long = tmp.1
   drop tmp.
   G.0me = filespec('name', G.0me_long)
   G.0meID = '['G.0me']'
   G.0me_ind = copies(' ', length(G.0me))
   G.0meID_ind = copies(' ', length(G.0meID))

   /* ----------------------------------------------------------------------------------------------- */
   /*  see if this is useful-- a set of "constants" that are available wherever                       */
   /*  needed, so I do not need to add to Shared_vars every time I invent a new global                */
   /*  variable.                                                                                      */
   /*  later-- Keep it like this. Granted, having to type 'G.0' every time is annoying. The           */
   /*  bigger advantage of it is that EXPOSEing a long list of variables means it takes more          */
   /*  time to call a procedure. The program runs much faster if most of the global variables         */
   /*  are combined into one compound variable.                                                       */
   /* ----------------------------------------------------------------------------------------------- */
   G.0version = '1.2'
   if G.0OpSystem == 'UNIX' then
      G.0CRLF = '0A'x
   else
      G.0CRLF = '0D0A'x
   G.0footerID = ' === End of '
   /*  never use the literal comment strings in a Rexx file                                           */
   G.0CommentBegin = '/'||'*'
   G.0CommentEnd = '*'||'/'
   parse value SysTextScreenSize() with G.0ScreenRows G.0ScreenCols
   /*  G.0sep is sometimes used as a field separator in strings                                       */
   G.0sep = x2c('1F')
   /*  Some functions will set G.0fatalerror if they have a serious problem                           */
   G.0fatalerror = 0
   /*  see if this is useful; a flag to show that a task made some change to the file                 */
   G.0FileChanged = 0

   /* =============================================================================================== */
   /*  keywords that start a clause:                                                                  */
   /*  note that '::' is not included. It is not needed and would break the ReCap task                */
   /*  because '::' would not be a legal variable name in Rexx                                        */
   G.0Keywords = 'Activate Alias Attrib Beep Call Cancel CD CDD CHCP Cls Color Copy' ,
                 'Date Del Delay Describe Detach Dir DirHistory Dirs DPath' ,
                 'DrawBox DrawHline DrawVline' ,
                 'Echo Echo. Echoerr Echoerr. Echos Echoserr EndLocal Eset Except Exit' ,
                 'Ffind For Free' ,
                 'Global Gosub Goto Help History If Inkey Input' ,
                 'Keybd Keys Keystack List LoadBTM Log MD MkDir Memory Move On Option' ,
                 'Path Pause Popd Prompt Pushd Quit RD Reboot Rem Ren Return' ,
                 'Screen Scrput Select Set SetDOS SetLocal Shift ShrAlias Start' ,
                 'Tee Time Timer Title Touch Tree Type UnAlias UnSet' ,
                 'Ver Verify Vol VscrPut Window Y' ,
                 'Do EndDo Leave Iterate' ,
                 'Switch Case Default EndSwitch' ,
                 'Iff ElseIff Else EndIff' ,
                 'Text EndText' ,
                 '?'

   /*  fix: add more subkeywords. This is for ReCap task                                              */
   G.0Keywords2 = 'Then In Do Min Max Restore Close Break Error ErrorMsg To While Until Forever'

   G.0OPLogical = 'Not .And. .Or. .Xor.'
   G.0OPCompare = 'eq lt gt le ge ne == !='
   G.0Conditions = 'Exist Defined Errorlevel IsAlias IsDir DirExist IsInternal IsLabel'

   /* =============================================================================================== */

   /* ----------------------------------------------------------------------------------------------- */
   /*  set some other variables                                                                       */
   /* ----------------------------------------------------------------------------------------------- */
   /*  most procedures will need to use "expose (Shared_vars)"                                        */
   Shared_vars = 'sigl Opt. G. all_tokens line_vars debug.'
   /*  line_vars is only needed by procedures that use the arrays of input and output lines           */
   line_vars = 'SrcIn. SrcOut.'
   /*  all_tokens is a list of all the t_xxx. arrays that describe the tokens.                        */
   /*  anything that uses tokens must expose them.                                                    */
   all_tokens = 't_class. t_type. t_prefix. t_val. t_line. t_col. t_indent. t_readonly.'

   /*  G.0token_vars tells the insert/delete functions which token arrays to act on.                  */
   /*  Since inserting/deleting tokens is so slow, only include the arrays that are                   */
   /*  needed by the tasks. The t_col. array is not included, because it is only                      */
   /*  used at the very beginning. The t_line array is included so it can be used for                 */
   /*  error messages.                                                                                */
   /*  If you create another token array, remember to add it to the lists                             */
   G.0token_vars = 't_class. t_type. t_prefix. t_val. t_line.'
   /*    G.0track_vars is a temporary variable I can use for debugging purposes                       */
   G.0track_vars = ''
   return 0
   /* === End of Init_Global ======================================================================== */

   Init_Config: /*fold00*/
   /* =============================================================================================== */
   /*  a list of configuration variables that can be changed on command line                          */
   /* ----------------------------------------------------------------------------------------------- */
   /*  To create a new command-line switch "-xyz" you need to                                         */
   /*     1) add a line to the list in the form                                                       */
   /*             Opt.0xyz    = <default value>    /@  a short description $ <possible values>  @/    */
   /*        note the $ that separates the two parts of the comment                                   */
   /*     2) add the string 'xyz' to the variable ValidSwitch                                         */
   /* =============================================================================================== */
   /*  This routine is read by ShowConfig to get a list of all the variable names. It looks for       */
   /*  lines that match                                                                               */
   /*       Opt.0<name> = xxx                                                                         */
   /*  the values listed below are defaults. Showconfig will show the actual values after             */
   /*  the command line switches are read.                                                            */
   /*                                                                                                 */
   /*  Doing it this way means I do not have to maintain a separate list for documentation,           */
   /*  and it lets me show the names with the proper capitalization. The alternative would be         */
   /*  having Rexx dump the variable list as OPT.0WIDTH, OPT.0FORMATCOMMENTS, etc.,                   */
   /*  which is hard to read.                                                                         */
   /*                                                                                                 */
   /*  IMPORTANT: do not change the layout of this routine unless you understand how                  */
   /*  ShowConfig parses it                                                                           */
   /* ----------------------------------------------------------------------------------------------- */

   /* =============================================================================================== */
   if translate(arg(1)) = 'SHOW' then do
         call ShowConfig
         call Egress 'Okay'
      end
   /* ----------------------------------------------------------------------------------------------- */
   /*  $format comment off                                                                            */
   Opt. = ''
   /*  General options:                                                                               */
   Opt.0config          = 0        /*  show this list of options                          $ 0 or 1    */
   Opt.0debug           = 'loop,fatal'    /*  debug categories   See "help debug"  $ <list>    */
   Opt.0help            = ''       /*  -help or help <switch> or help <subject>           $ <string>  */
   Opt.0idle            = 1        /*  use lower (idle) priority                          $ 0 or 1    */
   Opt.0Progress        = 1        /*  show progress                                      $ 0 or 1    */
   Opt.0verbose         = 0        /*  verbose error messages                             $ 0 or 1    */
   Opt.0Warn            = 1        /*  Tokeniser warns about faulty code                  $ 0 or 1    */
   /*  Formatting options:                                                                            */
   Opt.0Bare            = 0        /*  do a bare format only                              $ 0 or 1    */
   Opt.0IndentContinued = .5       /*  how much to indent continued lines                 $ <number>  */
   Opt.0Indent          = 1        /*  do indenting of Rexx code                          $ 0 or 1    */
   Opt.0Keywords        = ''       /*  how to recap keywords               $ UPPER,LOWER,MIXED, or '' */
   Opt.0Margin          = 2        /*  number of leading spaces (left margin)             $ <number>  */
   Opt.0ReCap           = 1        /*  recapitalize symbols and keywords                  $ 0 or 1    */
   Opt.0Style           = ''       /*  built-in formatting style to use             $ '' or <string>  */
   Opt.0Tab             = 3        /*  number of spaces per indent level                  $ <number>  */
   Opt.0FTE             = 1        /*  protect FTE/eFTE comments                          $ 0 or 1    */
   /*  Check options:                                                                                 */
   Opt.0Check           = 'ERROR,WARN' /*  what warnings to show                          $ <list>    */
   /*  $format comment on                                                                             */

   ValidSwitch = 'config debug help idle progress verbose warn' ,
                 'bare indentcontinued indent keywords margin recap' ,
                 'style tab check fte'

   return 0
   /* === End of Init_Config ======================================================================== */

   Init_CommandLine: /*fold00*/
   /* =============================================================================================== */
   /*  interpreting the command line:                                                                 */
   /*  Override default values of Opt.0xxx variables by reading the command line.                     */
   /*  respond to -version switch                                                                     */
   /*  determine name of input file and verify that it exists                                         */
   /*  determine name of output file                                                                  */
   /* =============================================================================================== */

   /*  Usually I just use "interpret GetOpts(MainArgs, ValidSwitch)", but I want a script             */
   /*  to be able to set its own defaults, then re-evaluate the command-line switches with            */
   /*        "interpret G.0ConfigString"                                                              */
   G.0ConfigString = GetOpts(MainArgs, ValidSwitch)
   interpret G.0ConfigString
   if Opt.0ShowGetOpts \== '' then do
         call ShowGetOpts
         call Egress 'Okay'
      end
   if Opt.0errors > 0 then
      /*  the message was already displayed when GetOpts() was interpreted                            */
      call Egress 'Usage'

   if Opt.0version \= '' then do
         say G.0me G.0version
         call Egress 'OKAY'
      end
   /* ----------------------------------------------------------------------------------------------- */
   do c = 1 to ArgC
      if pos('=', ArgV.c) > 0 then do
            call Msg ' "'ArgV.c'"'
            call Msg '     looks like a mistake. Did you mean'
            call Msg ' "-'ArgV.c'" ?'
            call Egress 'USAGE'
         end
   end c

   if FileExist(ArgV.1) then do
         call Msg '"'MainArgs'"'
         call Msg 'I need a command, like'
         call Msg '   'G.0me 'format' ArgV.1
         call Egress 'usage'
      end

   G.0Command = ArgV.1
   /*  I am shifting from using "-<option>" to "<OPTION>" as a command, but                           */
   /*  some commands do not use the normal <command> <infile> <outfile> syntax                        */
   select
      when Opt.0config ,
         | translate(G.0Command) == 'CONFIG' then do
            Opt.0config = 1
            G.0InfileN = G.0me_long
            G.0OutFileN = 'stdout'
         end
      when translate(G.0Command) == 'HELP' then do
            G.0InfileN = G.0me_long
            G.0OutFileN = 'stdout'
            if ArgV.2 \== '' then
               Opt.0help = ArgV.2
            else
               Opt.0help = 1
         end
      when Opt.0help \== '' ,
         & G.0Command == '' then do
            /*  this allows "-help=format" as an equivalent of the newer "HELP format"                */
            /*  one could argue that "help" should be a command instead of an option                  */
            /*  but many people will expect the program to understand "-help"                         */
            G.0Command = 'HELP'
            G.0InfileN = G.0me_long
            G.0OutFileN = 'stdout'
         end
      when translate(G.0Command) == 'DEVHELP' then do
            /*  this allows "BTMtool.cmd devhelp <where>" as an equivalent of                         */
            /*  "BTMtool.cmd showusage <path>BTMtool.cmd -where=<where>"                              */
            G.0InfileN = G.0me_long
            G.0OutFileN = 'stdout'
            if ArgV.2 \== '' then
               Opt.0Where = ArgV.2
            else
               Opt.0Where = ''
         end
      otherwise
         G.0InfileN = ArgV.2
         G.0OutFileN = ArgV.3
         if G.0OutFileN = '' then G.0OutFileN = 'stdout'
         if ArgV.4 \= '' then do
               list = ''
               do c = 2 to ArgC
                  list = list ArgV.c
               end c
               list = Quoted(strip(list))
               call Msg 'too many filenames:' list
               call Egress 'USAGE'
            end
   end   /*  select  */

   if G.0InfileN == '' then do
         call Msg 'I need a file name'
         call Egress 'Usage'
      end   /*  if InfileN == ''  */

   if \FileExist(G.0InfileN) then do
         call Msg 'Cannot find source file "'||G.0InfileN||'"'
         call Egress 'USAGE File_Not_Found'
      end

   /*  if outfile exists, do not overwrite it, add a suffix                                           */
   if FileExist(G.0OutFileN) then do
         suff = 2
         do while FileExist(G.0OutFileN||suff)
            suff = suff +1
         end
         NewN = G.0OutFileN||suff
         call Msg 'file exists:' G.0OutFileN
         call Msg 'changed to: ' NewN
         G.0OutFileN = NewN
      end
   return 0
   /* === End of Init_CommandLine =================================================================== */

   Setup: /*fold00*/
   /* =============================================================================================== */
   /*  get ready to start tasks:                                                                      */
   /*     re-interpret command line (to override any default values from the script)                  */
   /*     read the file and verify that it is not compiled or corrupt                                 */
   /*     determine the Rexx dialect, and set some variables based on it                              */
   /*     do a few other things based on command-line switches                                        */
   /* =============================================================================================== */
   interpret G.0ConfigString
   Opt.0debug = translate(Opt.0debug, ' ', ',')
   debug. = 0
   do w = 1 to words(Opt.0debug)
      wordw = translate(word(Opt.0debug, w))
      if wordw == 'ALL' then do
            debug. = 1
            debug.0fatal = 0
         end
      interpret 'debug.0'||wordw '= 1'
   end w
   /* ----------------------------------------------------------------------------------------------- */
   if Opt.0config then do
         call Init_Config 'SHOW'
         call Egress ret
      end
   /* ----------------------------------------------------------------------------------------------- */
   /*  initialize the output array. No task should initialize it or assume it is empty                */
   SrcOut. = ''
   SrcOut.0 = 0
   /*  read the input file into the SrcIn. array                                                      */
   SrcIn. = '' ; SrcIn.0 = 0
   if G.0InfileN \= '' then do
         call FileToArray 'SrcIn.', G.0InfileN
         do ln = 1 to SrcIn.0
            /*  StripStr() strips <Space> and <Tab>                                                   */
            SrcIn.ln = StripStr(SrcIn.ln, 't')
         end
         drop ln
      end   /*  if G.0InfileN \= ''  */
   /* ----------------------------------------------------------------------------------------------- */

   /* ----------------------------------------------------------------------------------------------- */
   if Opt.0idle then do
         /*  This is the highest priority that doesn't slow down my text editor                       */
         /*  in another window. You could also try                                                    */
         /*        SysSetPriority(0, -31)                                                             */
         ret = SysSetPriority(1, 31)
         if ret \= 0 then call Mark 'ret', 'from SysSetPriority, OS/2'
      end
   /* ----------------------------------------------------------------------------------------------- */
   t_readonly. = 0 ; t_readonly.0 = 0
   return 0
   /* === End of Setup ============================================================================== */

   ShowConfig: /*fold00*/
   /* =============================================================================================== */
   /*  Shows the current value of the configuration variables in the Opt. stem                        */
   /* =============================================================================================== */
   start = sigl
   varN. = ''
   varN.0 = 0
   v = 0
   do ln = start to sourceline()
      text = StripStr(sourceline(ln))
      if left(text, 2) = G.0CommentBegin then iterate
      if word(translate(text), 1) = 'RETURN' then leave
      parse var text varname '=' . (G.0CommentBegin) descr (G.0CommentEnd) .
      varname = strip(varname)
      if abbrev(translate(varname), 'OPT.0') then do
            v = v +1
            descr = strip(descr)
            varN.v = varname||' '||descr
            varN.0 = v
         end
   end
   ret = SysStemSort('varN.', 'a', 'i')
   varNameLen = 20
   Indent = varNameLen +4
   do v = 1 to varN.0
      parse var varN.v varN.v ' ' descr
      varVal = value(varN.v)

      varname = left('  '||'-'||substr(varN.v, 6), varNameLen)||'= '
      varVal =  left("'"||varVal||"'", 10)
      parse var descr descr '$' possible
      call lineout 'stdout', varname||varVal||left(possible, 30)||descr

      /*  alternate version for two lines per option:                                                 */
      /*        varname = left('  '||'-'||substr(varN.v, 6), varNameLen)||'= '                        */
      /*        varVal =  left("'"||varVal||"'", 10)                                                  */
      /*        descr = ChangeString('$', descr, G.0CRLF||copies(' ', 36))                            */
      /*        descr = ChangeString('$', descr, ''||copies(' ', 36))                                 */
      /*        call lineout 'stdout', varname||varVal||descr                                         */
   end
   call lineout 'stdout', '  For help, type 'G.0me' -help=<switch>'
   return 0
   /* === End of ShowConfig ========================================================================= */

   Prep_MarkRO: procedure expose (Shared_vars) (all_tokens) (line_vars) /*fold00*/
   /* =============================================================================================== */
   /*  HELP: Format $Format FTE eFTE readonly                                                         */
   /*                                                                                                 */
   /*    -FTE=1       do not change comments added by the FTE or eFTE folding text editor             */
   /*                 On by default. If you do not use FTE/eFTE, it should be harmless.               */
   /*                                                                                                 */
   /*  About the $FORMAT directive:                                                                   */
   /*                                                                                                 */
   /*  If none of the command-line switches do what you want, you can prevent part of                 */
   /*  a file from being formatted with a comment in the form                                         */
   /*      REM $FORMAT <class> <state>                                                                */
   /*  where <class> is 'comment' '4OS2' or 'all', and <state> is either 'on' or 'off'                */
   /*  "$FORMAT off" is understood as short for "$FORMAT all off".                                    */
   /*  The first word of the comment must be "$FORMAT". "COMMENTS" is the same as "COMMENT".          */
   /*  Case does not matter, and there can be other text after <state>.                               */
   /*                                                                                                 */
   /*  Examples:                                                                                      */
   /*    |  REM  $FORMAT all off                                                                      */
   /*    |  REM  $FORMAT comment off                                                                  */
   /*    |  REM  $FORMAT 4OS2 on but do not touch my carefully arranged comments                      */
   /*                                                                                                 */
   /*  ==> In this context, "formatting" means "changing the spacing or order of words".              */
   /*  ==> $FORMAT does not control recapitalization of words. For now, the only way to               */
   /*  ==> turn off recapitalization is to do it globally with the -ReCap=0 switch.                   */
   /*  ==> $FORMAT also will not prevent changes to variable names by the Change command.             */
   /* =============================================================================================== */
   /*  a formatting task should check the t_readonly attribute before changing a token:               */
   /*            if \t_readonly.t then                                                                */
   /*               <do something>                                                                    */
   /*                                                                                                 */
   /*  However, an indenting task should be allowed to set the t_indent.N attribute                   */
   /*  of token N. It will be ignored when applying the indenting, and it is                          */
   /*  information that might (conceivably) be needed for something.                                  */
   /* ----------------------------------------------------------------------------------------------- */
   if Opt.0Progress then call Progress SubID()

   t_readonly. = 0 ; t_readonly.0 = t_type.0
   /*  map "ON","OFF" to the t_readonly.t flag. you could allow synonyms like "true" or "yes"         */
   Formatting.on = 0
   Formatting.off = 1
   RO_4OS2 = 0
   RO_Comment = 0

   t = 1
   do while t < t_type.0
      t = t +1
      /*  set permission for the token:                                                               */
      select
         when t_class.t == 'COMMENT' then
            t_readonly.t = RO_Comment
         when t_class.t == 'TEXT' then
            t_readonly.t = 1
         otherwise
            t_readonly.t = RO_4OS2
      end

      /*  check comment for new $FORMAT instruction                                                   */
      if abbrev(t_type.t, 'COMMENT_VAL') then do
            Tval = StripStr(translate(t_val.t))
            select
               /*  if a $format instruction, read it                                                  */
               when abbrev(Tval, '$FORMAT ') then do
                     BadFmt = 0
                     parse upper var t_val.t . class  state  .
                     if class = 'ON' ,
                      | class = 'OFF' then do ; state = class ; class = 'ALL' ; end
                     /*  allow "comments" as well as "comment"                                        */
                     if class = 'COMMENTS' then class = 'COMMENT'
                     if state \= 'ON' ,
                      & state \= 'OFF' then BadFmt = 1
                     if class \= 'COMMENT' ,
                      & class \= '4OS2' ,
                      & class \= 'ALL' then BadFmt = 1

                     if \BadFmt then do
                           if class = 'ALL' ,
                            | class = '4OS2' then
                              RO_4OS2 = Formatting.state
                           if class = 'ALL' ,
                            | class = 'COMMENT' then do
                                 RO_Comment = Formatting.state

                                 /*  are $FORMAT comments themselves formattable? Let's say No:       */
                                 do s = t -1 to t
                                    t_readonly.s = 1
                                 end

                              end   /*  end if class  */
                        end   /*  if BadFmt  */
                     else do
                           call Msg G.0meID 'unrecognized $FORMAT instruction at line' t_line.t':'
                           call Msg '   "'||strip(t_val.t)||'"'
                        end   /*  end else  */
                  end   /*  when $format  */

               otherwise
                  /*  akm: this is for the FTE folding editor that I use; I do not want               */
                  /*  to tamper with its comments. Mark them as ReadOnly.                             */
                  /*  by default FTE and eFTE use "fold00", "fold01", etc.                            */
                  /*  On by default; if you do not use FTE/eFTE, it should be harmless.               */
                  if Opt.0FTE then
                     if length(t_val.t) == 6 then
                        if left(translate(t_val.t), 4) == 'FOLD' then
                           if datatype(substr(t_val.t, 5), 'W') then do
                                 do s = t -1 to t
                                    t_readonly.s = 1
                                 end
                              end
            end   /*  select  */
         end   /*  if a COMMENT_VAL  */
      /* -------------------------------------------------------------------------------------------- */
   end   /*  of t loop  */
   if Opt.0Progress then call Progress
   return 0
   /* === End of Prep_MarkRO ======================================================================== */

   Script_Bare: /*fold00*/
   /* =============================================================================================== */
   /*  HELP: Bare                                                                                     */
   /*    -Bare                   makes a new file of code only, without comments or empty lines       */
   /*                            other formatting tasks like Check, ReCap, and Indent are done        */
   /* =============================================================================================== */

   call Setup
   call Call_Toker
   if debug.0token then call SaveTokensT G.0InfileN||'_Tokens_Begin'

   call Task_Check
   call Task_Recap
   call Task_Indent
   /*  apply the indenting that was calculated above                                                  */
   call Apply_Indent
   if debug.0token then call SaveTokensF G.0InfileN||'_Tokens_End'

   /*  put the tokens together into lines and send to a file                                          */
   call Build_Bare
   call ArrayToFile 'SrcOut.', G.0OutFileN
   call Msg ''
   call Msg G.0meID 'wrote new file 'G.0OutFileN
   call Egress 'okay'
   /* === End of Script_Bare ======================================================================== */

   Script_CheckOnly: /*fold00*/
   /* =============================================================================================== */
   /*  HELP: Check CheckOnly                                                                          */
   /*                                                                                                 */
   /*     -CheckOnly              only checks the file for errors, no formatting                      */
   /*                                                                                                 */
   /* =============================================================================================== */
   call Setup
   call Call_Toker
   if debug.0token then call SaveTokensT G.0InfileN||'_Tokens_Begin'

   /* ----------------------------------------------------------------------------------------------- */
   /*  first, tasks that need to know line numbers, because the t_line. array will not be             */
   /*  available later. And anything else that does not change the tokens.                            */
   ret = Task_Check()
   if ret > 0 then
      call Egress 'file'
   else
      call Egress 'okay'
   /* === End of Script_CheckOnly =================================================================== */

   Script_Depends: /*fold00*/
   /* =============================================================================================== */
   /*  HELP: Depend Depends External Alias                                                            */
   /*                                                                                                 */
   /*     Depends                show a list of commands that are not recognized as 4OS2 keywords     */
   /*                            They may be aliases, external commands, or variables                 */
   /*                            For aliases and variables, tries to show their values                */
   /*                                                                                                 */
   /* =============================================================================================== */
   call Setup

   indent1 = copies(' ', 4)
   indent2 = copies(' ', 8)
   indent3 = copies(' ', 12)
   indent4 = copies(' ', 16)
   indent5 = copies(' ', 20)
   call Call_Toker
   if debug.0token then call SaveTokensT G.0InfileN||'_Tokens_Begin'

   call Task_CmdLists
   aliases = translate(cmd.alias)
   call Out 'external commands used:'
   do w = 1 to words(cmd.external)
      cmdN = word(cmd.external, w)
      if wordpos(translate(cmdN), aliases) > 0 then do
            call Out indent1||cmdN
            !aliasN = c2x(translate(cmdN))
            do x = 1 to WordsQ(alias.!aliasN, '1F'x)
               call Out indent2||'alias for' WordQ(alias.!aliasN, x, '1F'x)
            end x
         end
      else
         call Out indent1||cmdN
   end w
   /*  if a variable is defined in terms of another variable, expand it.                              */
   /*  4 levels of expansion should be enough, I hope                                                 */
   call Out 'variables used as commands:'
   do w = 1 to words(cmd.var)
      cmdN = word(cmd.var, w)
      call Out indent1||cmdN
      !setN = c2x(translate(word(strip(cmdN, 'l', '%'), 1)))
      do x = 1 to WordsQ(set.!setN, '1F'x)
         setval = WordQ(set.!setN, x, '1F'x)
         call Out indent2||'which is' setval
         if left(setval, 1) == '%' then do
               !setN2 = c2x(translate(word(strip(setval, 'l', '%'), 1)))
               do y = 1 to WordsQ(set.!setN2, '1F'x)
                  setval = WordQ(set.!setN2, y, '1F'x)
                  call Out indent3||'which is' setval
                  if left(setval, 1) == '%' then do
                        !setN3 = c2x(translate(word(strip(setval, 'l', '%'), 1)))
                        do z = 1 to WordsQ(set.!setN3, '1F'x)
                           setval = WordQ(set.!setN3, z, '1F'x)
                           call Out indent4||'which is' setval
                           if left(setval, 1) == '%' then do
                                 !setN4 = c2x(translate(word(strip(setval, 'l', '%'), 1)))
                                 do a = 1 to WordsQ(set.!setN4, '1F'x)
                                    setval = WordQ(set.!setN4, a, '1F'x)
                                    call Out indent5||'which is' setval
                                 end a
                              end
                        end z
                     end
               end y
            end
      end x
   end w


   if SrcOut.0 > 0 then do
         call ArrayToFile 'SrcOut.', G.0OutFileN
      end

   call Egress 'okay'
   /* === End of Script_Depends ===================================================================== */

   Script_Format: /*fold00*/
   /* =============================================================================================== */
   /* =============================================================================================== */

   call Setup
   /*  this is the tokenising stage.                                                                  */
   /*  If you want to do something special you can skip or replace some of these, but                 */
   /*  you should know what you are doing.                                                            */
   call Call_Toker
   if debug.0token then call SaveTokensT G.0InfileN||'_Tokens_Begin'

   /*  things that prepare the tokens for the formatting tasks                                        */
   call Prep_MarkRO
   /* ----------------------------------------------------------------------------------------------- */
   /*  first, tasks that need to know line numbers, because the t_line. array will not be             */
   /*  available later. And anything else that does not change the tokens.                            */
   call Task_Check
   /* ----------------------------------------------------------------------------------------------- */

   /* ----------------------------------------------------------------------------------------------- */
   /*  then tasks that make changes to tokens:                                                        */
   /*  inserting, deleting, or moving tokens must be done before indenting is calculated              */
   /*  re-spacing should come after tasks that move tokens or change spacing                          */
   /*  re-capitalizing could be done almost any time, but I put it here                               */
   /* ----------------------------------------------------------------------------------------------- */
   /*     call Task_Headers                                                                           */
   /*     call Task_ReSpace                                                                           */
   call Task_Recap
   /* ----------------------------------------------------------------------------------------------- */

   /* ----------------------------------------------------------------------------------------------- */
   /*  calculate indenting                                                                            */
   /*  This must come before anything that needs to know the line length                              */
   call Task_Indent
   call Task_Indent2
   /* ----------------------------------------------------------------------------------------------- */

   /* ----------------------------------------------------------------------------------------------- */
   /*  things that may depend on knowing the length of lines, like adjusting the right side           */
   /*  of comments, wrapping long lines.                                                              */
   /*     call Task_Wrap                        /*  must be after Indent   */                         */
   /*     call Task_UnIndentComments            /*  must be after Indent   */                         */

   /* ----------------------------------------------------------------------------------------------- */
   /*  apply the indenting that was calculated above                                                  */
   call Apply_Indent
   /*  if needed, save a list of tokens in their final state                                          */
   if debug.0token then call SaveTokensF G.0InfileN||'_Tokens_End'
   /*  put the tokens together into lines and send to a file                                          */
   call Output
   call Egress 'okay'
   /* === End of Script_Format ====================================================================== */

   Script_Help: /*fold00*/
   /*=================================================================================================*/
   /*=================================================================================================*/
   select
      when Opt.0help == 1 then do
            /*  $format off                                                                           */
         say  ''
         say  ' 'G.0me 'indents and reformats a 4OS2 BTM file'
         say  '    Usage:'
         say  '        'G.0me '<command> <old file> [<new file>] [options]'
         say  '    if <new file> is not given, output will go to STDOUT'
         say  '  '
         say  '    Commands:'
         say  '       help             this is it'
         say  '       help <subject>   longer description of a switch or subject'
         say  '                        The descriptions are extracted from subroutine headers'
         say  '       Depends          show a list of external commands, aliases, and variables used'
         say  '                        as commands. Aliases and variables will be expanded when possible'
         say  ''
         say  ''
         say  '    some options:'
         say  '       -Config          shows all the options, their default values, and a short description'
         say  '                        any of them can be changed on the command line'
         say  ''
         say  '    Options that are mostly for development and testing:'
         say  '       -debug="<type>[,<type>]"'
         say  '                        a list of debugging types. Currently "token", "valid", and "loop".'
         say  '       -ShowGetOpts     show how the command line was interpreted. Try this if you don''t'
         say  '                        get the results you expect from a switch.'
         say  ''
   /*           say  '       -ShowTools       shows header comments of the utility functions. Type'   */
   /*           say  '                           "'G.0me '-ShowTools"'                                */
   /*           say  '       -ShowTasks     shows header comments of the task functions. Type'        */
   /*           say  '                           "'G.0me '-ShowTasks"'                                */
            /*  $format on                                                                            */
         end
      otherwise do
            /*  show help for specific subject from a Task_xxx or Script_xxx routine                  */
            /*  the first line of the subroutine header must be  HELP: <subject> <subject> ...        */
            SrcIn. = '' ; SrcIn.0 = 0
            call FileToArray 'SrcIn.', G.0me_long
            l = 0                                                           /* current line in SrcIn. */
            Opt.0help = strip(Opt.0help, 'L', '-')
            do while l < SrcIn.0
               l = l +1
               if Rexx_IsLabel(SrcIn.l) then do
                     /*  remember the name of the procedure                                           */
                     parse value SrcIn.l with label ':' .
                     label = strip(label)
                     if abbrev(translate(label), 'PREP_') ,
                      | abbrev(translate(label), 'HELP_') ,
                      | abbrev(translate(label), 'TASK_') ,
                      | abbrev(translate(label), 'SCRIPT_') then do
                           l = l +1
                           if Rexx_DividerType(SrcIn.l) == 3 then do
                                 /*  it is a "===" divider after the label                            */
                                 l = l +1
                                 HelpLine = translate(strip(Rexx_UnComment(SrcIn.l)))
                                 parse var HelpLine Word1 topics

                                 if Word1 == 'HELP:' then do
                                       if wordpos(translate(Opt.0help), topics) > 0 then do
                                             l = l +1
                                             say 'HELP for' label':'
                                             /*                                            say ''     */
                                             do while l < SrcIn.0
                                                /*  copy everything until the next divider or non-comment */
                                                /*  in order to use multiline comments in the header, */
                                                /*  I had to remove "if \Rexx_IsComment()", which means */
                                                /*  the header =must= end with a divider              */
                                                /*  if \Rexx_IsComment(SrcIn.l) then leave            */
                                                if Rexx_DividerType(SrcIn.l) > 0 then do
                                                      say '   'Rexx_UnComment(SrcIn.l)
                                                      leave
                                                   end
                                                say Rexx_UnComment(SrcIn.l)
                                                l = l +1
                                             end
                                          end   /*  if wordpos()  */
                                    end   /*  if HELP:  */
                              end   /*  if found starting divider  */
                        end   /*  if TASK or SCRIPT  */
                  end   /*  if is a label  */
            end   /*  while l SrcIn.0  */
         end   /*  otherwise do  */
   end   /*  select  */
   call Egress 'okay'
   /* === End of Script_Help ======================================================================== */

   Script_Test: /*fold00*/
   /* =============================================================================================== */
   /*  Help: Test                                                                                     */
   /*  the Test script is for testing whatever bit I am currently working on                          */
   /* =============================================================================================== */
   call Setup
   call Call_Toker

   /*     do i = 2 to t_type.0                                                                        */
   /*        say i '"'t_val.i'"'  'in function "'T_Infunction(i)'"'                                   */
   /*     end                                                                                         */

   do  t = 2 to t_type.0 -1
      if t_class.t == 'QUOTED' ,
       | t_type.t == 'W_LITERAL' then do
            s = R_StartQuote(t)
            u = R_StopQuote(t)
            str = S_Quoted(t)
            call Mark 't t_val.t t_type.t s t_type.s u t_type.u str', 'found quotes'
         end
   end t

   if debug.0token then call SaveTokensT G.0InfileN||'_Tokens_Begin'

   call Egress 'okay'
   /* === End of Script_Test ======================================================================== */

   Task_CmdLists: procedure expose (Shared_vars) (line_vars) (all_tokens) cmd. alias. set. /*fold00*/
   /* =============================================================================================== */
   /*  returns                                                                                        */
   /*     cmd.external        a list of commands that are not recognized as keywords                  */
   /*     cmd.alias           a list of aliases that were defined                                     */
   /*     cmd.var             a list of variables used as commands                                    */
   /*     set.xxx             value of variable xxx                                                   */
   /*     alias.xxx           value of alias xxx                                                      */
   /* =============================================================================================== */
   if Opt.0Progress then call Progress G.0meID SubID()
   cmd. = ''
   alias. = ''
   set. = ''
   do t = 2 to t_type.0 -2
      select
         when abbrev(t_type.t, 'W_CMD') then do
               if MatchCmd(word(t_val.t, 1), cmd.external) == 0 then
                  cmd.external = cmd.external word(t_val.t, 1)
               /*  fix: why would t_val.t be more than one word?                                      */
            end
         when t_type.t == 'W_KEY' then do
               if translate(t_val.t) == 'ALIAS' then do
                     aliasN = T_Val(t +1)
                     if wordpos(translate(aliasN), translate(cmd.alias)) == 0 then
                        cmd.alias = cmd.alias aliasN
                     !aliasN = c2x(translate(aliasN))
                     alias.!aliasN = alias.!aliasN||'1F'x||T_AsString(t +3, R_LastWord(t +3))
                  end
               if translate(t_val.t) == 'SET' then do
                     setN = T_Val(t +1)
                     if wordpos(translate(setN), translate(cmd.set)) == 0 then
                        cmd.set = cmd.set setN
                     !setN = c2x(translate(setN))
                     thisval = T_AsString(t +3, R_LastWord(t))
                     if thisval \= '' then do
                           /*  in the case of "set foo=", thisval will be ''                          */
                           thisval = '1F'x||thisval
                           if pos(thisval, set.!setN) == 0 then
                              set.!setN = set.!setN||thisval
                        end
                  end
               if translate(t_val.t) == 'CALL' then do
                     if T_Type(t +1) == 'W_%' then do
                           varN = R_Expression(t +2)
                           if wordpos(translate(varN), translate(cmd.var)) == 0 then
                              cmd.var = cmd.var varN
                        end
                  end
            end   /*  when W_KEY  */
         when t_type.t == 'EOC' then do
               if T_Type(t +1) == 'W_%' then do
                     varN = R_Expression(t +2)
                     if wordpos(translate(varN), translate(cmd.var)) == 0 then
                        cmd.var = cmd.var varN
                  end
            end
         otherwise nop
      end   /*  select  */
   end t
   cmd.external = WordSort(cmd.external)
   cmd.alias = WordSort(cmd.alias)
   cmd.var = WordSort(cmd.var)
   if Opt.0Progress then call Progress
   return 0
   /* === End of Task_CmdLists ====================================================================== */

   Task_Indent: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Help: Indent Margin Tab IndentContinued Continued                                              */
   /*                                                                                                 */
   /*    -Indent=0               do not change indenting                                              */
   /*                                                                                                 */
   /*    -Margin=3               add a three-space left margin in addition to any indenting           */
   /*    -Margin=0               no margin                                                            */
   /*                                                                                                 */
   /*    -Tab=3                  indent each level by 3 spaces                                        */
   /*                                                                                                 */
   /*    -IndentContinued=2      indent continued lines by 2 spaces                                   */
   /*                                                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  This routine only calculates the logical indenting for tokens.  Later tasks may                */
   /*  change the indenting. After the lines are re-assembled the leading spaces will                 */
   /*  be added by Apply_Indent().                                                                    */
   /* =============================================================================================== */
   /*  As much as possible I try to not think in terms of lines here. The indent value is how much    */
   /*  a word =would= be indented, if that word were at the beginning of a line, if it were legal     */
   /*  for that word to start a line. Most tokens do not start a line, but it is much easier to       */
   /*  store an indent value for a token than to figure out whether it really needs it.               */
   /* ----------------------------------------------------------------------------------------------- */

   if \Opt.0Indent then return 0
   if Opt.0Progress then call Progress G.0meID SubID()

   if Opt.0Tab < 0 then do
         call ShowError 'Error: cannot use -Tab='Opt.0Tab
         call Egress 'usage'
      end
   if Opt.0Margin < 0 then do
         call ShowError 'Error: cannot use -Margin='Opt.0Margin
         call Egress 'usage'
      end

   t_indent. = 0
   t_indent.0 = t_type.0
   Tval = ''

   Indent = 0
   do t = 2 to t_type.0
      if t_class.t \= '4OS2' then do
            t_indent.t = Indent
            iterate
         end
      Tval = translate(t_val.t)
      select
         when Tval == 'IFF' then do
               t_indent.t = Indent
               Indent = Indent +1
            end
         when Tval == 'ELSEIFF' then do
               t_indent.t = Indent -1
            end
         when Tval == 'ELSE' then do
               t_indent.t = Indent -1
            end
         when Tval == 'ENDIFF' then do
               t_indent.t = Indent -1
               Indent = Indent -1
            end
         when Tval == 'SWITCH' then do
               t_indent.t = Indent
               Indent = Indent +1
            end
         when Tval == 'CASE' then do
               t_indent.t = Indent -1
            end
         when Tval == 'DEFAULT' then do
               t_indent.t = Indent -1
            end
         when Tval == 'ENDSWITCH' then do
               t_indent.t = Indent -1
               Indent = Indent -1
            end
         when Tval == 'DO' then do
               t_indent.t = Indent
               Indent = Indent +1
            end
         when Tval == 'ENDDO' then do
               t_indent.t = Indent -1
               Indent = Indent -1
            end
         when Tval == '(' then do
               t_indent.t = Indent
               Indent = Indent +1
            end
         when Tval == ')' then do
               t_indent.t = Indent -1
               Indent = Indent -1
            end
         otherwise
            t_indent.t = Indent
      end   /*  select  */

      if Indent < 0 then do
            call ShowError 'Warning: negative indent at token 't '"'t_val.t'" near line 't_line.t
            call ShowError 'Probably defective code, like a missing DO', ''
            call Mark 't t_class.t t_type.t t_val.t t_line.t t_indent.t Indent'
            t_indent.t = 0
            Indent = 0
         end
   end t   /*  next token  */

   /*  quick fix for the new-style tokenising, copy indent of first word to "*" or "@"                */
   /*  work backwards in case someone uses "@*cd "                                                    */
   do t = t_type.0 to 1 by -1
      if t_type.t == '@' ,
       | t_type.t == '*' then
         t_indent.t = T_Indent(t +1)
   end t


   if Opt.0Progress then call Progress

   return 0
   /* === End of Task_Indent ======================================================================== */

   Task_Indent2: procedure expose (Shared_vars) (line_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Help: Indent IndentContinued Continued Outdent ThenDo OutdentThendo Format                     */
   /*                                                                                                 */
   /*    -IndentContinued=''     If indenting is needed, continued lines will be indented             */
   /*                            as a block, to preserve their relative position.                     */
   /*                            In other words, do not change indenting of continued lines.          */
   /*                                                                                                 */
   /*    -IndentContinued=<amount>                                                                    */
   /*                            where <amount> is a multiple of the Tab value.                       */
   /*                            It does not need to be an integer.                                   */
   /*                                                                                                 */
   /*   For example, if the Tab value is 3 spaces:                                                    */
   /*       -IndentContinued=''     do not change indenting of continued lines                        */
   /*       -IndentContinued=0      no extra indent; align continued line with previous line          */
   /*       -IndentContinued=1      indent continued lines 3 spaces                                   */
   /*       -IndentContinued=.3     indent continued lines 1 space                                    */
   /*       -IndentContinued=1.6    indent continued lines 5 spaces                                   */
   /*                                                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  This task will contain adjustments to the default (logical) indenting to allow                 */
   /*  for individual styles. If you want add new options to the indenting, this is                   */
   /*  probably the place to do it.                                                                   */
   /* ----------------------------------------------------------------------------------------------- */
   if Opt.0Progress then call Progress SubID()

   /* ----------------------------------------------------------------------------------------------- */

   /* ----------------------------------------------------------------------------------------------- */
   /*  indent CONTINUED lines                                                                         */
   do t = 2 to t_type.0 -1
      if t_class.t == 'FILE_EOL' then do
            Word1 = t +1
            if L_IsContinued(Word1, -1) then
               t_indent.Word1 = t_indent.Word1 +Opt.0IndentContinued
         end   /*  if FILE_EOL  */
   end t
   /* ----------------------------------------------------------------------------------------------- */
   if Opt.0Progress then call Progress
   return 0
   /* === End of Task_Indent2 ======================================================================= */

   Task_Check: procedure expose (Shared_vars) (line_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  HELP: Check CheckOnly                                                                          */
   /*                                                                                                 */
   /*    -Check                  check for some common errors in addition to formatting               */
   /*                                                                                                 */
   /*        It looks for:                                                                            */
   /*              unbalanced or mismatched quotes                                                    */
   /*              unmatched DO/ENDDO or SWITCH/ENDSWITCH                                             */
   /*              Rexx keywords I might use by mistake                                               */
   /*              Duplicate labels                                                                   */
   /*                                                                                                 */
   /* =============================================================================================== */
   if Opt.0Check == '' then return 0

   /*  change "error,warn" to "ERROR WARN"                                                            */
   /*     Opt.0Check = translate(Opt.0Check)                                                          */
   /*     Opt.0Check = translate(Opt.0Check, ' ', ',')                                                */

   if Opt.0Progress then call Progress G.0meID SubID()
   ErrorCount = 0

   /* ----------------------------------------------------------------------------------------------- */
   /*  check for unbalanced quotes                                                                    */
   t = 1
   do while t <= t_type.0 -1
      t = t +1
      if t_type.t == 'W_"' then do
            str = S_Quoted(t)
            if right(str, 1) \= '"' then do
                  call CheckOut 'missing double quote at line' t_line.t
                  ErrorCount = ErrorCount +1
               end
            t = R_StopQuote(t)
         end
   end   /*  while t  */

   t = 1
   do while t <= t_type.0 -1
      t = t +1
      if t_class.t == 'COMMENT' then
         iterate
      if left(t_val.t, 1) == '`' then do
            t = R_StopQuote(t)
            if length(t_val.t) == 1 ,
             | right(t_val.t, 1) \= '`' then do
                  call CheckOut 'missing back quote at line' t_line.t
                  ErrorCount = ErrorCount +1
               end
         end
   end   /*  while t  */
   /* ----------------------------------------------------------------------------------------------- */
   /*  check for unmatched DO/ENDDO or SWITCH/ENDSWITCH                                               */
   /*                                                                                                 */
   Dolevel = 0
   t_from = 2
   if t_type.t_from == 'LABEL' then
      labelN = t_val.t_from
   else
      labelN = '<none>'
   t_to = 0
   do while t_to < t_type.0
      t_to = R_NextLabel(t_from)
      Dolevel = 0
      do t = t_from to t_to
         if t_type.t = 'W_KEY' then do
               if Dolevel == 0 then Last0 = t_line.t
               Tval = translate(t_val.t)
               select
                  when Tval == 'DO' then Dolevel = Dolevel +1
                  when Tval == 'SWITCH' then Dolevel = Dolevel +1
                  when Tval == 'ENDDO' ,
                     | Tval == 'ENDSWITCH' then do
                        Dolevel = Dolevel -1
                        if Dolevel < 0 then do
                              call CheckOut ,
                                   'Warning: unmatched' Tval 'at line 't_line.t' in label "'labelN'"'
                              ErrorCount = ErrorCount +1
                              Dolevel = 0
                           end
                     end
                  otherwise nop
               end   /*  end select Tval  */
            end
      end t   /*  next token  */
      if Dolevel > 0 then do
            Dolevel = '+'||Dolevel
            call CheckOut 'Warning: unbalanced DO/ENDDO or SWITCH/ENDSWITCH'
            call CheckOut '         in label "'labelN'" missing ENDDO or ENDSWITCH somewhere after line 'Last0
            call CheckOut '         the ending indent level is 'Dolevel
            ErrorCount = ErrorCount +1
         end
      t_from = t_to
   end   /*  while t_to  */
   /* ----------------------------------------------------------------------------------------------- */
   /*  check for things that I might use by mistake, mostly Rexx-isms                                 */
   RexxWords = 'SAY END'
   do t = 2 to t_type.0 -1
      /* -------------------------------------------------------------------------------------------- */
      select
         when t_type.t == 'W_CMD' then
            if translate(T_Val(t -1)) \= 'ALIAS' then do
                  /*  a REXX KEYWORD could be an alias, but more likely is a mistake                  */
                  keyVal = translate(t_val.t)
                  if wordpos(keyVal, RexxWords) > 0 then do
                        call CheckOut 'Warning: Rexx keyword "'t_val.t'" at line' t_line.t
                        call CheckOut Qline(t_line.t)
                        ErrorCount = ErrorCount +1
                     end
               end   /*  if W_CMD  */

         when t_type.t == 'W_KEY' then do
               keyVal = translate(t_val.t)
               nextVal = translate(T_Val(t +1))
               select
                  when (keyVal == 'IF' ,
                     | keyVal == 'IFF') ,
                     & nextVal == '\' then do
                        /*  "IF \" instead of "IF NOT"                                                */
                        call CheckOut 'Warning: Rexx keyword "'nextVal'" at line' t_line.t
                        call CheckOut Qline(t_line.t)
                        ErrorCount = ErrorCount +1
                     end
                  when keyVal == 'IF' ,
                     & translate(T_Val(R_LastWord(t))) == 'THEN' then do
                        /*  IF instead of IFF                                                         */
                        call CheckOut 'Warning: IF...THEN should be IFF...THEN, or just IF'
                        call CheckOut '         at line' t_line.t
                        call CheckOut Qline(t_line.t)
                        ErrorCount = ErrorCount +1
                     end
                  when keyVal == 'IFF' ,
                     & translate(T_Val(R_LastWord(t))) \= 'THEN' then do
                        /*  IFF without THEN                                                          */
                        call CheckOut 'Warning: IFF without THEN at line' t_line.t
                        call CheckOut Qline(t_line.t)
                        ErrorCount = ErrorCount +1
                     end
                  otherwise nop
               end   /*  select  */
            end   /*  if W_KEY  */

         when abbrev(t_type.t, 'W_OP') ,
            & t_val.t == '=' then do
               if T_Type(t -1) \= 'W_%' then do
                     /*  ASSIGNMENT                                                                   */
                     Word1 = translate(T_Val(R_FirstWord(t)))
                     if \MatchCmd(Word1, 'DO SET ALIAS OPTION') then do
                           if t < R_FirstWord(t) +2 then do
                                 call CheckOut 'Warning: missing SET or ALIAS at line 't_line.t
                                 call CheckOut Qline(t_line.t)
                                 ErrorCount = ErrorCount +1
                              end
                        end
                     if t_prefix.t \== '' ,
                      | T_Prefix(t +1) \== '' then do
                           if \MatchCmd(Word1, 'DO OPTION') then do
                                 call CheckOut 'Warning: space(s) around "=" at line 't_line.t
                                 call CheckOut Qline(t_line.t)
                                 ErrorCount = ErrorCount +1
                              end
                        end
                  end   /*  if \= "%"  */
            end   /*  if "="  */
         when t_type.t == 'W_SYMBOL' then do
               /*  SINGLE QUOTES are legal, but likely to be a mistake                                */
               if t_class.t \= 'QUOTED' then do
                     if left(t_val.t, 1) == "'" ,
                      | right(t_val.t, 1) == "'" then
                        if translate(T_Infunction(t)) \= '@REXX' then do
                              call CheckOut 'Warning: Rexx-style single quote at line 't_line.t
                              call CheckOut Qline(t_line.t)
                              ErrorCount = ErrorCount +1
                           end
                  end
            end
         otherwise nop
      end   /*  select  */
   end t

   /* ----------------------------------------------------------------------------------------------- */
   /*  check for duplicate labels                                                                     */
   LabelList = ''
   do t = 2 to t_type.0
      if t_type.t == 'LABEL' then do
            if wordpos(translate(t_val.t), LabelList) > 0 then do
                  call CheckOut 'Warning: duplicate label "'t_val.t'" at line 't_line.t
                  call CheckOut Qline(t_line.t)
                  ErrorCount = ErrorCount +1
               end
            else
               LabelList = LabelList||' '||translate(t_val.t)
         end
   end t
   /* ----------------------------------------------------------------------------------------------- */
   if Opt.0Progress then call Progress
   if ErrorCount > 0 then
      call CheckOut 'found' ErrorCount 'errors'
   else
      if translate(G.0Command) == 'CHECKONLY' then
         call CheckOut 'no errors found'
   return ErrorCount
   /* === End of Task_Check ========================================================================= */

   Task_Recap: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Help: ReCap Keywords Capitalization                                                            */
   /*                                                                                                 */
   /*    -ReCap                  recapitalize variables, procedures, Built-In Functions               */
   /*    -ReCap=0                do not recapitalize anything                                         */
   /*                                                                                                 */
   /*    -Keywords=upper         make all keywords upper case                                         */
   /*    -Keywords=lower         make all keywords lower case                                         */
   /*    -Keywords=mixed         make all keywords mixed case                                         */
   /*    -Keywords=''            do not change keywords                                               */
   /*                                                                                                 */
   /*    the name of an internal procedure is always capitalized to match its Label.                  */
   /*                                                                                                 */
   /*    the first place a variable is used will determine how it is capitalized.                     */
   /* =============================================================================================== */
   if \Opt.0ReCap then return 0
   Esigl = sigl
   ValidOpts = 'UPPER LOWER MIXED'
   if Opt.0Keywords \= '' ,
    & wordpos(translate(Opt.0Keywords), ValidOpts) == 0 then do
         call Msg 'unrecognized option: "-Keywords='Opt.0Keywords'"'
      end
   if Opt.0Progress then call Progress G.0meID SubID()

   /* ----------------------------------------------------------------------------------------------- */
   ProcList = ''
   FuncList = ''
   VarList = ''
   KeyList = ''
   LogicalList = ''
   CompareList = ''

   /*  Seed the procedure list                                                                        */
   /*  the label is the authoritative source for capitalization                                       */
   do t = 2 to t_type.0 -1
      if t_type.t == 'LABEL' then do
            if wordpos(translate(t_val.t), translate(ProcList)) == 0 then
               ProcList = ProcList t_val.t
         end   /*  if LABEL  */
   end t

   /*     BIFs are recognized by '%@', no need for a built-in list of names                           */

   /*  seed the keyword list with all the keywords that I know of:                                    */
   if Opt.0Keywords \= '' then do
         KeyList = G.0Keywords G.0Keywords2
         select
            when translate(Opt.0Keywords) == 'UPPER' then
               KeyList = translate(KeyList)
            when translate(Opt.0Keywords) == 'LOWER' then
               KeyList = translate(KeyList, 'abcdefghijklmnopqrstuvwxyz', 'ABCDEFGHIJKLMNOPQRSTUVWXYZ')
            otherwise nop
         end
      end   /*  if Opt.0keywords  */

   if Opt.0Logical \= '' then do
         OPlogical = G.0OPLogical
         select
            when translate(Opt.0Logical) == 'UPPER' then
               OPlogical = translate(OPlogical)
            when translate(Opt.0Logical) == 'LOWER' then
               OPlogical = translate(OPlogical, 'abcdefghijklmnopqrstuvwxyz', 'ABCDEFGHIJKLMNOPQRSTUVWXYZ')
            otherwise nop
         end
      end   /*  if Opt.0logical  */

   if Opt.0compare \= '' then do
         OPcompare = G.0OPCompare
         select
            when translate(Opt.0compare) == 'UPPER' then
               OPcompare = translate(OPcompare)
            when translate(Opt.0compare) == 'LOWER' then
               OPcompare = translate(OPcompare, 'abcdefghijklmnopqrstuvwxyz', 'ABCDEFGHIJKLMNOPQRSTUVWXYZ')
            otherwise nop
         end
      end   /*  if Opt.0compare  */

   /* ----------------------------------------------------------------------------------------------- */
   /*  if a symbol is recognized, correct the capitalization. If not, save it to the list of known    */
   /*  symbols. The result is that the symbol will be capitalized the way it was first used.          */
   do t = 2 to t_type.0
      select
         /* ----------------------------------------------------------------------------------------- */
         when t_type.t == 'W_PROCEDURE' then do
               NamePos = wordpos(translate(t_val.t), translate(ProcList))
               if NamePos == 0 then
                  ProcList = ProcList t_val.t
               else
                  if \t_readonly.t then
                     t_val.t = word(ProcList, NamePos)
            end
         /* ----------------------------------------------------------------------------------------- */
         when t_type.t == 'W_FUNCTION' then do
               NamePos = wordpos(translate(t_val.t), translate(FuncList))
               if NamePos == 0 then
                  FuncList = FuncList t_val.t
               else
                  if \t_readonly.t then
                     t_val.t = word(FuncList, NamePos)
            end
         /* ----------------------------------------------------------------------------------------- */
         when t_type.t == 'W_VARIABLE' then do
               /*  if it is already in the list of variable names, recap it according to the list.    */
               /*  otherwise, add it to the list                                                      */
               NamePos = wordpos(translate(t_val.t), translate(VarList))
               if NamePos == 0 then
                  VarList = VarList t_val.t
               else
                  if \t_readonly.t then
                     t_val.t = word(VarList, NamePos)
            end
         /* ----------------------------------------------------------------------------------------- */
         when abbrev(t_type.t, 'W_KEY') then do
               NamePos = wordpos(translate(t_val.t), translate(KeyList))
               if NamePos == 0 then
                  KeyList = KeyList t_val.t
               else
                  if \t_readonly.t then
                     t_val.t = word(KeyList, NamePos)
            end
         /* ----------------------------------------------------------------------------------------- */
         when abbrev(t_type.t, 'W_LOGIC') then do
               NamePos = wordpos(translate(t_val.t), translate(LogicalList))
               if NamePos == 0 then
                  LogicalList = LogicalList t_val.t
               else
                  if \t_readonly.t then
                     t_val.t = word(LogicalList, NamePos)
            end
         /* ----------------------------------------------------------------------------------------- */
         /* ----------------------------------------------------------------------------------------- */
         when abbrev(t_type.t, 'W_COMPARE') then do
               NamePos = wordpos(translate(t_val.t), translate(CompareList))
               if NamePos == 0 then
                  CompareList = CompareList t_val.t
               else
                  if \t_readonly.t then
                     t_val.t = word(CompareList, NamePos)
            end
         /* ----------------------------------------------------------------------------------------- */
         otherwise nop
      end
   end t
   if Opt.0Progress then call Progress
   return 0
   /* === End of Task_Recap ========================================================================= */

   Task_ShowTasks: procedure expose (Shared_vars) (line_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Help: ShowTasks Tasks                                                                          */
   /*                                                                                                 */
   /*    -ShowTasks              extracts headers from all of the Script_xxx and Task_xxx routines    */
   /*                                                                                                 */
   /* =============================================================================================== */
   if \Opt.0ShowTasks then return 0
   if Opt.0Progress then call Progress SubID()

   SrcOut. = '' ; SrcOut.0 = 0
   h = 0                                                      /* current line in header array SrcOut. */
   l = 0                                                       /* current line in source array SrcIn. */
   do while l <= SrcIn.0
      l = l +1
      if Rexx_IsLabel(SrcIn.l) then do
            /*  remember the name of the procedure                                                    */
            parse var SrcIn.l label ':' .
            label = strip(label)
            if abbrev(translate(label), 'TASK_') ,
             | abbrev(translate(label), 'SCRIPT_') then do
                  l = l +1
                  u = l +1
                  if Rexx_DividerType(SrcIn.l) == 3 then do
                        /*  subroutine header starts with "===" divider after the label               */
                        h = h +1 ; SrcOut.h = '' ; SrcOut.0 = h
                        h = h +1 ; SrcOut.h = '  '||Rexx_UnComment(SrcIn.l) ; SrcOut.0 = h
                        h = h +1 ; SrcOut.h = '  '||' ['label']' ; SrcOut.0 = h
                        l = l +1
                        do while l <= SrcIn.0
                           /*  copy everything until the next divider or non-comment                  */
                           if \Rexx_IsComment(SrcIn.l) then leave
                           h = h +1 ; SrcOut.h = '  '||Rexx_UnComment(SrcIn.l) ; SrcOut.0 = h
                           if Rexx_DividerType(SrcIn.l) == 3 then leave
                           l = l +1
                        end
                        h = h +1 ; SrcOut.h = '' ; SrcOut.0 = h
                     end   /*  if found starting divider  */
               end   /*  if TASK or SCRIPT  */
         end   /*  if is a label  */
   end   /*  while l <= SrcIn.0  */
   if Opt.0Progress then call Progress
   return 0
   /* === End of Task_ShowTasks ===================================================================== */

   Task_ShowTools: procedure expose (Shared_vars) (line_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Help: ShowTools                                                                                */
   /*                                                                                                 */
   /*    -ShowTools              show all Usage headers from procedures                               */
   /*    -ShowTools='L_'         show only functions that start with "L_"                             */
   /*    -ShowTools='L_First'    show only functions that start with "L_First"                        */
   /*                                                                                                 */
   /* =============================================================================================== */
   /*  extracts the usage headers from procedures.                                                    */
   /*  A usage header looks like this:                                                                */
   /*                                                                                                 */
   /*          Label:                                                                                 */
   /*          /@ ===================== @/                                                            */
   /*          Usage:                                                                                 */
   /*                <text>                                                                           */
   /*                                                                                                 */
   /*  This is intended for extracting headers from the formatter, but it could be                    */
   /*  used on any file that uses the same header style.                                              */
   /* =============================================================================================== */
   if Opt.0ShowTools == 0 then return 0
   if Opt.0Progress then call Progress SubID()
   if Opt.0ShowTools == 1 then
      Filter = ''
   else
      Filter = translate(Opt.0ShowTools)
   SrcOut. = '' ; SrcOut.0 = 0
   h = 0                                                      /* current line in header array SrcOut. */
   l = 0                                                       /* current line in source array SrcIn. */
   do while l <= SrcIn.0
      l = l +1
      if Rexx_IsLabel(SrcIn.l) then do
            parse var SrcIn.l label ':' .                       /* remember the name of the procedure */
            label = strip(label)
            /*  note that if Filter == '', anything will pass the abbrev() test,                      */
            /*  which is unintuitive but useful                                                       */
            if abbrev(translate(label), Filter) then do
                  l = l +1
                  u = l +1
                  Usage = strip((Rexx_UnComment(SrcIn.u)))
                  if Rexx_DividerType(SrcIn.l) == 3 ,
                   & abbrev(translate(Usage), 'USAGE:') then do
                        /*  subroutine header starts with "===" divider after the label               */
                        /*  and the next line is "Usage:"                                             */
                        h = h +1 ; SrcOut.h = '  '||Rexx_UnComment(SrcIn.l) ; SrcOut.0 = h
                        h = h +1 ; SrcOut.h = '  '||' ['label']' ; SrcOut.0 = h
                        l = l +1
                        do while l <= SrcIn.0
                           /*  copy everything until the next divider or non-comment                  */
                           if \Rexx_IsComment(SrcIn.l) then leave

                           if Rexx_DividerType(SrcIn.l) == 3 then leave
                           h = h +1 ; SrcOut.h = '  '||Rexx_UnComment(SrcIn.l) ; SrcOut.0 = h
                           l = l +1
                        end
                        if strip(SrcOut.h) \= '' then do
                              h = h +1 ; SrcOut.h = '' ; SrcOut.0 = h
                           end
                     end   /*  if Usage: header  */
                  else
                     if Opt.0verbose then call Msg 'no Usage: in header at' label
               end   /*  if abbrev()  */
         end   /*  if is a label  */
   end   /*  while l <= SrcIn.0  */
   if Opt.0Progress then call Progress
   return 0
   /* === End of Task_ShowTools ===================================================================== */

   Apply_Indent: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  applies indenting to each line by simply changing the prefix of the first word                 */
   /*  this is also where continued lines are indented                                                */
   /*  if the token is marked ReadOnly then the line is not indented                                  */
   /* =============================================================================================== */
   if Opt.0Progress then call Progress G.0meID SubID()
   Added = ''
   do t = 1 to t_type.0 -1
      /*  if token is an EOL, then change prefix of first token in this line                          */
      if abbrev(t_class.t, 'FILE_') then do
            Word1 = t +1
            /*  bugfix-- make sure word1 is the first visible token, not a hidden EOC                 */
            do while Word1 < t_type.0
               if t_val.Word1 \= '' then leave
               if abbrev(t_class.Word1, 'FILE_') then
                  iterate t
               t_prefix.Word1 = ''
               Word1 = Word1 +1
            end
            if t_class.Word1 == 'FILE_EOF' then leave t
            if \t_readonly.Word1 then do
                  /*  if previous line was a continued line, add an extra indent to this line.        */
                  spaces = t_indent.Word1 * Opt.0Tab
                  /*  round to the nearest whole space                                                */
                  spaces = format(spaces, , 0)
                  t_prefix.Word1 = copies(' ', spaces)
                  /*  add margin                                                                      */
                  t_prefix.Word1 = t_prefix.Word1||copies(' ', Opt.0Margin)
               end   /*  if not readonly  */

         end   /*  if FILE_EOL  */
   end t
   if Opt.0Progress then call Progress
   return 0
   /* === End of Apply_Indent ======================================================================= */

   Output: /*fold00*/
   /* =============================================================================================== */
   /*  This is the standard set of routines that combine the tokens into an array of lines            */
   /*  and sends them to a file. Most scripts will want to call this.                                 */
   /* =============================================================================================== */
   if Opt.0Progress then call Progress G.0meID SubID()
   /*  combine tokens into a new array of lines:                                                      */
   call Build_Lines
   /* ----------------------------------------------------------------------------------------------- */
   /*  Save to the new file and quit                                                                  */
   call ArrayToFile 'SrcOut.', G.0OutFileN
   call Msg ''
   call Msg G.0meID 'wrote new file 'G.0OutFileN
   return 0
   /* === End of Output ============================================================================= */

   Build_Lines: procedure expose (Shared_vars) (line_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  build each line by combining the prefix and value of each token in the line                    */
   /*  then add the margin to the left side                                                           */
   /* =============================================================================================== */
   if Opt.0Progress then call Progress G.0meID SubID()
   if Opt.0Margin < 0 then do
         call ShowError 'Error: bad value for Opt.0Margin "'Opt.0Margin'"'
         call Egress 'usage'
      end

   SrcOut. = ''
   ln = 1
   SrcOut.ln = ''

   do t = 1 to t_type.0
      SrcOut.ln = SrcOut.ln||t_prefix.t||t_val.t
      if t_class.t == 'FILE_EOL' then do
            SrcOut.ln = strip(SrcOut.ln, 't')
            ln = ln +1
         end
   end
   SrcOut.0 = ln -1

   /*  make sure the file ends with a CRLF                                                            */
   Last = SrcOut.0
   if SrcOut.Last \== '' then do
         ln = SrcOut.0 +1
         SrcOut.ln = ''
         SrcOut.0 = ln
      end

   if Opt.0Progress then call Progress
   return 0
   /* === End of Build_Lines ======================================================================== */

   Build_Bare: procedure expose (Shared_vars) (line_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  HELP: Bare                                                                                     */
   /*                                                                                                 */
   /*  Build_Bare makes a new line array of code only, without comments or empty lines.               */
   /*  Indenting is preserved.                                                                        */
   /*                                                                                                 */
   /* =============================================================================================== */
   if Opt.0Progress then call Progress SubID()
   SrcOut. = ''
   ln = 1
   do t = 1 to t_type.0
      if t_type.t == 'W_KEY' ,
       & MatchCmd(t_val.t, 'REM') then iterate
      if t_class.t == 'COMMENT' then iterate
      if t_type.t == 'COLON2' then iterate
      SrcOut.ln = SrcOut.ln||t_prefix.t||t_val.t
      if t_class.t == 'FILE_EOL' then do
            /*  to strip only trailing spaces:                                                        */
            SrcOut.ln = strip(SrcOut.ln, 'T')
            /*  or to strip indenting also:                                                           */
            /*   SrcOut.ln = strip(SrcOut.ln, 'B')                                                    */
            /*  ignore blank line                                                                     */
            if SrcOut.ln \= '' then ln = ln +1
            SrcOut.ln = copies(' ', Opt.0Margin)
         end
   end
   SrcOut.0 = ln -1
   if Opt.0Progress then call Progress
   return 0
   /* === End of Build_Bare ========================================================================= */

   SaveTokensF: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  send a lot of information about the tokens to a file                                           */
   /* =============================================================================================== */
   if Opt.0Progress then call Progress G.0meID SubID()

   parse arg TokenFileN
   if TokenFileN == '' then do
         call ShowError 'BUG: in SaveTokensF, no file name given.  Called from line' Esigl
         call ShowError '      'sourceline(Esigl)
         call Egress 'BUG'
      end

   if FileExist(TokenFileN) then do
         suff = 2
         do while FileExist(TokenFileN||'_'||suff)
            suff = suff +1
         end
         NewN = TokenFileN||'_'||suff
         call Msg G.0meID 'token file' TokenFileN 'changed to' NewN
         TokenFileN = NewN
      end

   /*  bugfix-- this might be called before the t_indent array has been created. If so,               */
   /*  initialize it to avoid errors below                                                            */
   if symbol('t_indent.0') == 'LIT' then
      t_indent. = 0

   /*   header lines and tab positions                                                                */
   tab_sigline = 0
   tab_Token     = tab_sigline +3
   tab_RO        = tab_Token   +10
   tab_indent    = tab_RO      +12
   tab_class     = tab_indent  +12
   tab_type      = tab_class   +12
   tab_val       = tab_type    +22
   Header_3 = ' '
   Header_3 = overlay('Token #', Header_3, tab_Token)
   Header_3 = overlay('ReadOnly:', Header_3, tab_RO)
   Header_3 = overlay('Indent:', Header_3, tab_indent)
   Header_3 = overlay('Class:', Header_3, tab_class)
   Header_3 = overlay('Type:', Header_3, tab_type)
   Header_3 = overlay('{Prefix}{Value}:', Header_3, tab_val)

   call lineout TokenFileN, Header_3
   do t = 1 to t_type.0
      /*        Header_3 = overlay('Toker line:', Header_3, 1)                                        */
      Outline = copies(' ', 140)
      Outline = overlay(format(t, 7), Outline, tab_Token)
      Outline = overlay(format(t_readonly.t, 6), Outline, tab_RO)
      Outline = overlay(format(t_indent.t, 4), Outline, tab_indent)
      Outline = overlay(t_class.t, Outline, tab_class)
      Outline = overlay(t_type.t, Outline, tab_type)
      Outline = overlay("{"t_prefix.t"}{"t_val.t"}", Outline, tab_val)

      if t // 25 == 0 then do
            call lineout TokenFileN, ''
            call lineout TokenFileN, Header_3
         end
      Outline = strip(Outline, 't')
      call lineout TokenFileN, Outline
   end t
   call lineout TokenFileN
   if Opt.0Progress then call Progress
   say G.0meID 'Token list was written to 'TokenFileN
   return 0
   /* === End of SaveTokensF ======================================================================== */

   Call_Toker: procedure expose (Shared_vars) (line_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  call the Tokeniser and do some processing of the tokens it returned                            */
   /* ----------------------------------------------------------------------------------------------- */


   /*  the Toker will return a value:                                                                 */
   /*  OKAY     --tokenising succeeded without problems                                               */
   /*  FILE     --indicates something wrong with the source file. We already checked it               */
   /*             for binary nulls, so it is probably defective code. Should not                      */
   /*             be considered a fatal error, but the result of trying to format                     */
   /*             defective code should be checked carefully.                                         */
   /*  USAGE    --something was wrong with the way we called the Toker, so we quit.                   */
   /*  BUG      --indicates an error in the Toker itself, so we should probably quit.                 */
   /*                                                                                                 */

   G.0TokerExit = Toker()
   if debug.0symbol then
      call debugSymbol
   select
      when G.0TokerExit = '' then nop
      when wordpos('BUG', G.0TokerExit) > 0 then do
            call ShowError 'Error: Tokeniser returned "BUG"'
            call ShowError 'aborting because this is probably a bug in the Tokeniser', ''
            if debug.0token then call SaveTokensT G.0InfileN||'_Tokens_Begin'
            call Egress 'BUG_Toker'
         end
      when wordpos('FILE', G.0TokerExit) > 0 then do
            /*  some kind of defective file. If the Toker continued, we will also,                    */
            /*  but there is no telling what will happen next                                         */
            if translate(G.0Command) \== 'CHECKONLY'then do
                  call ShowError 'Warning: Tokeniser returned error "FILE"'
                  call ShowError 'Continuing, but there may be more errors', ''
               end
         end
      otherwise
         /*  if Toker returned something else, there is a bug in the Toker                            */
         call ShowError 'Toker returned "'G.0TokerExit'", aborting'
         if debug.0token then call SaveTokensT G.0InfileN||'_Tokens_Begin'
         call Egress 'BUG'
   end
   return 0
   /* === End of Call_Toker ========================================================================= */

   debugSymbol: procedure expose (Shared_vars) (line_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  unlike in RexxTool.cmd, a type W_SYMBOL here means a token that the Toker did not              */
   /*  identify as some other type. Use this to look for words that it should have identified.        */
   /* =============================================================================================== */

   do t = 1 to t_type.0
      if t_type.t == 'W_SYMBOL' then do
            call ShowError '[debug] type W_SYMBOL found for token' t', "'t_val.t'" at line' t_line.t
            call ShowError Qline(t_line.t), ''
         end
   end t
   return
   /* === End of debugSymbol ======================================================================== */

   L_AsString: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       r = L_AsString(<token> [, <line>])                                                        */
   /*       r = L_AsString(T)                                                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns an entire line as a string                                                             */
   /*  if no <line> is given, returns the line that token T is in                                     */
   /*  if <line> is negative, returns the line that is <line> lines above token T                     */
   /*  if <line> is positive, returns the line that is <line> lines below token T                     */
   /* =============================================================================================== */
   Esigl = sigl
   t = arg(1)
   count = arg(2)
   if debug.0arg then call T_Valid t
   if count = '' then count = 0

   p = 0
   c = 0
   select
      when count < 0 then do
            t = L_PrevEOL(t)
            do p = t to 1 by -1
               if abbrev(t_class.p, 'FILE_') then c = c +1
               if c == abs(count) then leave
            end
         end

      when count > 0 then
         do p = t to t_type.0 by 1
            if abbrev(t_class.p, 'FILE_') then c = c +1
            if c == count +1 then leave
         end

      otherwise
         p = t
   end

   if p > t_type.0 then return ''
   if p < 1 then return ''

   First = L_PrevEOL(p) +1
   Last = L_LineEOL(p) -1
   /*  for the first token, use the indent value instead of its old prefix                            */
   outstring = copies(' ', t_indent.First * Opt.0Tab)
   outstring = outstring||t_val.First
   do i = First +1 to Last
      outstring = outstring||t_prefix.i||t_val.i
   end
   return outstring
   /* === End of L_AsString ========================================================================= */

   L_IsContinued: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       r = L_IsContinued(T [,<line>])                                                            */
   /*  Example:                                                                                       */
   /*       IF L_IsContinued(t, -1) THEN     /@ if previous line is continued @/                      */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns True if the line is a continued line ending with an escape character                   */
   /*  <line> is relative to the line that T is in                                                    */
   /* =============================================================================================== */
   t = arg(1)
   count = arg(2)
   if count = '' then count = 0
   p = L_LineEOL(t, count)
   if t_type.p == 'NULL' then return 1
   return 0
   /* === End of L_IsContinued ====================================================================== */

   L_Last4OS2Token: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       r = L_Last4OS2Token(T [,<line>])                                                          */
   /*       if T_Val(L_Last4OS2Token(T)) == ',' ...                                                   */
   /* ----------------------------------------------------------------------------------------------- */
   /*  given token T, returns the number of the last 4OS2 token in the line.                          */
   /*  That is, Comment tokens are ignored                                                            */
   /*  <line> is relative to the line that T is in                                                    */
   /*  if line has no 4OS2 words it returns '', which may cause an error later                        */
   /* =============================================================================================== */
   Esigl = sigl
   t = arg(1)
   count = arg(2)
   if count = '' then count = 0
   if debug.0arg then call T_Valid t

   p = L_LineEOL(t, count)
   do c = L_LineEOL(p) to L_PrevEOL(p) by -1
      if t_class.c == '4OS2' then return c
   end c
   return ''
   /* === End of L_Last4OS2Token ==================================================================== */

   L_LineEOL: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       r = L_LineEOL(T [, <line>])                                                               */
   /*       r = L_LineEOL(T)                                                                          */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns the token number that marks the end of a line (EOL)                                    */
   /*  if no <line> is given, the line that token T is in                                             */
   /*  if <line> is negative, the line is <line> lines above where token T is                         */
   /*  if <line> is positive, the line is <line> lines below where token T is                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  in all of the L_xxx() functions a line is defined as a string + EOL, where                     */
   /*  the string may be empty. A line is identified by the token number of its EOL.                  */
   /* =============================================================================================== */
   Esigl = sigl
   t = arg(1)
   count = arg(2)
   if count = '' then count = 0
   if debug.0arg then call T_Valid t

   p = 0
   c = 0
   select
      when count < 0 then do
            t = L_PrevEOL(t)
            do p = t to 1 by -1
               if abbrev(t_class.p, 'FILE_') then c = c +1
               if c == abs(count) then leave
            end
         end
      when count >= 0 then
         do p = t to t_type.0 by 1
            if abbrev(t_class.p, 'FILE_') then c = c +1
            if c == count +1 then leave
         end
      otherwise nop
   end
   return p
   /* === End of L_LineEOL ========================================================================== */

   L_PrevEOL: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        r = L_PrevEOL(T)                                                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  given token T, returns the EOL at the start of the line that T is in.                          */
   /*  Or more accurately, the EOL of the previous line.                                              */
   /*  If T is an EOL, returns the previous EOL                                                       */
   /*                                                                                                 */
   /*  this is equivalent to L_LineEOL(t, -1)                                                         */
   /* =============================================================================================== */
   Esigl = sigl
   arg t
   if debug.0arg then call T_Valid t

   if abbrev(t_class.t, 'FILE_') then t = t -1
   do c = t to 1 by -1
      if abbrev(t_class.c, 'FILE_') then return c
   end c
   return 1
   /* === End of L_PrevEOL ========================================================================== */

   R_Expression: procedure expose (Shared_vars) (line_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = R_Expression(t)                                                                 */
   /*                                                                                                 */
   /*  returns a string representing the expression that token t is part of                           */
   /*  Examples:                                                                                      */
   /*     for a line                                                                                  */
   /*        %aa = %@filename[%aa]                                                                    */
   /*     where the first token is 1                                                                  */
   /*        R_Expression(1) returns "%aa"                                                            */
   /*        R_Expression(2) returns "%aa"                                                            */
   /*        R_Expression(3) returns "="                                                              */
   /*        R_Expression(4) returns "%@filename[%aa]"                                                */
   /*                                                                                                 */
   /*  This is very experimental. Will take some tweaking to see what it is useful for                */
   /* =============================================================================================== */
   Esigl = sigl
   /*  if quoted, move start back to begin quote                                                      */
   arg t
   start = t
   if t_class.start == 'QUOTED' then do
         do while T_Class(start -1) == 'QUOTED'
            start = start -1
         end
      end

   /*  fix: what else should be considered a word boundary?                                           */
   do while t_prefix.start == '' ,
          & T_Type(start -1) \= 'W_(' ,
          & T_Val(start -1) \= '=' ,
          & T_Type(start -1) \= 'EOC'
      start = start -1
   end

   wLast = R_LastWord(t)
   stop = wLast
   s = start
   do while s <= wLast
      select
         when t_val.s == '"' then do
               do until T_Class(s +1) \= 'QUOTED'
                  s = s +1
               end
            end
         when t_val.s == '[' then
            s = R_MatchingBracket(s)
         when t_type.s == 'W_)' then do
               stop = s -1
               leave
            end
         when t_val.s == '=' then do
               stop = s -1
               leave
            end
         when t_val.s == '(' then do
               stop = s
               leave
            end
         otherwise nop
      end   /*  select  */
      if T_Prefix(s +1) \== '' then do
            /*  found the next space-separated "word"                                                 */
            stop = s
            leave
         end
      s = s +1
   end   /*  while s < WLast  */

   string = ''
   do s = start to stop
      string = string||t_prefix.s||t_val.s
   end s
   string = strip(string, 'l')
   return string
   /* === End of R_Expression ======================================================================= */

   R_MatchingBracket: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        t = R_MatchingBracket(t)                                                                 */
   /*                                                                                                 */
   /*  if token t is '(' returns the matching ')'                                                     */
   /*  if token t is ')' returns the matching '('                                                     */
   /*  if token t is '[' returns the matching ']'                                                     */
   /*  if token t is ']' returns the matching '['                                                     */
   /* =============================================================================================== */
   /*  '[' and ']' cannot be in separate statements, so if the line ends without                      */
   /*  finding a match,  return the end-of-line token.                                                */
   /*  A missing ')' would be harder to detect.                                                       */

   Esigl = sigl
   arg t
   inside = 1
   select
      when t_val.t == '(' then do
            do s = t +1 to t_type.0
               select
                  when t_val.s == '(' then
                     inside = inside +1
                  when t_val.s == ')' then do
                        inside = inside -1
                        if inside == 0 then
                           return s
                     end
                  otherwise nop
               end   /*  select  */
            end s
            /*  no match found                                                                        */
            return t_type.0
         end
      when t_val.t == '[' then do
            do s = t +1 to t_type.0
               select
                  when t_type.s == 'EOC' ,
                     & abbrev(t_class.s, 'FILE_') then do
                        /*  end of line/statement without finding the ']'                             */
                        call ShowError 'Warning: R_MatchingBracket: no matching "]" found on line' t_line.t
                        call ShowError Qline(t_line.t)
                        return s
                     end
                  when t_val.s == '[' then
                     inside = inside +1
                  when t_val.s == ']' then do
                        inside = inside -1
                        if inside == 0 then
                           return s
                     end
                  otherwise nop
               end   /*  select  */
            end s
            /*  no match found                                                                        */
            return t_type.0
         end
      when t_val.t == ')' then do
            do s = t -1 to 1 by -1
               select
                  when t_val.s == ')' then
                     inside = inside +1
                  when t_val.s == '(' then do
                        inside = inside -1
                        if inside == 0 then
                           return s
                     end
                  otherwise nop
               end   /*  select  */
            end s
            /*  no match found                                                                        */
            return 1
         end
      when t_val.t == ']' then do
            do s = t -1 to 1 by -1
               select
                  when t_type.s == 'EOC' ,
                     & abbrev(t_class.s, 'FILE_') then do
                        /*  beginning of line/statement without finding the ']'                       */
                        call ShowError 'Warning: R_MatchingBracket: no matching "[" found on line' t_line.t
                        call ShowError Qline(t_line.t)
                        return s
                     end
                  when t_val.s == ']' then
                     inside = inside +1
                  when t_val.s == '[' then do
                        inside = inside -1
                        if inside == 0 then
                           return s
                     end
                  otherwise nop
               end   /*  select  */
            end s
            /*  no match found                                                                        */
            return 1
         end
      otherwise
         call ShowError 'BUG: invalid argument to Toker_Match at line' Esigl
         return t
   end   /*  select  */
   /* === End of R_MatchingBracket ================================================================== */

   R_NextEOC: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       r = R_NextEOC(T)                                                                          */
   /*                                                                                                 */
   /*  returns the next End of Clause after token T                                                   */
   /* =============================================================================================== */
   Esigl = sigl
   arg t
   if debug.0arg then call T_Valid t
   do c = t +1 to t_type.0
      if t_type.c == 'EOC' then return c
   end c
   return t
   /* === End of R_NextEOC ========================================================================== */

   R_PrevEOC: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       r = R_PrevEOC(T)                                                                          */
   /*                                                                                                 */
   /*  returns the End of Clause before the statement that token T is in                              */
   /* =============================================================================================== */
   Esigl = sigl
   arg t
   if debug.0arg then call T_Valid t
   do c = t -1 to 1 by -1
      if t_type.c == 'EOC' then return c
   end c
   return t
   /* === End of R_PrevEOC ========================================================================== */

   R_NextKeyword: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        r = NextKeyword(T [, <string>])                                                          */
   /*                                                                                                 */
   /*  returns the token of the next keyword that matches any word in <string>                        */
   /*  a string of '' matches any keyword                                                             */
   /*  if no match was found, returns t_type.0                                                        */
   /* =============================================================================================== */
   t = arg(1)
   ToMatch = translate(arg(2))
   do i = t +1 to t_type.0
      if t_type.i == 'W_KEY' then do
            if ToMatch == '' then return i
            if MatchCmd(t_val.i, ToMatch) then return 1
         end
   end i
   return t_type.0
   /* === End of R_NextKeyword ====================================================================== */

   R_PrevKeyword: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        r = PrevKeyword(T [, <string>])                                                          */
   /*                                                                                                 */
   /*  returns the token of the previous keyword that matches any word in <string>                    */
   /*  a string of '' matches any keyword                                                             */
   /*  if no match was found, returns 1                                                               */
   /* =============================================================================================== */
   t = arg(1)
   ToMatch = translate(arg(2))
   do i = t -1 to 1 by -1
      if t_type.i == 'W_KEY' then do
            if ToMatch == '' then return i
            if MatchCmd(t_val.i, ToMatch) then return 1
         end
   end i
   return 1
   /* === End of R_PrevKeyword ====================================================================== */

   R_NextLabel: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       r = R_NextLabel(T)                                                                        */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns the token number of the next label, If none, returns the token number of EOF           */
   /* =============================================================================================== */
   arg t
   do s = t +1 to t_type.0 -1
      if t_type.s == 'LABEL' then return s
   end
   return t_type.0
   /* === End of R_NextLabel ======================================================================== */

   R_NextWord: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       r = R_NextWord(T)                                                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  finds the next =word=                                                                          */
   /*  it ignores End-Of-Statement, End-Of-Line, Continuation, VAR, and comments                      */
   /* =============================================================================================== */
   Esigl = sigl
   arg t
   if debug.0arg then call T_Valid t
   do c = t +1 to t_type.0
      if abbrev(t_type.c, 'W_') then return c
   end c
   /*  fix: what should it return if at end of file where there is no next Word?                      */
   return t
   /* === End of R_NextWord ========================================================================= */

   S_Quoted: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = S_Quoted(t)                                                                     */
   /*                                                                                                 */
   /*  returns the entire quoted string that token t is part of                                       */
   /*  the string may be spread over more than one continued line                                     */
   /* =============================================================================================== */
   Esigl = sigl
   arg t

   str = ''
   do s = R_StartQuote(t) to R_StopQuote(t)
      if t_type.s \= 'CONTINUE' ,
       & t_type.s \= 'NULL' then
         str = str||t_val.s
   end s
   return str
   /* === End of S_Quoted =========================================================================== */

   R_StopQuote: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     q = R_StopQuote(t)                                                                          */
   /*     where t is any token of a quoted string                                                     */
   /*                                                                                                 */
   /*  returns the token that ends a string in either double quotes or back quotes                    */
   /* =============================================================================================== */
   Esigl = sigl
   arg t

   select
      /*  the order is important here-- in the case of a backquoted string inside                     */
      /*  double quotes, we want to give priority to the backquotes                                   */
      when t_type.t == 'W_LITERAL' then do
            s = t
            do while t_type.s == 'W_LITERAL' ,
                   | t_type.s == 'NULL' ,
                   | t_type.s == 'CONTINUE'
               s = s +1
            end
            return s -1
         end
      when t_class.t == 'QUOTED' then do
            s = t
            do while t_class.s == 'QUOTED' ,
                   | t_type.s == 'NULL' ,
                   | t_type.s == 'CONTINUE'
               s = s +1
            end
            return s -1
         end
      otherwise
         call ShowError 'BUG: invalid argument to R_StopQuote at line' Esigl
         call Mark 't t_class.t t_type.t t_val.t', ''
         call Egress 'bug'
   end   /*  select  */
   /* === End of R_StopQuote ======================================================================== */

   R_StartQuote: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     q = R_StartQuote(t)                                                                         */
   /*     where t is any token of a quoted string                                                     */
   /*                                                                                                 */
   /*  returns the token that begins a string in either double quotes or back quotes                  */
   /* =============================================================================================== */
   Esigl = sigl
   arg t
   select
      /*  the order is important here-- in the case of a backquoted string inside                     */
      /*  double quotes, we want to give priority to the backquotes                                   */
      when t_type.t == 'W_LITERAL' then do
            s = t
            do while t_type.s == 'W_LITERAL' ,
                   | t_type.s == 'NULL' ,
                   | t_type.s == 'CONTINUE'
               s = s -1
            end
            return s +1
         end
      when t_class.t == 'QUOTED' then do
            s = t
            do while t_class.s == 'QUOTED' ,
                   | t_type.s == 'NULL' ,
                   | t_type.s == 'CONTINUE'
               s = s -1
            end
            return s +1
         end
      otherwise
         call ShowError 'BUG: invalid argument to R_StartQuote at line' Esigl
         call Mark 't t_class.t t_type.t t_val.t', ''
         call Egress 'bug'
   end   /*  select  */
   /* === End of R_StartQuote ======================================================================= */

   R_FirstWord: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       r = R_FirstWord(T)                                                                        */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns the first word of the statement that token T is in                                     */
   /* =============================================================================================== */
   Esigl = sigl
   arg t
   if debug.0arg then call T_Valid t
   if \abbrev(t_type.t, 'W_') then do
         call ShowError 'BUG: invalid call to R_FirstWord() from line' Esigl
         call ShowError 'token' t 'is not a 4OS2 word'
         call Mark 't t_type.t t_val.t'
         call Egress 'BUG_Function'
      end

   do c = t -1 to 1 by -1
      if t_type.c == 'EOC' then leave
   end c
   do w = c to t
      if abbrev(t_type.w, 'W_') then return w
   end w
   return t
   /* === End of R_FirstWord ======================================================================== */

   R_LastWord: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       r = R_LastWord(T)                                                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns the last word of the statement that token T is in                                      */
   /* =============================================================================================== */
   Esigl = sigl
   arg t
   if debug.0arg then call T_Valid t
   if \abbrev(t_type.t, 'W_') then do
         call ShowError 'BUG: invalid call to R_LastWord() from line' Esigl
         call ShowError 'token' t 'is not a 4OS2 word'
         call Mark 't t_type.t t_val.t'
         call Egress 'BUG_Function'
      end

   do c = t +1 to t_type.0
      if t_type.c == 'EOC' then leave
   end c
   do w = c to t by -1
      if abbrev(t_type.w, 'W_') then return w
   end w
   return t
   /* === End of R_LastWord ========================================================================= */

   R_OtherEnd: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       r = R_OtherEnd(T)                                                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  if token t is DO or SWITCH, this finds the matching ENDDO or ENDSWITCH                         */
   /*  if token t is ENDDO or ENDSWITCH, this finds the matching DO or SWITCH                         */
   /* =============================================================================================== */
   /*  this requires valid code. To check for unbalanced DO/END, use MatchDoEnd()                     */
   /* ----------------------------------------------------------------------------------------------- */
   Esigl = sigl
   arg t
   if debug.0arg then call T_Valid t
   Tval = translate(t_val.t)
   select
      when t_type.t == 'W_KEY' ,
         & wordpos(Tval, 'DO SWITCH') > 0 then do
            t_start = t +1 ; t_end = t_type.0 ; step = 1
         end
      when t_type.t == 'W_KEY' ,
         & wordpos(Tval, 'ENDDO ENDSWITCH') > 0 then do
            t_start = t -1 ; t_end = 1 ; step = -1
         end
      otherwise
         call ShowError 'BUG: invalid call to R_OtherEnd from line' Esigl
         call Mark 't t_type.t t_val.t'
         call Egress 'BUG_Function'
   end
   level = 1
   do e = t_start to t_end by step
      Tval = translate(t_val.e)
      select
         when t_type.e \== 'W_KEY' then nop
         when Tval = 'DO' then level = level + step
         when Tval = 'SWITCH' then level = level + step
         when Tval = 'ENDDO' then level = level - step
         when Tval = 'ENDSWITCH' then level = level - step
         otherwise nop
      end
      if level = 0 then return e
   end e
   /*  fix: what to return if the other end is not found?                                             */
   return t
   /* === End of R_OtherEnd ========================================================================= */

   Rexx_DividerType: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       IF Rexx_DividerType(string) > 3 then ...                                                  */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns a number to show if a string looks like a divider comment--                            */
   /*     5 like /@ ############# @/                                                                  */
   /*     4 like /@ ************* @/                                                                  */
   /*     3 like /@ ============= @/                                                                  */
   /*     2 like /@ ------------- @/                                                                  */
   /*     1 like /@ _____________ @/                                                                  */
   /*     0 string is not a divider                                                                   */
   /* =============================================================================================== */
   /*  akm: 'divider' defined as a one-line comment in the style I use in headers                     */
   /*        if you add letters to Div_chars, note that verify() is case-sensitive                    */
   /* ----------------------------------------------------------------------------------------------- */

   parse arg str
   str = strip(str)
   if left(str, 2) \== G.0CommentBegin then return 0
   if right(str, 2) \== G.0CommentEnd then return 0
   str = MidStr(str, 3, length(str) -4)
   len = length(str)
   str = StripStr(str)
   if length(str) +2 < len then return 0                       /* more than one blank at each end: No */
   Div_chars = '*=-_#'
   if length(str) < 20 then return 0                                                 /* too short, No */
   if verify(str, Div_chars) > 0 then return 0                  /* has other chars than Div_chars, No */
   c1 = left(str, 1)                                          /* compare first char to rest of string */
   if verify(str, c1) > 0 then return 0                               /* if not all one character, No */
   select
      when c1 == '#' then return 5
      when c1 == '*' then return 4
      when c1 == '=' then return 3
      when c1 == '-' then return 2
      when c1 == '_' then return 1
      otherwise return 0
   end
   return 0
   /* === End of Rexx_DividerType =================================================================== */

   Rexx_IsComment: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       IF Rexx_IsComment(string) then ...                                                        */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns True if string looks like a Rexx comment                                               */
   /* =============================================================================================== */
   parse arg str
   str = strip(str)
   if left(str, 2) \= G.0CommentBegin then return 0
   if right(str, 2) \= G.0CommentEnd then return 0
   return  1
   /* === End of Rexx_IsComment ===================================================================== */

   Rexx_IsLabel: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       IF Rexx_IsLabel('foo:') then ...                                                          */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns 1 if a string looks like a Rexx label                                                  */
   /* =============================================================================================== */
   /*  contrary to some documentation, the ':' does not need to follow immediately after the label,   */
   /*  and the label can start with a digit or '.'                                                    */
   /* ----------------------------------------------------------------------------------------------- */
   parse arg str
   if pos(':', str) == 0 then return 0
   parse var str label ':' .
   label = strip(label)
   if symbol(label) \== 'BAD' then return 1
   return 0
   /* === End of Rexx_IsLabel ======================================================================= */

   Rexx_UnComment: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       string = Rexx_UnComment <string>                                                          */
   /* ----------------------------------------------------------------------------------------------- */
   /*  strips the beginning /@ and ending @/ from a comment.                                          */
   /*  it does not strip spaces from the remaining text                                               */
   /* =============================================================================================== */
   parse arg str
   /*  only need one call to StripStr() because Begin and End comment use the same characters         */
   if left(strip(str), 2) == G.0CommentBegin then
      str = StripStr(strip(str), 'B', G.0CommentBegin)
   return str
   /* === End of Rexx_UnComment ===================================================================== */

   T_AsString: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = T_AsString(T1, T2)                                                              */
   /*                                                                                                 */
   /*  returns a string composed of all prefixes and values of tokens <T1> through <T2>               */
   /*  the calling routine might want to trim the leading blanks                                      */
   /* =============================================================================================== */
   t_from = arg(1)
   t_to = arg(2)
   if t_to == '' then t_to = t_from
   outstring = ''
   do i = t_from to t_to
      outstring = outstring||t_prefix.i||t_val.i
   end
   return outstring
   /* === End of T_AsString ========================================================================= */

   T_Class: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       string = T_Class(<token>, [<new prefix>])                                                 */
   /*       string = T_Class(t)                                                                       */
   /*       CALL T_Class t, 'COMMENT'                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns the class of token t                                                                   */
   /*  if there is a second argument, the class will be changed to it first                           */
   /* =============================================================================================== */
   Esigl = sigl
   t = arg(1)
   if debug.0arg then call T_Valid t
   if arg(2, 'e') then t_class.t = arg(2)
   return t_class.t
   /* === End of T_Class ============================================================================ */

   T_Col: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       string = T_Col(<token>, [<new prefix>])                                                   */
   /*       string = T_Col(t)                                                                         */
   /*       CALL T_Col t, '10'                                                                        */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns the column of token t                                                                  */
   /*  if there is a second argument, the t_col.t will be changed to it first                         */
   /* =============================================================================================== */
   Esigl = sigl
   t = arg(1)
   if debug.0arg then call T_Valid t
   if arg(2, 'e') then t_col.t = arg(2)
   return t_col.t
   /* === End of T_Col ============================================================================== */

   T_FindToken: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage :                                                                                        */
   /*          t = T_FindToken(<start> , <stop> , [<value>] , [<type>] , [<class>] )                  */
   /*                                                                                                 */
   /*  Finds the next/previous token that has the specified value, type, and class                    */
   /*  where <start> and <stop> are token numbers between 0 and t_type.0                              */
   /*        <value> is the t_val to match                                                            */
   /*        <type>  is the t_type to match                                                           */
   /*        <class> is the t_class to match                                                          */
   /*                                                                                                 */
   /*  a <value> may be '', which will only match a token with a null value                           */
   /*  a <type> or <class> is an abbreviation, so 'FILE' will match                                   */
   /*    'FILE_BOF' or 'FILE_EOL'                                                                     */
   /*  a missing <type> or <class> will match anything                                                */
   /*                                                                                                 */
   /*  all matching is case-insensitive                                                               */
   /*                                                                                                 */
   /*  if <start> is greater than <stop> the search will work backwards                               */
   /*  if <stop> is missing, the search will work forward from <start> to t_type.0                    */
   /*                                                                                                 */
   /*  Examples:                                                                                      */
   /*        to find the next 'ENDDO' after token T                                                   */
   /*             NextEndDo = T_FindToken(T +1, , 'ENDDO' , 'W_KEY' , '4OS2')                         */
   /*        to find the beginning of a comment on this line                                          */
   /*              Rem = T_FindToken(t, L_LineEOL(t), 'REM', 'W_KEY')                                 */
   /*        to find the last ")" before the end of this line                                         */
   /*              cp = T_FindToken(L_LineEOL(t). L_PrevEOL(t), ')', 'W_)', '4OS2')                   */
   /*           or more simply                                                                        */
   /*              cp = T_FindToken(L_LineEOL(t). L_PrevEOL(t), , 'W_)')                              */
   /*                                                                                                 */
   /*  Returns:                                                                                       */
   /*      the token number if found                                                                  */
   /*      -1 if not found                                                                            */
   /* =============================================================================================== */

   parse upper arg t_start , t_stop , m_val , m_type , m_class , .
   t_start = arg(1)
   t_stop = arg(2)
   m_val = arg(3)
   m_type = arg(4)
   m_class = arg(5)

   if t_stop = '' then
      t_stop = t_type.0 -1
   if debug.0arg then do
         call T_Valid t_start
         call T_Valid t_stop
         if arg(4, 'e') ,
          & arg(4)  == '' then
            call ShowError 'Warning: invalid type argument to T_FindToken()'
         if arg(5, 'e') ,
          & arg(5)  == '' then
            call ShowError 'Warning: invalid class argument to T_FindToken()'
      end

   if t_stop > t_start then do
         step = 1
      end
   else do
         step = -1
      end

   do s = t_start to t_stop by step
      /*  t_val comparisons are not case-sensitive, and must match whole string                       */
      if arg(3, 'e') then
         if translate(t_val.s) \= m_val then do
               iterate
            end
      if arg(4, 'e') then
         if \abbrev(t_type.s, translate(m_type)) then do
               iterate
            end
      if arg(5, 'e') then
         if \abbrev(t_class.s, translate(m_class)) then do
               iterate
            end
      /*  if it passed all the tests, it is a match                                                   */
      return s
   end s
   /*   ended without finding anything                                                                */
   return -1
   /* === End of T_FindToken ======================================================================== */

   T_Indent: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       r = T_Indent(<token>, [new value])                                                        */
   /*       r = T_Indent(t)                                                                           */
   /*       CALL T_Indent t, 1                                                                        */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns the Indent value status of token <n>                                                   */
   /*  if there is a second argument, the Indent will be changed to it first                          */
   /* =============================================================================================== */
   Esigl = sigl
   t = arg(1)
   if debug.0arg then call T_Valid t
   if arg(2, 'e') then t_indent.t = arg(2)
   return t_indent.t
   /* === End of T_Indent =========================================================================== */

   T_Infunction: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        FunctionName = T_Infunction(T)                                                           */
   /*  if token T is inside a function, returns the name of the function                              */
   /*  otherwise returns ''                                                                           */
   /* =============================================================================================== */
   parse arg t

   level = -1
   do i = t to 1 by -1
      select
         when t_type.i == 'W_[' then
            level = level +1
         when t_type.i == 'W_]' then
            level = level -1
         when left(t_val.i, 1) == '@' then do
               if level == 0 then
                  return t_val.i
            end
         when t_type.i == 'EOC' ,
            & abbrev(t_class.i, 'FILE_') then
            leave i
         otherwise
            nop
      end   /*  select  */
   end i
   return ''
   /* === End of T_Infunction ======================================================================= */

   T_Line: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       string = T_Line(<token>, [<new prefix>])                                                  */
   /*       string = T_Line(t)                                                                        */
   /*       CALL T_Line t, '10'                                                                       */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns the line of token t                                                                    */
   /*  if there is a second argument, the t_line.t will be changed to it first                        */
   /* =============================================================================================== */
   Esigl = sigl
   t = arg(1)
   if debug.0arg then call T_Valid t
   if arg(2, 'e') then t_line.t = arg(2)
   return t_line.t
   /* === End of T_Line ============================================================================= */

   T_Prefix: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       string = T_Prefix(<token>, [<new prefix>])                                                */
   /*       string = T_Prefix(t)                                                                      */
   /*        CALL T_Prefix t, ' '                                                                     */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns the prefix of token t                                                                  */
   /*  if there is a second argument, the prefix will be changed to it first                          */
   /* =============================================================================================== */
   Esigl = sigl
   t = arg(1)
   if debug.0arg then call T_Valid t
   if arg(2, 'e') then t_prefix.t = arg(2)
   return t_prefix.t
   /* === End of T_Prefix =========================================================================== */

   T_ReadOnly: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       r = T_ReadOnly(<token>, [new value])                                                      */
   /*       r = T_ReadOnly(t)                                                                         */
   /*       CALL T_ReadOnly t, 1                                                                      */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns the ReadOnly status of token <n>                                                       */
   /*  if there is a second argument, the ReadOnly flag will be changed to it first                   */
   /* =============================================================================================== */
   Esigl = sigl
   t = arg(1)
   if debug.0arg then call T_Valid t
   if arg(2, 'e') then t_readonly.t = arg(2)
   return t_readonly.t
   /* === End of T_ReadOnly ========================================================================= */

   T_Type: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       string = T_Type(<token>, [<new prefix>])                                                  */
   /*       r = T_Type(t)                                                                             */
   /*       CALL T_Type t, 'COMMENT_VAL'                                                              */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns the type of token <n>                                                                  */
   /*  if there is a second argument, the type will be changed to it first                            */
   /* =============================================================================================== */
   Esigl = sigl
   t = arg(1)
   if debug.0arg then call T_Valid t
   if arg(2, 'e') then t_type.t = arg(2)
   return t_type.t
   /* === End of T_Type ============================================================================= */

   T_Val: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       string = T_Val(<token>, [<new prefix>])                                                   */
   /*       string = T_Val(t)                                                                         */
   /*       CALL T_Val t, 'this is a comment'                                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns the value of token t                                                                   */
   /*  if there is a second argument, the value will be changed to it first                           */
   /* =============================================================================================== */
   Esigl = sigl
   t = arg(1)
   if debug.0arg then call T_Valid t
   if arg(2, 'e') then t_val.t = arg(2)
   return t_val.t
   /* === End of T_Val ============================================================================== */

   T_Valid: /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        call T_Valid T                                                                           */
   /*        if T_Valid(T) then ...                                                                   */
   /* ----------------------------------------------------------------------------------------------- */
   /*  verifies that T is a whole number between 1 and t_type.0                                       */
   /*  If not, it shows a diagnostic message and exits 'BUG'                                          */
   /*  Useful when writing a new task, to validate out-of-range arguments to the T_xxx functions      */
   /* =============================================================================================== */
   t = arg(1)
   if datatype(arg(1), 'W') ,
    & arg(1) >= 1 ,
    & arg(1) <= t_type.0 then return 1
   TVsigl = sigl
   call ShowError '', ''
   call ShowError 'BUG: invalid token number "'t'" in "'LastLabel(TVsigl)'"'
   call ShowError '      'sourceline(TVsigl)
   call ShowError 'Esigl == line' Esigl '(in 'LastLabel(Esigl)')', ''
   call ShowError '      'sourceline(Esigl)
   call Egress 'BUG'
   return 0
   /* === End of T_Valid ============================================================================ */

   MatchCmd: procedure expose (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     if MatchCmd(<keyword>, <list>) then                                                         */
   /*  Example:                                                                                       */
   /*     if MatchCmd('set', 'SET ALIAS')     --returns True                                          */
   /*                                                                                                 */
   /*  returns True if arg(1) is matches a word in the list in arg(2)                                 */
   /*  the match is case-insensitive                                                                  */
   /* =============================================================================================== */
   parse arg key , list
   /*     if wordpos(StripStr(translate(key), 'l', '*@'), translate(list)) > 0 then                   */
   if wordpos(translate(key), translate(list)) > 0 then
      return 1
   else
      return 0
   /* === End of MatchCmd =========================================================================== */

   ArrayInsert: procedure expose  (Shared_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        CALL ArrayInsert <array name> , <where> , <how many>                                     */
   /*        CALL ArrayInsert 't_type.' , 100 , 3                                                     */
   /* ----------------------------------------------------------------------------------------------- */
   /*  this is intended to be called by the T_InsertToken function, but I suppose it might            */
   /*  have other uses.                                                                               */
   /* ----------------------------------------------------------------------------------------------- */
   /*  inserts 1 or more new elements into an array                                                   */
   /*  the array name must be included in the token_vars list                                         */
   /*  this is faster than making multiple calls to SysStemInsert()                                   */
   /* =============================================================================================== */

   ArrayN = arg(1)
   where = arg(2)
   HowMany = arg(3)
   if HowMany == '' then HowMany = 1
   if ArrayN == '' ,
    | \datatype(where, 'w') ,
    | \datatype(HowMany, 'w') then do
         call ShowError 'In ArrayInsert, invalid argument(s)'
         call Egress 'BUG_Function'
      end

   !Array = value('ArrayN')

   /*  shift lower part of array down to make room:                                                   */
   call value !Array||0, value(!Array||0) + HowMany
   do n = value(!Array||0) to where by -1
      o = n - HowMany
      call value !Array||n, value(!Array||o)
   end

   /*  make new empty tokens by overwriting with default value of this array                          */
   do n = where to where + HowMany -1
      call value !Array||n, value(!Array)
   end

   return 0
   /* === End of ArrayInsert ======================================================================== */

   ArrayToFile: procedure expose (Shared_vars) (line_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       CALL ArrayToFile 'Stemvar.' , 'FileName'                                                  */
   /* ----------------------------------------------------------------------------------------------- */
   /*  writes a stem array to a file                                                                  */
   /*  the stem =name= is passed, but the stem =variable= itself must be exposed                      */
   /*  if filename is not given, default is stdout                                                    */
   /* =============================================================================================== */
   parse arg stemN , fileN
   if fileN == '' then fileN = 'stdout'
   if stemN == '' then do
         call ShowError 'In ArrayToFile, invalid arguments'
         call Egress 'BUG_Function'
      end
   !listN = value('stemN')
   if !listN == '' then do
         call ShowError 'In ArrayToFile, invalid arguments'
         call Egress 'BUG_Function'
      end

   /*  cosmetic improvement, keep the progress and output from stepping on each other                 */
   if fileN == 'stdout' then
      call lineout 'stderr', ''
   else
      if Opt.0Progress then call Progress G.0meID SubID()

   if right(!listN, 1) \== '.' then !listN = !listN||'.'
   ListHigh = value(!listN||0)
   do i = 1 to ListHigh
      ThisLine = value(!listN||i)
      call lineout fileN, ThisLine
   end
   call lineout fileN
   if Opt.0Progress then call Progress
   return 0
   /* === End of ArrayToFile ======================================================================== */

   FileToArray: procedure expose (Shared_vars) (line_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       CALL FileToArray 'Stemvar.' , 'FileName'                                                  */
   /* ----------------------------------------------------------------------------------------------- */
   /*  reads input file into an array, return 0 if successful                                         */
   /*                                                                                                 */
   /*  The stem name is passed, but the stem variable itself must be exposed                          */
   /*  The stem variable should already be initialized                                                */
   /* ----------------------------------------------------------------------------------------------- */
   if Opt.0Progress then call Progress G.0meID SubID()
   parse arg stemN , fileN
   !listN = value('stemN')
   if !listN = '' ,
    | fileN = '' ,
    | right(!listN, 1) \== '.' then do
         call ShowError 'In FileToArray, invalid argument(s)'
         call Egress 'BUG_Function'
      end

   if Opt.0Progress then call Progress G.0meID SubID()
   i = value(!listN||0)
   if i == '' then i = 0
   /*  interpreting a line is slightly faster than a loop using 'value' for each assignment           */
   FileArray_script = ,
     'do while chars(fileN) > 0 ; i = i +1 ; '!listN||'i = linein(fileN) ; end ; '!listN||'0 = i'
   interpret FileArray_script
   call stream fileN, 'c', 'CLOSE'
   if Opt.0Progress then call Progress
   return 0
   /* === End of FileToArray ======================================================================== */

   CheckOut: procedure expose (Shared_vars) SrcOut. /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        call CheckOut <string>                                                                   */
   /*                                                                                                 */
   /*  if formatting, sends <string> to stderr                                                        */
   /*  if command is CHECKONLY, adds <string> to the SrcOut array                                     */
   /*                                                                                                 */
   /*  this is only used by the Task_CheckXXX routines                                                */
   /* =============================================================================================== */
   /*  this is a workaround-- the Task_CheckXXX routines were written to send warnings                */
   /*  to stderr when formatting a file. The formatted text is written to the SrcOut                  */
   /*  array, which may go to either stdout or a file. But when using the CheckOnly                   */
   /*  command, I want TaskCheckXXX to send output to the SrcOut array like the other                 */
   /*  informational commands do.                                                                     */
   /* =============================================================================================== */
   parse arg args
   if translate(G.0Command) \== 'CHECKONLY' then
      call lineout 'stderr', args
   else do
         ln = SrcOut.0 +1
         SrcOut.ln = args
         SrcOut.0 = ln
      end
   return 0
   /* === End of CheckOut =========================================================================== */

   Out: procedure expose (Shared_vars) SrcOut. /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        call Out <string>                                                                        */
   /*                                                                                                 */
   /*  adds <string> to the SrcOut. array                                                             */
   /*  the array must already have been initialized with                                              */
   /*        SrcOut. = ''                                                                             */
   /*        SrcOut.0 = 0                                                                             */
   /* =============================================================================================== */
   parse arg args
   ln = SrcOut.0 +1
   SrcOut.ln = args
   SrcOut.0 = ln
   return 0
   /* === End of Out ================================================================================ */

   Msg: /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        call Msg <text>                                                                          */
   /* ----------------------------------------------------------------------------------------------- */
   /*   sends <text> to stderr.                                                                       */
   /*  this exists because I got tired of typing "call lineout 'stderr',"                             */
   /* =============================================================================================== */
   call lineout 'stderr', arg(1)
   return 0
   /* === End of Msg ================================================================================ */

   Qline: procedure expose (Shared_vars) (line_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        say Qline(ln)                                                                            */
   /*  where ln is a line number of the SrcIn. array                                                  */
   /* ----------------------------------------------------------------------------------------------- */
   /*  this provides a consistent way to quote a line from the input file, for use                    */
   /*  in error messages. It is easier than typing                                                    */
   /*        ln = t_line.t +1                                                                         */
   /*        call lineout 'stderr', '   ==>  'strip(ScrIn.ln)                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  Many error messages use "ShowError ==> L_AsString(<token>)". The difference is                 */
   /*  a matter of who the error message is for. If your message is to tell a user                    */
   /*  about a problem line in his original file, use                                                 */
   /*           call Msg <message>                                                                    */
   /*           call Msg Qline(<line number>)                                                         */
   /*  If your message is for a developer about a possible bug in this program,                       */
   /*  use ShowError because it tells exactly where the error message comes from.                     */
   /*  If you want to show the line, L_AsString() reflects the current state of                       */
   /*  the token array.                                                                               */
   /*           call ShowError <message>                                                              */
   /*           call ShowError Qline(<line number>)                                                   */
   /*           call ShowError '  ==> 'L_AsString(<token>)                                            */
   /* =============================================================================================== */
   Esigl = sigl
   arg ln
   if ln < 1 ,
    | ln > SrcIn.0 then do
         call ShowError 'Invalid argument "'ln'" to Qline at line' Esigl
         call ShowError '  ==> 'sourceline(Esigl)
         call Egress 'BUG'
      end
   return '   ==>  'strip(SrcIn.ln)
   /* === End of Qline ============================================================================== */

   /***************************************************************************************************/
   /*  STANDARD INTERNAL FUNCTIONS:                                                                   */
   /***************************************************************************************************/

   Egress: /*fold00*/
   /* =============================================================================================== */
   /*  all exits come through here                                                                    */
   /* =============================================================================================== */
   G.0CalledFrom = sigl
   signal off NoValue
   signal off Error
   signal off Syntax
   signal off Halt
   if Opt.0idle == 1 then do
         /*  return priority to what we presume it started at:                                        */
         ret = SysSetPriority(2, 0)
         if ret \== 0 then call Mark 'ret'
      end
   parse arg ExitCode ReturnText
   ExitCode = translate(ExitCode)
   if Opt.0Progress then call Progress ''

   /*    handle my typing mistakes                                                                    */
   if ExitCode = 'OKAY' then exit 0
   if ExitCode = '' then exit 0
   if ExitCode = 0 then exit 0
   if ExitCode = 'HALT' then exit 1
   /*  note that the Halt handler does not work                                                       */
   if abbrev(ExitCode, 'USAGE') then exit 1
   if abbrev(ExitCode, 'FILE') then exit 2
   if abbrev(ExitCode, 'ERROR') then exit 3
   if abbrev(ExitCode, 'ERRORHANDLER') then exit 9
   if abbrev(ExitCode, 'BUG') then exit 99
   /*    in case I used some other exit string and forgot to add it here:                             */
   exit 99
   /* === End of Egress ============================================================================= */

   DumpVars: /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        CALL DumpVars <pattern>                                                                  */
   /*        ret = DumpVars(<pattern>)                                                                */
   /*  Dumps current variables to stdout. This can be called from anywhere.                           */
   /*  The optional parameter is a name pattern; only variables that start with that string           */
   /*  will be shown.                                                                                 */
   /* Most of the time 'CALL Mark' is better                                                          */
   /*                                                                                                 */
   /*  fix: I have not looked at this in a long time; probably should remove it.                      */
   /* =============================================================================================== */
   do while queued() ; call linein 'queue:' ; end
   call RxVlist arg(1), 'V'
   return DumpVars_2()
   /* === End of DumpVars =========================================================================== */

   DumpVars_2: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  ret = DumpVars_2(<[list of variable names>])                                                   */
   /* ----------------------------------------------------------------------------------------------- */
   /*  I want a one-line command to dump all variables in the current environment and filter them.    */
   /*  fix: WrapVar() assumes the value is a string of words and that spaces are not significant,     */
   /*       which is not good enough. It should preserve all spaces, splitting string at spaces if    */
   /*       possible, insert single quotes, "||", and line-continuation comma                         */
   /*       and it should replace non-printable characters with hex strings                           */
   /*       and it should replace comment_beg with '/'||'*'                                           */
   /*       and maybe deal with other dangerous strings                                               */
   /*                                                                                                 */
   /* =============================================================================================== */

   v = 0
   varN. = '' ; varN.0 = 0
   varV. = '' ; varV.0 = 0

   do while queued()  > 0
      v = v +1
      varN.v = left(linein('queue:'), 50)||v
      varV.v = linein('queue:')
      varN.0 = v
      varV.0 = v
   end

   if varV.0 == 0 then do
         call ShowError 'no variables found'
         return 99
      end

   ret = SysStemSort('varN.', 'A', 'I')
   varNameLen = 30
   Indent = varNameLen +4
   maxValLen = G.0ScreenCols - varNameLen -20
   do v = 1 to varN.0
      parse var varN.v varname varNum .
      if varname = 'SIGL' then iterate
      if varname = 'RESULT' then iterate
      if varname = 'RC' then iterate
      varname = left('  '||varname, varNameLen)||'=   '
      varVal = varV.varNum
      if WordsQ(varVal) > 1 then
         varVal = WrapVar("'"||varVal||"'", maxValLen, Indent)
      else
         varVal = "'"||varVal||"'"
      call lineout 'stdout', varname||varVal
   end
   return 0
   /* === End of DumpVars_2 ========================================================================= */

   WrapVar: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  this is used only by DumpVars_2                                                                */
   /* ----------------------------------------------------------------------------------------------- */
   /*  splits a long variable value into shorter strings, adding single quotes and                    */
   /*  line-continuation to produce a variable assignment statement                                   */
   /* ----------------------------------------------------------------------------------------------- */
   /*  fix: it assumes the string is made of words and that spaces are not significant.               */
   /*  fix: find or write a general-purpose text wrapper                                              */
   /* =============================================================================================== */

   len = 80
   Indent = 0

   string = arg(1)
   if arg(2, 'e') then len = arg(2)
   if arg(3, 'e') then Indent = arg(3)
   if length(string) < len then return string
   oldstring = ''
   newstring = ''
   do while string \= ''
      parse var string aword string
      if length(newstring||' '||aword) < len then newstring = newstring||' '||aword
      else do
            oldstring = strip(oldstring, 'l')||newstring||"'"||" ,"||G.0CRLF||copies(' ', Indent)
            newstring = "'"||aword
         end
   end
   return strip(oldstring||newstring)
   /* === End of WrapVar ============================================================================ */

   ShowError: procedure expose (Shared_vars) /*fold00*/
   0ShowError:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        call ShowError <Arg1> [, <Arg2>]                                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*   sends a message to stderr, adding a header with the program name, function name,              */
   /*   and line number to show where the message comes from.                                         */
   /*   Use '' as the second arg to improve the display of a multi-line message:                      */
   /*                                                                                                 */
   /*   Example: in function Add, to show a multi-line message:                                       */
   /*       call ShowError 'Warning: unexpected value for variable varA: "'varA'"'                    */
   /*       call ShowError 'It should be a number', ''                                                */
   /*   displays:                                                                                     */
   /*       [Foo.cmd:Add:33] Warning: unexpected value for variable varA: "foo"                       */
   /*                        It should be a number                                                    */
   /*                                                                                                 */
   /* =============================================================================================== */
   seSigl = sigl

   if Opt.0verbose \= 0 then
      seID = '['G.0me':'LastLabel(seSigl)':'seSigl']'
   else
      seID = '['G.0me':'seSigl']'

   if arg(2, 'e') ,
    & arg(2) = '' then
      seID = copies(' ', length(seID))
   else
      call lineout 'stderr', ''

   call lineout 'stderr', seID arg(1)
   return 0
   /* === End of 0ShowError ========================================================================= */

   Mark: /*fold00*/
   0Mark:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       CALL Mark [<list of variable names>] [, <string>]                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  Shows the filename and line number it was called from, the current values of the               */
   /*  variables named, and an informational string                                                   */
   /*  Arg(1) is a list of variable names                                                             */
   /*  Arg(2) can be any descriptive text                                                             */
   /*  Examples:                                                                                      */
   /*      CALL mark 'varA varB', 'At top of the "DO x = 1 to 5" loop'                                */
   /*  If no variable names are given, it will just show a message:                                   */
   /*      CALL mark '', 'At top of the "DO x = 1 to 5" loop'                                         */
   /*                                                                                                 */
   /*  This must share all the variables of whatever routine it was called from.                      */
   /*  I cannot think of a way to do it without setting at least one variable here,                   */
   /*  so reserve the variable '_' for the use of Mark and similar routines.                          */
   /* =============================================================================================== */
   _ = sigl
   call lineout 'stderr', ''
   if arg(2, 'e') ,
    & arg(2) = '' then nop
   else call lineout 'stderr', '[Mark]['value('G.0me')':'_']' arg(2)
   do _ = 1 to 99
      if word(arg(1), _) = '' then leave _
      if symbol(word(arg(1), _)) = 'VAR' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||'= "'||value(word(arg(1), _))||'"'
      if symbol(word(arg(1), _)) = 'LIT' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||' is not defined'
      if symbol(word(arg(1), _)) = 'BAD' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||' is not a variable'
   end _
   drop _
   return 0
   /* === End of 0Mark ============================================================================== */

   LastLabel: procedure expose (Shared_vars) /*fold00*/
   0LastLabel:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       ret = LastLabel()                                                                         */
   /*       ret = LastLabel(<line number>)                                                            */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns the name of the nearest label =in this program= previous to <line number>              */
   /*  if <line> is not given then it is the current line                                             */
   /*  returns '' if no label was found or if <line> is not a valid line number                       */
   /*                                                                                                 */
   /*  it is normally used to determine the name of a routine or function:                            */
   /*      Foo: procedure                                                                             */
   /*      say 'Now in procedure 'Lastlabel()         > "Now in procedure Foo"                        */
   /*                                                                                                 */
   /*  akm: This partly depends on the style of writing I use, where                                  */
   /*  multiline comments begin at the beginning of a line and end at the end of a                    */
   /*  line. It could be fooled by something that looks like a label inside other                     */
   /*  kinds of multiline comments. But it is useful, and if it is wrong it will only                 */
   /*  affect the progress and debugging messages. This is about as reliable as I can                 */
   /*  make it without parsing every comment letter by letter.                                        */
   /* =============================================================================================== */
   if arg(1, 'o') then sline = sigl ; else sline = arg(1)
   if datatype(sline, 'W') \= 1 then return ''
   if sline > sourceline() then return ''
   incomment = 0
   do l = sline to 1 by -1
      aline = strip(sourceline(l))
      if left(aline, 2) == '/'||'*' ,
       | right(aline, 2) == '*'||'/' then do
            /*  counter-intuitively, incomment is incremented by the end of comment,                  */
            /*  because we are working backwards                                                      */
            incomment = incomment - CountStr('/'||'*', aline) + CountStr('*'||'/', aline)
         end

      if incomment <= 0 then do
            if HasLabel(aline) then do
                  parse var aline label ':' .
                  label = strip(label)
                  return label
               end
         end
   end l
   return ''
   /* === End of 0LastLabel ========================================================================= */

   HasLabel: procedure expose (Shared_vars) /*fold00*/
   0HasLabel:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        if HasLabel(<string>) then ...                                                           */
   /*  returns True if <string> is a line that looks like a label                                     */
   /*  the calling routine is responsible for identifying multi-line comments                         */
   /* =============================================================================================== */
   /*  it is a label if str is one word followed by ':', followed by '', 'procedure', or a comment    */
   parse arg str
   str = strip(str)
   if pos(':', str) == 0 then return 0
   if left(str, 2) == '/'||'*' then return 0
   parse var str label ':' rest
   rest = strip(rest)
   if words(label) \= 1 then return 0
   if symbol(label) == 'BAD' then return 0
   if rest \= '' then
      select
         when translate(word(rest, 1)) == 'PROCEDURE' then
            return 1
         when left(rest, 2) == '/'||'*' then
            return 1
         otherwise
            return 0
      end   /*  select  */
   return 1
   /* === End of 0HasLabel ========================================================================== */

   SubID: /*fold00*/
   /* =============================================================================================== */
   /*  SubID()                                                                                        */
   /*  returns a string identifying the label name of the caller                                      */
   /* ----------------------------------------------------------------------------------------------- */
   /*  depends on: SIGL, LastLabel()                                                                  */
   /* =============================================================================================== */
   return '['LastLabel(sigl)']'
   /* === End of SubID ============================================================================== */

   /***************************************************************************************************/
   /* Error handlers section:                                                                         */
   /***************************************************************************************************/

   RunOut_handler: /*fold00*/
   /* =============================================================================================== */
   /*  signal Handler_RunOut                                                                          */
   /* ----------------------------------------------------------------------------------------------- */
   /*  depends on variables: G.0meID                                                                  */
   /*  depends on functions: LastLabel(), Egress(), ShowError()                                       */
   /* ----------------------------------------------------------------------------------------------- */
   /*  can be called when a routine does not return and execution falls through into the              */
   /*  next routine. "signal RunOut_handler" after the end of a routine can catch the error before    */
   /*  Rexx does, and give a more accurate idea of where the problem is.                              */
   /*  It does not usually catch errors caused by an unbalanced DO/END.                               */
   /* =============================================================================================== */
   Esigl = sigl
   Err_func = LastLabel(Esigl)
   IDstring = G.0meID||'[RunOut_Handler]'
   call ShowError IDstring 'Failed to return from procedure"'Err_func'" at line 'Esigl
   call Egress 'Bug_Runout'
   /* === End of RunOut_handler ===================================================================== */

   Toker: procedure expose (Shared_vars) (line_vars) (all_tokens) /*fold00*/
   /* =============================================================================================== */
   /*  this is the entry point of the tokeniser                                                       */
   /*  The tokeniser breaks up 4OS2 code into a list of tokens, identified by type, and               */
   /*  adds them to the list represented by stem variables t_type., t_val., etc.                      */
   /*                                                                                                 */
   /*  akm: the various Toker_XXX routines are all part of the tokeniser; I split it up into          */
   /*        parts so I can understand what I am doing.                                               */
   /* =============================================================================================== */
   /*    It took me a long time to understand the Tokeniser. These are notes to myself                */
   /*    in case I need to understand it again.                                                       */
   /*                                                                                                 */
   /*    variables that contain a line of text start with "L_"                                        */
   /*        SrcIn.      == an array of lines from the source file                                    */
   /*        SrcIn.0     == total number of lines                                                     */
   /*                                                                                                 */
   /*   the working variables in the Tokeniser loop:                                                  */
   /*        st_line      == current line number                                                      */
   /*        c            == a 1-character string; the current character being examined               */
   /*        c2           == a 1-character string; the next after the current character               */
   /*        cc           == a 2-character string; 'c' plus the following character                   */
   /*        p            == active position in the line                                              */
   /*                        note that 'p' is one position to the right of 'c'                        */
   /*                                                                                                 */
   /*        st_col       == the column where the current token starts.                               */
   /*                        When the end of token is found, the token is stored and                  */
   /*                        st_col is updated to the start of the next token.                        */
   /*        incomment    == comment nesting level, it can be > 1 if in a nested comment              */
   /*                                                                                                 */
   /*    most of the variables that will be used by Toker_store start with "st_"                      */
   /*                                                                                                 */
   /*  The end result is a set of arrays, holding the tokens' class, type, line, column,              */
   /*  prefix, and value.                                                                             */
   /*                                                                                                 */
   /*  the tokeniser loop does something like this:                                                   */
   /*        Initialize the search for the next token                                                 */
   /*            set 'st_col' to the current position 'p', presumably the start of the token          */
   /*            set 'c' to the 1 character at position 'p'                                           */
   /*            set 'cc' to the 2 characters at position 'p'                                         */
   /*            increment 'p'                                                                        */
   /*        If there are leading spaces before a token then                                          */
   /*            move p to where the token really starts                                              */
   /*            save the string of spaces. they are the 'prefix' attribute of the token              */
   /*            Initialize again                                                                     */
   /*        Examine 'c' and 'cc' to identify a token.                                                */
   /*            Sometimes that involves subordinate loops, incrementing 'p', and using               */
   /*            extra variables like 'c2' to see what characters come after 'c'.                     */
   /*        At any place in the loop where a token is identified--                                   */
   /*            set 'st_val' and 'st_prefix' to the contents of the token and its prefix             */
   /*            set st_class and st_type to describe the token                                       */
   /*            set st_line and st_col to the line and column where the token was found              */
   /*            call Toker_Store to add the token to the list                                        */
   /*            ensure that 'p' is set correctly for the next search                                 */
   /*            ==> and do nothing else                                                              */
   /*            iterate                                                                              */
   /*                                                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  Routines that start with "Toker_" are all part of the tokeniser.                               */
   /* =============================================================================================== */

   /* =============================================================================================== */
   /*  initialize many variables used by Toker_* routines                                             */
   /* =============================================================================================== */
   /*  the arrays that will hold the tokens and attributes:                                           */
   t_class.      = ''   ; t_class.0      = 0                                   /* class of each token */
   t_type.       = ''   ; t_type.0       = 0                                    /* type of each token */
   t_line.       = 0    ; t_line.0       = 0                                    /* line of each token */
   t_col.        = 0    ; t_col.0        = 0                                  /* column of each token */
   t_prefix.     = ''   ; t_prefix.0     = 0                                  /* prefix of each token */
   t_val.        = ''   ; t_val.0        = 0                                   /* value of each token */
   /*                                                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  $format rexx off                                                                               */
   Tkr. = ''
   Tkr.0mode = '4OS2'
   /*  note the Tab and Space at start, and escaped single quote                                      */
   Tkr.0charset = '09'x||' @$_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ?!' ,
    ||'1234567890+/-*%&|=<>\' ,
    ||'.()[]:`"'';,'

   /*  this is the normal list of character types, where words and numbers are all type W_SYMBOL.     */
   Tkr.0typeSymbol = 'bbsssssssssssssssssssssssssssssssssssssssssssssssssssssssss' ,
    ||'ssssssssssssssoooooos' ,
    ||'s()[]s`"ss,'

   /*  fix: needs testing                                                                             */
   /*  types as used inside most functions, where blanks are symbols, not prefixes                    */
   Tkr.0typeFunction = 'sssssssssssssssssssssssssssssssssssssssssssssssssssssssssss' ,
    ||'ssssssssssssssossssss' ,
    ||'s()[]s`"s;,'

   /*  types as used inside @eval[], @inc[], @dec[] functions                                         */
   /*  note that comma is also a number                                                               */
   Tkr.0typeMath = 'bbsssssssssssssssssssssssssssssssssssssssssssssssssssssssss' ,
    ||'nnnnnnnnnnooooooooooo' ,
    ||'n()[]s`"ssn'

   /*  types as used inside a TEXT/ENDTEXT block                                                      */
   /*  fix: if newer version of 4OS2 interprets variables inside text, this will need to be changed   */
   Tkr.0typeText = 'sssssssssssssssssssssssssssssssssssssssssssssssssssssssssss' ,
    ||'ssssssssssssssossssss' ,
    ||'sssssssssss'

   Tkr.0typeset = Tkr.0typeSymbol
   /*  $format rexx on                                                                                */
   /* ----------------------------------------------------------------------------------------------- */

   /*  This list includes ELSE, CASE, ENDSWITCH, etc. as Keywords. That is not how                    */
   /*  4OS2.exe works, but it is simpler for our purposes.  For indenting, any word                   */
   /*  that can begin a line is relevant, even if it is not a real keyword.                           */
   /*  '::' is considered a synonym of the keyword 'REM'                                              */
   Tkr.0Keywords = translate(G.0Keywords)

   Tkr.0token_vars = 't_class. t_type. t_prefix. t_val. t_line. t_col.'

   Tkr.0ExitCode = ''

   /* ----------------------------------------------------------------------------------------------- */
   /*  the DEBUGGING section:                                                                         */
   /*  When working on something I can set G.0track_vars to the names of whatever variables I am      */
   /*  interested in. If one of the errorhandlers is triggered, it shows the values of whatever       */
   /*  variables are listed in G.0track_vars.                                                         */
   G.0track_vars = ''
   /* ----------------------------------------------------------------------------------------------- */

   t = 0                                                                      /* current token number */
   c = ''                                                                  /* character at position p */
   cc = ''                                                              /* 2 characters at position p */
   c2 = ''
   c3 = ''
   p = 1                                                                      /* position in the line */
   st_class = '4OS2'                         /* st_xxx variables are set here and read by Toker_store */
   st_type = ''
   st_prefix = ''
   st_val = ''
   st_line = 1                                                                  /* source line number */
   st_col = p                                                            /* start position of a token */

   st_vars = 'st_class st_type st_prefix st_val st_line st_col'
   toker_tokens = 't_class. t_type. t_prefix. t_val. t_line. t_col.'
   loop_ctr = 0

   /*  I add a Beginning of File token just to make it easier to write some things.                   */
   /*  Without it, a loop that uses token t-1 on each pass will error if t = 1.                       */
   if t = 0 then do
         st_class = 'FILE_BOF'
         st_type = 'EOC'
         st_val = ''
         st_col = 1
         call Toker_Store
      end

   inbrackets = 0             /* counts recursion level when calling itself for bracketed expressions */
   BackQ = 0                                            /* 1 if current position is inside backquotes */
   DoubleQ = 0                                       /* 1 if current position is inside double quotes */
   continuation = ''                                                          /* not a continued line */

   /*  get ready for the first line. This is the same thing that Toker_do_EOL does at                 */
   /*  the beginning of each line.                                                                    */
   select
      when right(SrcIn.st_line, 2) == '%=' then do
            continuation = '%='
            ThisLine = left(SrcIn.st_line, length(SrcIn.st_line) -2)
         end
      when right(SrcIn.st_line, 1) == '^' then do
            continuation = '^'
            ThisLine = left(SrcIn.st_line, length(SrcIn.st_line) -1)
         end
      otherwise
         continuation = ''
         ThisLine = SrcIn.st_line
   end   /*  select  */

   L_len = length(ThisLine)
   p = 1

   /*  do not change the order of the Toker_xxx procedures; some depend on the ones before            */
   call Toker_4OS2
   call Toker_KeyWords
   return Tkr.0ExitCode
   /* === End of Toker ============================================================================== */

   Toker_4OS2: /*fold00*/
   /* =============================================================================================== */
   /*  IMPORTANT: to prevent stupid 'endless loop' mistakes, do the tests in this order:              */
   /*      check for EOF (is line > last line)                                                        */
   /*      check for EOL (is p > last column of line)                                                 */
   /*      initialize for a new token                                                                 */
   /*      then do the string comparisons                                                             */
   /* ----------------------------------------------------------------------------------------------- */
   /*  Text inside double quotes needs special handling, because we have to                           */
   /*  distinguish between literal strings and variables/functions, which are not                     */
   /*  protected by quotes. Hence the recursion here-- when Toker_4OS2 finds a string                 */
   /*  in double quotes, it calls Toker_do_Quote. When Toker_do_Quote finds an                        */
   /*  expression inside square brackets (in a function), it calls Toker_4OS2 to parse                */
   /*  that section. This saves me from adding a lot of things to Toker_do_Quote that                 */
   /*  would duplicate Toker_4OS2. Variables inbrackets and inquotes track the nesting                */
   /*  levels.                                                                                        */
   /*                                                                                                 */
   /*  Note that most of the Toker_xxx routines are not procedures-- all variables are                */
   /*  shared and there are no local variables, even when they are called recursively.                */
   /*                                                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   if Opt.0Progress then call Progress G.0meID'[Toker] looking for tokens...'
   /*
   new plan:
      at each new line
         check for continued line;
            remove ^ or %= from line and save as "continuation"
      at each end of line,
         if continuation \= '' then
            save continuation as token with type CONTINUE
            save EOL as token with type NULL
            keep status flags-- inside backquotes, inside double quotes, inside brackets
         else
            save EOL as type EOC
            if in a quoteed string or brackets, unset flag and warn
         set up for new line

      for quoted strings, keep separate inquote flags
   */
   /* ----------------------------------------------------------------------------------------------- */
   /*  fix: organize the initialization lines in Toker, maybe move some here                          */
   do forever
      /* -------------------------------------------------------------------------------------------- */
      /*  EOF  (end of file)                                                                          */
      /*  make sure the file ends with EOL token                                                      */
      if st_line > SrcIn.0  then do
            tokHigh = t_type.0
            if t_class.tokHigh \== 'FILE_EOL' then do
                  call ShowError  'Tokeniser error: reached the last line but there is no EOL token'
                  call ShowError  'This is probably a bug in the tokeniser', ''
                  st_class = 'FILE_EOL'
                  st_type = 'EOC'
                  st_val = ''
                  st_col = L_len +1
                  call Toker_Store
                  Tkr.0ExitCode = Tkr.0ExitCode 'BUG'
               end
            if inbrackets \== 0 ,
             | BackQ \== 0 ,
             | DoubleQ \== 0 then do
                  call ShowError 'Tokeniser error: these variables should be 0 at the end of the file.'
                  call ShowError 'This could be a defective file or a bug in the Tokeniser.', ''
                  call Mark 'inbrackets BackQ DoubleQ'
                  Tkr.0ExitCode = Tkr.0ExitCode 'BUG'
               end
            st_class = 'FILE_EOF'
            st_type = 'EOC'
            st_val = ''
            st_col = L_len +1
            call Toker_Store
            leave
         end
      /* -------------------------------------------------------------------------------------------- */
      /*  EOL (end of line)                                                                           */
      if p > L_len then do

            ret = Toker_do_EOL()
            /*  bugfix for a missing "]". Toker_do_EOL will cancel inbrackets and                     */
            /*  tell us to return now                                                                 */
            if translate(ret) == 'RETURN' then
               return 0
            else
               iterate
         end
      /* -------------------------------------------------------------------------------------------- */
      if Tkr.0mode == 'TEXT' then do
            if translate(strip(SrcIn.st_line)) == 'ENDTEXT' then do
                  st_class = '4OS2'
                  Tkr.0typeset = Tkr.0typeSymbol
                  Tkr.0mode = '4OS2'
               end
            else do
                  st_val = SrcIn.st_line
                  st_class = 'TEXT'
                  st_type = 'TEXT'
                  call Toker_Store
                  p = L_len +1
                  iterate
               end
         end
      /* -------------------------------------------------------------------------------------------- */
      /*  set up for the next token:                                                                  */
      if \DoubleQ then
         st_class = '4OS2'
      else
         st_class = 'QUOTED'
      st_type = ''
      st_prefix = ''
      st_val = ''
      st_col = p
      c = MidStr(ThisLine, p, 1)
      c2 = MidStr(ThisLine, p +1, 1)
      cc = MidStr(ThisLine, p, 2)
      p = p +1

      if debug.0loop then call LoopCounter
      /* -------------------------------------------------------------------------------------------- */
      /*  BLANK is a <space> or <tab> character.                                                      */
      /*  The leading spaces before a token are the prefix. No need to iterate-- just save the        */
      /*  blanks to the prefix variable, move p to the start of the token, and keep going.            */
      type = translate(c, Tkr.0typeset, Tkr.0charset)
      if type == 'b' then do
            do while MidStr(SrcIn.st_line, p, 1) == ' ' ,
                   | substr(SrcIn.st_line, p, 1) == '09'x
               p = p +1
            end
            st_prefix = MidStr(ThisLine, st_col, p - st_col)
            st_col = p
            c = MidStr(ThisLine, p, 1)
            cc = MidStr(ThisLine, p, 2)
            c2 = MidStr(ThisLine, p +1, 1)
            p = p +1
         end
      /* -------------------------------------------------------------------------------------------- */

      /* -------------------------------------------------------------------------------------------- */
      /*  EMPTY LINE (next pass will store the EOL token)                                             */
      if c == '' then iterate
      /* -------------------------------------------------------------------------------------------- */
      /*  check type again because c may have changed.                                                */
      /*  at this point we know it is the start of some kind of 4OS2 token.                           */
      type = translate(c, Tkr.0typeset, Tkr.0charset)

      /* -------------------------------------------------------------------------------------------- */
      /*  COMMENT is the rest of the line,                                                            */
      /*  that includes the continuation, so use SrcIn.st_line, not ThisLine                          */
      if Tkr.0mode == 'REM' then do
            st_val = MidStr(SrcIn.st_line, p -1)
            continuation = ''
            st_class = 'COMMENT'
            st_type = 'COMMENT_VAL'
            call Toker_Store
            p = L_len +1
            Tkr.0mode = '4OS2'
            iterate
         end
      /* -------------------------------------------------------------------------------------------- */
      if c == ']' then do
            st_val = MidStr(ThisLine, st_col, p - st_col)
            st_type = 'W_]'
            call Toker_Store
            if inbrackets > 0 then
               return 0
            else
               iterate
         end
      /* -------------------------------------------------------------------------------------------- */
      /*  LITERAL in BACKQUOTES                                                                       */
      if c == '`' then do
            p = p -1
            call Toker_do_BackQuote
            iterate                                                            /* to top of main loop */
         end   /*  end if quoted string  */
      /* -------------------------------------------------------------------------------------------- */
      /*  string in DOUBLE QUOTES                                                                     */
      if c == '"' then do
            p = p -1
            st_class = 'QUOTED'
            call Toker_do_Quote
            if \DoubleQ then
               st_class = '4OS2'
            iterate                                                            /* to top of main loop */
         end
      /* -------------------------------------------------------------------------------------------- */

      /* -------------------------------------------------------------------------------------------- */
      /*  SYMBOL                                                                                      */
      if type == 's' then do
            /*  bugfix--  "!=" means "ne".  "!" is a symbol but "=" is not                            */
            if cc == '!=' then do
                  p = p +1
                  st_val = MidStr(ThisLine, st_col, p - st_col)
                  st_type = 'W_COMPARE'
                  call Toker_Store
                  iterate
               end

            /*  @ and * are separate tokens only if at the start of a clause                          */
            if c == '@' then do
                  Last = t_type.0
                  if t_type.Last == 'EOC' then do
                        st_val = MidStr(ThisLine, st_col, p - st_col)
                        st_type = '@'
                        call Toker_Store
                        iterate
                     end
               end

            if c == '*' then do
                  Last = t_type.0
                  /*  anything else that would mean the beginning of a statement?                     */
                  if wordpos(t_type.Last, 'EOC @ W_( W_SEP') > 0 then do
                        st_val = MidStr(ThisLine, st_col, p - st_col)
                        st_type = '*'
                        call Toker_Store
                        iterate
                     end
               end

            /*  COLON                                                                                 */
            /*  recognize '::' as one token because it is sometimes used instead of 'rem'             */
            if c == ':' then do
                  c2 = MidStr(ThisLine, p, 1)
                  if c \= c2 then do
                        st_val = MidStr(ThisLine, st_col, p - st_col)
                        st_type = 'COLON'
                        call Toker_Store
                        iterate
                     end
                  p = p +1
                  st_val = MidStr(ThisLine, st_col, p - st_col)
                  st_type = 'COLON2'
                  call Toker_Store
                  iterate
               end   /*  if c == ':'  */

            /*  advance position to a character that is not a legal symbol character                  */
            do forever
               if p > L_len then leave
               c = MidStr(ThisLine, p, 1)
               cc = MidStr(ThisLine, p, 2)
               type = translate(c, Tkr.0typeset, Tkr.0charset)
               if verify(type, 'sn.') > 0 then leave
               p = p +1
            end
            st_val = MidStr(ThisLine, st_col, p - st_col)
            st_type = 'W_SYMBOL'
            call Toker_Store
            iterate
         end   /*  if type == 's'  */
      /* -------------------------------------------------------------------------------------------- */

      /* -------------------------------------------------------------------------------------------- */
      if type == 'o' then do
            if Tkr.0mode == 'MATH' then do
                  if pos(c, '+-/\') > 0 then do
                        /* one-character operator                                                     */
                        st_val = MidStr(ThisLine, st_col, p - st_col)
                        st_type = 'W_MATH'
                        call Toker_Store
                        iterate
                     end
                  if pos(c, '%*') > 0 then do
                        /*  could be one or two-character operator                                    */
                        if c2 == c then
                           p = p +1
                        st_val = MidStr(ThisLine, st_col, p - st_col)
                        st_type = 'W_MATH'
                        call Toker_Store
                        iterate
                     end
                  call ShowError  'Unexpected MATH operator character: "'c'"'
                  call ShowError  'at position 'p' of line 'st_line, ''
                  call ShowError  '  ==> {'SrcIn.st_line'}', ''
                  Tkr.0ExitCode = Tkr.0ExitCode 'FILE'
                  st_val = MidStr(ThisLine, st_col, p - st_col)
                  st_type = 'W_UNKNOWN'
                  call Toker_Store
                  iterate
               end   /*  mode == MATH  */
            else do                                                                   /* mode is 4OS2 */
                  if c == '=' then do
                        /*  "==" is the same as "eq"                                                  */
                        if c2 == '=' then do
                              p = p +1
                              st_val = MidStr(ThisLine, st_col, p - st_col)
                              st_type = 'W_COMPARE'
                              call Toker_Store
                              iterate
                           end
                        else do
                              st_val = MidStr(ThisLine, st_col, p - st_col)
                              st_type = 'W_OP'
                              call Toker_Store
                              iterate
                           end
                     end
                  if c == '<' then do
                        /* redirection of stdin                                                       */
                        st_val = MidStr(ThisLine, st_col, p - st_col)
                        st_type = 'W_REDIR'
                        call Toker_Store
                        iterate
                     end

                  if c == '>' then do
                        /*            redirection of stdout and stderr                                */
                        do until pos(c, '>&') == 0
                           c = MidStr(ThisLine, p, 1)
                           p = p +1
                        end
                        p = p -1
                        st_val = MidStr(ThisLine, st_col, p - st_col)
                        st_type = 'W_REDIR'
                        call Toker_Store
                        iterate
                     end

                  /* EOL, must be end of word. This should never happen                               */
                  if p > L_len then do
                        st_val = MidStr(ThisLine, st_col, p - st_col)
                        st_type = 'W_UNKNOWN'
                        call Toker_Store
                        call ShowError 'BUG: unknown token type at end of line'
                        call ShowError Qline(st_line), ''
                        call Mark 'thisline st_val st_type'
                        Tkr.0ExitCode = Tkr.0ExitCode 'BUG'
                        iterate
                     end

                  c2 = MidStr(ThisLine, p, 1)
                  if c == '&' then do
                        /*  & or && separators                                                        */
                        if c2 == '&' then
                           p = p +1
                        st_val = MidStr(ThisLine, st_col, p - st_col)
                        st_type = 'W_SEP'
                        call Toker_Store
                        st_col = p
                        iterate
                     end

                  if c == '|' then
                     if c2 == '|' then do
                           /*  || conditional separator                                               */
                           p = p +1
                           st_val = MidStr(ThisLine, st_col, p - st_col)
                           st_type = 'W_SEP'
                           call Toker_Store
                           st_col = p
                           iterate
                        end
                     else do
                           /*  | or |& pipe separator                                                 */
                           if c2 == '&' then
                              p = p +1
                           st_val = MidStr(ThisLine, st_col, p - st_col)
                           st_type = 'W_PIPE'
                           call Toker_Store
                           st_col = p
                           iterate
                        end

                  /* % or %%                                                                          */
                  if c == '%' then do
                        select
                           when c2 == '+' then do
                                 p = p +1
                                 st_val = MidStr(ThisLine, st_col, p - st_col)
                                 st_type = 'W_SEP'
                                 call Toker_Store
                                 st_col = p
                                 iterate
                              end
                           when c2 == '=' then do
                                 p = p +1
                                 st_val = MidStr(ThisLine, st_col, p - st_col)
                                 st_type = 'W_ESC'
                                 call Toker_Store
                                 st_col = p
                                 iterate
                              end
                           otherwise
                              if c2 == '%' then
                                 p = p +1
                              st_val = MidStr(ThisLine, st_col, p - st_col)
                              st_type = 'W_%'
                              call Toker_Store
                              st_col = p
                              /*  Toker_do_variable will store the variable name or function name     */
                              call Toker_do_Variable
                              iterate
                        end   /*  select  */
                     end   /*  if c == %  */

                  /*  if execution gets to here without iterating, there is something wrong           */
                  call ShowError  'Unexpected operator character: "'c'"'
                  call ShowError  'at position 'p' of line 'st_line, ''
                  call ShowError  '  ==> {'SrcIn.st_line'}', ''
                  Tkr.0ExitCode = Tkr.0ExitCode 'FILE'
                  st_val = MidStr(ThisLine, st_col, p - st_col)
                  st_type = 'W_UNKNOWN'
                  call Toker_Store
                  iterate
               end   /*  mode == 4OS2  */
         end   /*  end if type == 'o'  */
      /* -------------------------------------------------------------------------------------------- */

      /* -------------------------------------------------------------------------------------------- */
      /*  CONSTANT                                                                                    */
      /*  This only applies when inside a math function.                                              */
      /*  For now, tag numbers as W_SYMBOL. If I decide I want a separate type for                    */
      /*  numbers, just change "W_SYMBOL" to "W_CONSTANT" here.                                       */
      if type == 'n' then do
            do forever
               if p > L_len then leave
               c = MidStr(ThisLine, p, 1)
               type = translate(c, Tkr.0typeset, Tkr.0charset)
               if verify(type, 'n.') > 0 then leave
               p = p +1
            end
            st_val = MidStr(ThisLine, st_col, p - st_col)
            st_type = 'W_SYMBOL'
            call Toker_Store
            iterate
         end
      /* -------------------------------------------------------------------------------------------- */

      /* -------------------------------------------------------------------------------------------- */
      /*  OTHER                                                                                       */
      /*  akm: I use longer names for some just because ':' ';' and ',' are hard to read              */
      st_val = MidStr(ThisLine, st_col, p - st_col)
      select
         when st_val = ',' then st_type = 'W_COMMA'
         when st_val = ':' then st_type = 'COLON'
         otherwise st_type = 'W_'||type
      end
      call Toker_Store
      iterate
      /* -------------------------------------------------------------------------------------------- */
   end   /*  of outer DO forever loop  */
   if Opt.0Progress then call Progress
   return Tkr.0ExitCode
   /* === End of Toker_4OS2 ========================================================================= */

   Toker_do_EOL: /*fold00*/
   /* =============================================================================================== */
   /*  store tokens at end of a line and set up for next line                                         */
   /* =============================================================================================== */
   Esigl = sigl
   ret = 0
   if continuation \= '' then do
         /*  tokeniser should already have set st_prefix from the trailing blanks of the line         */
         /*  store CONTINUE token first                                                               */
         st_val = continuation
         st_class = '4OS2'
         st_type = 'CONTINUE'
         st_col = L_len +1
         call Toker_Store
         st_type = 'NULL'
      end
   else do
         st_type = 'EOC'
      end
   st_class = 'FILE_EOL'
   st_val = ''
   st_col = L_len +1
   call Toker_Store
   st_line = st_line +1
   st_class = '4OS2'

   /*  if doing a bracketed expression, if we reached the EOL without finding                         */
   /*  the closing "]", it is probably a missing "]" or missing quote.                                */
   /*  Show a message, assume it is the end of the bracketed text, and return.                        */
   if continuation == '' then do
         if inbrackets > 0 then do
               if Opt.0Warn then do
                     call ShowError 'Tokeniser warning:' ,
                          'failed to find closing bracket at line' st_line -1
                     call ShowError Qline(st_line -1), ''
                     ret = 'RETURN'
                  end
            end
         /*  if not a continued line, the EOL cancels any unclosed quotes and brackets                */
         /*  also cancels REM mode (bugfix for REM with no comment), but does not cancel TEXT mode    */
         DoubleQ = 0
         BackQ = 0
         inbrackets = 0
         if Tkr.0mode == 'REM' then Tkr.0mode = '4OS2'
      end

   /* ----------------------------------------------------------------------------------------------- */
   /*  set up for the next line                                                                       */
   /*  does the new line end in a continuation character?                                             */
   select
      when right(SrcIn.st_line, 2) == '%=' then do
            continuation = '%='
            ThisLine = left(SrcIn.st_line, length(SrcIn.st_line) -2)
         end
      when right(SrcIn.st_line, 1) == '^' then do
            continuation = '^'
            ThisLine = left(SrcIn.st_line, length(SrcIn.st_line) -1)
         end
      otherwise
         continuation = ''
         ThisLine = SrcIn.st_line
   end   /*  select  */

   L_len = length(ThisLine)
   p = 1
   st_col = p

   return ret
   /* === End of Toker_do_EOL ======================================================================= */

   Toker_do_Variable: /*fold00*/
   /* =============================================================================================== */
   /*  does one variable. The caller has already stored the '%', and p is at the first                */
   /*  letter of the name                                                                             */
   /* =============================================================================================== */
   Esigl = sigl
   c = MidStr(ThisLine, p, 1)
   p = p +1
   if c == '[' then do
         /*  a string like "%[ab##]" needs special handling because it can include                    */
         /*  characters that are not normally allowed in a variable                                   */
         st_val = MidStr(ThisLine, st_col, p - st_col)
         st_type = 'W_['
         call Toker_Store
         st_col = p

         c2 = MidStr(ThisLine, p +1, 1)
         p = p +1
         do while c2 \== ']'
            c2 = MidStr(ThisLine, p +1, 1)
            p = p +1
            if p > L_len then leave
         end
         st_val = MidStr(ThisLine, st_col, p - st_col)
         st_type = 'W_VARIABLE'
         call Toker_Store
         st_col = p

         if p > L_len then do
               call ShowError 'Warning: incomplete variable at line' st_line
               call ShowError '  ==>' SrcIn.st_line, ''
               Tkr.0ExitCode = Tkr.0ExitCode 'BUG'
            end
         else do
               p = p +1
               st_val = MidStr(ThisLine, st_col, p - st_col)
               st_type = 'W_]'
               call Toker_Store
               st_col = p
            end
      end   /*  if c == [  */

   else do
         /*  a normal variable                                                                        */
         /*  why did I write the line about xranges instead of just verifying type?                   */
         /*        --because the xranges define the legal variable names in 4OS2.                     */
         /*          Added ? because it is used as internal var                                       */
         do forever
            if p > L_len then leave
            c = MidStr(ThisLine, p, 1)
            type = translate(c, Tkr.0typeset, Tkr.0charset)
            /*              if verify(type, 'sn') > 0 then leave                                      */
            if verify(c, xrange('a', 'z')||xrange('A', 'Z')||xrange('0', '9')||'_$?') > 0 then leave
            p = p +1
         end   /*  forever  */
         st_val = MidStr(ThisLine, st_col, p - st_col)
         st_type = 'W_VARIABLE'
         call Toker_Store
         st_col = p
         /*  fix: what to do about trailing %, as in "%aa%" and how to tell it from start of          */
         /*        next variable? For now, make it a separate token and call it "W_%2"                */
         c = MidStr(ThisLine, p, 1)
         if c == '%' then do
               p = p +1
               st_val = MidStr(ThisLine, st_col, p - st_col)
               st_type = 'W_%2'
               call Toker_Store
               st_col = p
            end

         /*  was that variable a function name? If so, do the part in brackets                        */
         c = MidStr(ThisLine, p, 1)
         if c == '[' then do
               p = p +1
               st_val = MidStr(ThisLine, st_col, p - st_col)
               st_type = 'W_['
               call Toker_Store
               /*  call Toker_4OS2 to do everything between '[' and ']'                               */
               inbrackets = inbrackets +1
               call Toker_4OS2
               inbrackets = inbrackets -1
            end

      end   /*  else  */
   return 0
   /* === End of Toker_do_Variable ================================================================== */

   Toker_do_BackQuote: /*fold00*/
   /* =============================================================================================== */
   /*  stores a string in backquotes as one W_LITERAL token                                           */
   /* =============================================================================================== */
   Esigl = sigl
   c = MidStr(ThisLine, p, 1)
   if c \= '`' then do
         call ShowError 'BUG: in Toker_do_BackQuote, starting position is not a backquote'
         call ShowError 'called from line' Esigl, ''
         call Mark 'Thisline c'
         call Egress 'bug'
      end
   p = p +1
   c = MidStr(ThisLine, p, 1)
   BackQ = 1
   do forever
      call LoopCounter
      c = MidStr(ThisLine, p, 1)
      p = p +1
      if c == '`' then do
            /*  the ending backquote                                                                  */
            st_val = MidStr(ThisLine, st_col, p - st_col)
            st_type = 'W_LITERAL'
            call Toker_Store
            st_col = p
            BackQ = 0
            return 0
            leave   /*  it is the ending quote  */
         end
      if p > L_len then do
            /*  EOL without closing backquote; store what we have                                     */
            st_val = MidStr(ThisLine, st_col, p - st_col)
            st_type = 'W_LITERAL'
            call Toker_Store
            st_col = p
            call Toker_do_EOL
            if \BackQ then
               return 0
            else
               iterate
         end
   end   /*  do forever  */
   /* === End of Toker_do_BackQuote ================================================================= */

   Toker_do_Quote: /*fold00*/
   /* =============================================================================================== */
   /*  to parse a string in double quotes                                                             */
   /* =============================================================================================== */
   Esigl = sigl
   if Opt.0Progress then call Progress

   st_col = p
   c = MidStr(ThisLine, p, 1)
   cc = MidStr(ThisLine, p, 2)
   p = p +1
   if c == '"' then do
         /*  this must be the opening quote:                                                          */
         DoubleQ = 1
         st_val = MidStr(ThisLine, st_col, p - st_col)
         st_type = 'W_"'
         st_class = 'QUOTED'
         call Toker_Store
         st_col = p
      end

   do forever
      call LoopCounter
      c = MidStr(ThisLine, p, 1)
      cc = MidStr(ThisLine, p, 2)
      p = p +1
      c2 = MidStr(ThisLine, p, 1)
      select
         when c2 == '"' then do
               /*  end of quoted string, store it                                                     */
               st_val = MidStr(ThisLine, st_col, p - st_col)
               st_type = 'W_SYMBOL'
               st_class = 'QUOTED'
               call Toker_Store
               st_col = p
            end   /*  when c2 == "  */

         when c == '"' then do
               /*  the closing quote:                                                                 */
               st_val = MidStr(ThisLine, st_col, p - st_col)
               st_type = 'W_"'
               st_class = 'QUOTED'
               call Toker_Store
               DoubleQ = 0
               st_class = '4OS2'
               st_col = p
               return 0
            end   /*  when c == "  */

         when c == '%' then do
               /*  begin variable or function; first store what we have, if anything                  */
               p = p -1
               if p > st_col then do
                     st_val = MidStr(ThisLine, st_col, p - st_col)
                     st_type = 'W_SYMBOL'
                     st_class = 'QUOTED'
                     call Toker_Store
                     st_col = p
                  end
               c2 = MidStr(ThisLine, p, 1)
               if c2 == '%' then
                  p = p +1
               st_val = MidStr(ThisLine, st_col, p - st_col)
               st_type = 'W_%'
               st_class = 'QUOTED'
               call Toker_Store
               st_col = p
               /*  Toker_do_variable will store the variable name or function name                    */
               call Toker_do_Variable
               st_col = p
            end   /*  when %  */

         when c2 == '`' then do
               /*  next character is start of backquoted string, store what we have                   */
               st_val = MidStr(ThisLine, st_col, p - st_col)
               st_type = 'W_SYMBOL'
               st_class = 'QUOTED'
               call Toker_Store
               st_col = p
               call Toker_do_BackQuote
               if \DoubleQ then do
                     st_class = '4OS2'
                     return 0
                  end
            end   /*  when c2 == `  */
         otherwise
            nop
      end   /*  select  */
      /*  EOL (end of line)                                                                           */
      if p > L_len then do
            /*  this could happen if the line does not have a closing quote; store what we have       */
            st_val = MidStr(ThisLine, st_col, p - st_col)
            st_type = 'W_SYMBOL'
            st_class = 'QUOTED'
            call Toker_Store
            st_col = p
            call Toker_do_EOL
            if \DoubleQ then do
                  st_class = '4OS2'
                  return 0
               end
            else
               iterate
         end
   end   /*  do forever  */
   return 0
   /* === End of Toker_do_Quote ===================================================================== */

   Toker_Store: /*fold00*/
   /* =============================================================================================== */
   /*  This is part of the tokeniser. It is  called by the Toker to                                   */
   /*  store a token in the arrays. It also identifies keywords and inserts hidden                    */
   /*  4OS2 End-Of-Clause tokens where they are implied. What I call a hidden EOC is a                */
   /*  token of t_type EOC and empty t_val. Having explicit separators between all                    */
   /*  clauses should make it easier for routines that deal with statements as opposed                */
   /*  to lines.                                                                                      */
   /* =============================================================================================== */
   Esigl = sigl
   /* ----------------------------------------------------------------------------------------------- */
   /*  Storing the EOCs here requires identifying keywords also.                                      */
   /*  h   == the token to be stored                                                                  */
   /*  hp  == the token just before h                                                                 */
   /*  hpp == the token just before hp                                                                */
   /* ----------------------------------------------------------------------------------------------- */
   h = t_type.0 +1
   hp = t_type.0
   hpp = hp -1

   if st_class == '4OS2' then do
         if hpp > 0 then do
               /*  THEN and ELSE usually end the line, but it is legal to put the next command        */
               /*  immediately after. In that case insert a hidden EOC token after THEN/ELSE          */
               if wordpos(translate(t_val.hp), 'THEN ELSE') > 0 then do
                     if \MatchCmd(t_val.hpp, '% SET DEFINED') then do
                           /*  might as well fix the type while we're at it:                          */
                           if translate(t_val.hp) == 'ELSE' then
                              t_type.hp = 'W_KEY'
                           if translate(t_val.hp) == 'THEN' then
                              t_type.hp = 'W_KEY2'
                           if st_type \== 'EOC' then do
                                 /*  not if "%then" or "%else"                                        */
                                 /*  not if "set THEN=..." or "if defined THEN"                       */
                                 call Toker_addEOC h
                                 h = h +1
                                 hp = h -1
                                 hpp = hp -1
                              end
                        end
                  end   /*  if THEN or ELSE  */

               /*  these indicate the beginning of a new statement                                    */
               /*  adding the EOC now is faster than inserting it into the array later                */
               select
                  when t_type.hp == 'W_SEP' ,
                     | t_type.hp == 'W_PIPE' then do
                        if st_type \= 'EOC' then do
                              call Toker_addEOC h
                              h = h +1
                              hp = h -1
                              hpp = hp -1
                           end
                     end

                  when translate(t_val.hpp) == 'ON' ,
                     & wordpos(translate(t_val.hp), 'BREAK ERROR ERRORMSG') > 0 ,
                     & st_type \= 'W_(' then do
                        if st_type \= 'EOC' then do
                              call Toker_addEOC h
                              h = h +1
                              hp = h -1
                              hpp = hp -1
                           end
                     end

                  /*  in "FOR %aa in (*.txt)", the parentheses are not a command group                */
                  /*  bugfix-- also exclude IF, as in "if (condition1 .or. condition2)"               */
                  /*  another bugfix-- exclude () in a rexx command                                   */
                  /*  a command group in an IF statement must be done later, in Toker_Fixup           */
                  when t_type.hp == 'W_(' then do
                        if translate(T_Infunction(h)) \= '@REXX' then
                           if translate(t_val.hpp) \= 'IN' then
                              if \MatchCmd(T_Val(R_FirstWord(hp)), 'IF') then do
                                    if st_type \= 'EOC' then do
                                          call Toker_addEOC h
                                          h = h +1
                                          hp = h -1
                                          hpp = hp -1
                                       end
                                 end
                     end

                  when t_type.hp == 'W_[' then do
                        funcN = translate(T_Infunction(h))
                        if funcN == '@EXEC' ,
                         | funcN == '@EXECSTR' then do
                              if st_type \= 'EOC' then
                                 if st_type \= 'W_(' then do
                                       call Toker_addEOC h
                                       h = h +1
                                       hp = h -1
                                       hpp = hp -1
                                    end
                           end
                     end
                  otherwise nop
               end   /*  select  */
            end   /*  if hpp > 0  */
      end   /*  if st_class == 4OS2  */
   /* ----------------------------------------------------------------------------------------------- */

   /*  store this token in all arrays:                                                                */
   t_class.h    = st_class        ; t_class.0 = h
   t_type.h     = st_type         ; t_type.0 = h
   t_line.h     = st_line         ; t_line.0 = h
   t_col.h      = st_col          ; t_col.0 = h
   t_prefix.h   = st_prefix       ; t_prefix.0 = h
   t_val.h      = st_val          ; t_val.0 = h
   /*  once the token is stored, we know its prefix is invalid. Toker manages the                     */
   /*  other st_xxx variables.                                                                        */
   st_prefix = ''
   if Opt.0Progress ,
    & h // 1000 = 0 then
      call Progress G.0meID' stored token '||h||' from line '||st_line' of 'SrcIn.0

   /* ----------------------------------------------------------------------------------------------- */
   /*  Tag any keywords. Sub-keywords must be done later, after all the tokens are stored             */

   /*  avoid error if there is no previous token                                                      */
   /*  also, the LastEOC might as well be initialized here as anywhere else                           */
   if h == 1 then do
         Tkr.0LastEOC = 1
         return
      end

   if t_class.h == '4OS2' then do
         select
            /*  if token h looks like a keyword, and is at the beginning of a clause,                 */
            /*  then it is a real keyword, unless I overlooked something.                             */
            when MatchCmd(t_val.h, Tkr.0Keywords) then do
                  /*              if t_type.hp == 'EOC' then do                                       */
                  if R_FirstWord(h) == h then do
                        t_type.h = 'W_KEY'
                     end   /*  when a keyword  */
               end   /*  when  */
            when h < 3 then nop
            otherwise nop
         end   /*  select  */

         funcN = T_Infunction(h)
         select
            /*  if inside a math function the character types are different                           */
            /*        when wordpos(translate(T_Infunction(h)), '@EVAL @INC @DEC') > 0 then do         */
            when wordpos(translate(funcN), '@EVAL @INC @DEC') > 0 then do
                  Tkr.0typeset = Tkr.0typeMath
                  Tkr.0mode = 'MATH'
               end
            when wordpos(translate(funcN), '@EXEC @EXECSTR') > 0 then do
                  Tkr.0typeset = Tkr.0typeSymbol
                  Tkr.0mode = '4OS2'
               end
            when funcN \= '' then do
                  Tkr.0typeset = Tkr.0typeFunction
                  Tkr.0mode = 'FUNCTION'
               end
            otherwise
               Tkr.0typeset = Tkr.0typeSymbol
               Tkr.0mode = '4OS2'
         end   /*  select  */

         if translate(t_val.h) == 'REM' ,
          | t_type.h == 'COLON2' then do
               /*  I have decided it will be more useful to consider a 'REM' as a                     */
               /*  part of the comment rather than as a keyword.                                      */
               if t_type.hp == 'EOC' ,
                | t_type.hp == 'LABEL' ,
                | wordpos(translate(t_val.hp), 'THEN ENDIFF ENDDO ENDSWITCH') > 0 then do
                     /*  tell Toker that the rest of this line is a comment                           */
                     t_class.h = 'COMMENT'
                     t_type.h = 'COMMENT_BEG'
                     Tkr.0mode = 'REM'
                  end
            end

         /*  identify a label:                                                                        */
         if t_type.hp == 'COLON' then
            t_type.h = 'LABEL'
      end   /*  if class == 4OS2  */

   /*  Fixups that involve inserting or deleting tokens are done at the end of each                   */
   /*  clause so they can look ahead at later tokens                                                  */
   if t_type.h == 'EOC' then do
         call Toker_Fixup
         Tkr.0LastEOC = t_type.0
      end
   return
   /* === End of Toker_Store ======================================================================== */

   Toker_Fixup: procedure expose (Shared_vars) (line_vars) (all_tokens) Tkr. /*fold00*/
   /* =============================================================================================== */
   /*  fixes some things that the main Toker routine does not or cannot do,                           */
   /* =============================================================================================== */
   Esigl = sigl
   if Opt.0Progress then call Progress G.0meID SubID()
   loop_ctr = 0
   t = Tkr.0LastEOC

   do while t < t_type.0
      t = t +1
      call LoopCounter
      if t_type.t == 'W_KEY' then do

            select
               /* ----------------------------------------------------------------------------------- */
               /*  an IF statement without THEN is tricky--                                           */
               /*        if (%a eq %b .or. %c eq %d) echo `yes`                                       */
               /*  Insert a hidden EOC where the THEN should be so the Toker can identify the         */
               /*  command easily.                                                                    */
               /*  fix: there is probably a simpler way to do this                                    */
               when translate(t_val.t) == 'IF' then do
                     /*  We need to identify the end of the condition being tested,                   */
                     /*  Start by finding the comparison word in the middle,  like "eq" or "lt".      */
                     /*  A comparison could also be a word like 'EXIST' or 'DEFINED'. There may be    */
                     /*  more than one comparison word, so work backwards to find the last one,       */
                     /*  then forward to find the spot where the command starts.                      */
                     comp = ''
                     w1 = R_FirstWord(t)
                     wLast = R_LastWord(t)
                     s = wLast
                     do while s > w1
                        select
                           when t_val.s == ']' then
                              s = R_MatchingBracket(s)
                           when t_val.s == '"' then
                              do until T_Class(s -1) \= 'QUOTED'
                                 s = s -1
                              end
                           when wordpos(translate(t_val.s), translate(G.0Conditions)) > 0 then do
                                 comp = s
                                 leave
                              end
                           otherwise
                              if wordpos(translate(t_val.s), translate(G.0OPCompare)) > 0 then do
                                    /*  found the comparison operator                                 */
                                    comp = s
                                    leave
                                 end
                        end   /*  select  */
                        s = s -1
                     end   /*  while s  */

                     eoc = ''
                     if comp \== '' then do
                           /*  skip the expression after the comparison op                            */
                           s = comp
                           do while s <= wLast
                              s = s +1
                              select
                                 when t_val.s == '"' then do
                                       do until T_Class(s +1) \= 'QUOTED'
                                          s = s +1
                                       end
                                    end
                                 when t_val.s == '[' then
                                    s = R_MatchingBracket(s)
                                 when t_val.s == ')' then
                                    /*  ignore closing parenthesis, as in "if (%a eq %b .or. %c eq %d)" */
                                    nop
                                 otherwise nop
                              end   /*  select  */
                              if T_Type(s +1) == 'W_SYMBOL' then do
                                    if T_Prefix(s +1) \== '' then do
                                          /*  found the next space-separated "word"                   */
                                          eoc = s +1
                                          leave
                                       end
                                    else
                                       if t_type.s == 'W_(' then do
                                             eoc = s +1
                                             leave
                                          end
                                 end   /*  if SYMBOL  */
                           end   /*  while s < WLast  */
                        end   /*  if comp == ''  */
                     else do
                           call ShowError 'Tokeniser warning: error parsing IF statement at line' t_line.t
                           call ShowError Qline(t_line.t), ''
                           Tkr.0ExitCode = Tkr.0ExitCode 'FILE'
                           iterate
                        end

                     if eoc \== '' then do
                           /*  eoc is beginning of the command, insert a hidden EOC                   */
                           /*  also check for THEN, which would mean either the IF is a               */
                           /*  mistake for IFF, or the writer mis-typed it as a                       */
                           /*  Rexx-style IF...THEN <command> statement. Either way, we don't         */
                           /*  want to compound the mistake by tokenising it wrong.                   */

                           if T_Type(eoc -1) \= 'EOC' ,
                            & t_type.eoc \= 'EOC' ,
                            & translate(t_val.eoc) \= 'THEN' then do
                                 call Toker_addEOC eoc
                              end
                        end
                     else do
                           call ShowError 'could not find command in IF statement at line' t_line.t
                           call Mark 't comp eoc'
                           call ShowError Qline(t_line.t), ''
                           call ShowError Qline(t_line.comp), ''
                           Tkr.0ExitCode = Tkr.0ExitCode 'FILE'
                        end
                  end   /*  when IF  */
               /* ----------------------------------------------------------------------------------- */
               /*  FOR %%var in (list) <command>                                                      */
               when translate(t_val.t) == 'FOR' then do
                     /*  like IF, we need to identify the place where the command(s) starts           */
                     /*  which is after the closing ")"                                               */
                     eoc = T_FindToken(t, R_LastWord(t), , 'W_)') +1
                     if eoc == -1 then do
                           call ShowError 'Tokeniser warning: error parsing FOR statement at line' t_line.t
                           call ShowError Qline(t_line.t), ''
                           Tkr.0ExitCode = Tkr.0ExitCode 'FILE'
                        end
                     else do
                           if translate(t_val.eoc) == 'DO' then eoc = eoc +1
                           if t_type.eoc == 'W_(' then eoc = eoc +1
                           /*  eoc is beginning of the command, insert a hidden EOC                   */
                           if T_Type(eoc -1) \= 'EOC' ,
                            & t_type.eoc \= 'EOC' then do
                                 call Toker_addEOC eoc
                              end
                        end
                  end   /*  when FOR  */
               /* ----------------------------------------------------------------------------------- */
               /*  START <command>                                                                    */
               when translate(t_val.t) == 'START' then do
                     title = ''
                     Command = ''
                     s = t
                     do while s <= R_NextEOC(t)
                        s = s +1
                        select
                           when t_type.s == 'W_LITERAL' then do
                                 if t_class.s == 'QUOTED' then do
                                       if title == '' then
                                          title = s
                                       else do
                                             Command = s
                                             leave
                                          end
                                    end
                              end
                           when t_val.s == '/' then do
                                 s = s +1
                                 if translate(t_val.s) == 'PGM' then do
                                       s = s +1
                                       if t_type.s == 'W_"' then s = s +1
                                       Command = s
                                       leave
                                    end
                              end
                           when t_type.s == 'EOC' then do
                                 /*  no need to check for quotes, it would have been caught above     */
                                 Command = s -1
                                 leave
                              end

                           otherwise nop
                        end   /*  select  */
                     end   /*  while s  */

                     if Command \= '' then
                        t_type.Command = 'W_CMD'
                     else do
                           call ShowError 'Tokeniser warning: could not find command in line' t_line.t
                           call ShowError Qline(t_line.t), ''
                           Tkr.0ExitCode = Tkr.0ExitCode 'FILE'
                        end
                  end   /*  if START  */

               /* ----------------------------------------------------------------------------------- */
               /*  DETACH <command>                                                                   */
               when translate(t_val.t) == 'DETACH' then do
                     Command = ''
                     s = t
                     do while s <= R_NextEOC(t)
                        s = s +1
                        select
                           when t_type.s == 'W_LITERAL' then do
                                 if t_class.s == 'QUOTED' then do
                                       Command = s
                                       leave
                                    end
                              end
                           when t_type.s == 'EOC' then do
                                 /*  no need to check for quotes, it would have been caught above     */
                                 Command = s -1
                                 leave
                              end

                           otherwise nop
                        end   /*  select  */
                     end   /*  while s  */

                     if Command \= '' then
                        t_type.Command = 'W_CMD'
                     else do
                           call ShowError 'Tokeniser warning: could not find command in line' t_line.t
                           call ShowError Qline(t_line.t), ''
                           Tkr.0ExitCode = Tkr.0ExitCode 'FILE'
                        end
                  end   /*  if DETACH  */
               /* ----------------------------------------------------------------------------------- */
               /*     CALL <COMMAND>                                                                  */
               when translate(t_val.t) == 'CALL' then
                  if abbrev(T_Type(t +1), 'W_') then
                     /*  exclude "call %filename"                                                     */
                     if T_Val(t +1) == R_Expression(t +1) then
                        call T_Type t +1, 'W_CMD'
               /* ----------------------------------------------------------------------------------- */
               /*     ALIAS or SET                                                                    */
               when MatchCmd(t_val.t, 'SET ALIAS') then
                  do s = t +1 to R_NextEOC(t)
                     if t_val.s == '=' then leave s
                     if t_type.s \= 'W_SYMBOL' then iterate s
                     if translate(t_val.s) == '/R' then leave s
                     if left(t_val.s, 1) == '/' then iterate s
                     if translate(t_val.t) == 'ALIAS' then
                        t_type.s = 'W_CMD'
                     else
                        t_type.s = 'W_VARIABLE'
                  end s
               /* ----------------------------------------------------------------------------------- */
               /*  UNALIAS or UNSET                                                                   */
               when MatchCmd(t_val.t, 'UNSET UNALIAS') then
                  do s = t +1 to R_NextEOC(t)
                     if t_type.s \= 'W_SYMBOL' then iterate s
                     if translate(t_val.s) == '/R' then leave
                     if left(t_val.s, 1) == '/' then iterate s
                     if translate(t_val.t) == 'ALIAS' then
                        t_type.s = 'W_CMD'
                     else
                        t_type.s = 'W_VARIABLE'
                  end s
               /* ----------------------------------------------------------------------------------- */
               /*  TEXT                                                                               */
               when translate(t_val.t) == 'TEXT' then do
                     Tkr.0typeset = Tkr.0typeText
                     Tkr.0mode = 'TEXT'
                  end
               otherwise
                  nop
            end   /*  select  */
         end   /*  if W_KEY  */
      /* -------------------------------------------------------------------------------------------- */

      /* -------------------------------------------------------------------------------------------- */
      if t_type.t == 'W_SYMBOL' then do
            /*  PROCEDURE                                                                             */
            /*  identify subroutine names and give them a special type:                               */
            if Toker_AtGosub(t) then t_type.t = 'W_PROCEDURE'
            /* -------------------------------------------------------------------------------------- */
            /*  LOGICAL and COMPARISON words                                                          */
            if MatchCmd(T_Val(R_PrevKeyword(t)), 'IF IFF DO') then do
                  if wordpos(translate(t_val.t), translate(G.0OPLogical)) > 0 then
                     t_type.t = 'W_LOGIC'
                  if wordpos(translate(t_val.t), translate(G.0OPCompare)) > 0 then
                     t_type.t = 'W_COMPARE'
               end   /*  if IF or IFF  */
         end   /*  if W_SYMBOL  */
      /* -------------------------------------------------------------------------------------------- */
      /*  FUNCTION                                                                                    */
      if T_Type(t -1) == 'W_%' then do
            if t_prefix.t == '' then
               if left(t_val.t, 1) == '@' then
                  t_type.t = 'W_FUNCTION'
         end

      /* -------------------------------------------------------------------------------------------- */

      /* -------------------------------------------------------------------------------------------- */
      /*  ALIAS, KEYWORD, or external COMMAND                                                         */
      /*  this is to identify commands that were missed before. Usually because                       */
      /*  of an IF statement, where the EOC could not be identified in Toker_Store.                   */
      /*  This must be done after the section above that inserts the EOC into an IF                   */
      /*  statement.                                                                                  */
      if t_type.t == 'W_SYMBOL' then do
            if R_FirstWord(t) == t then do
                  /*  in "if a eq b *del file.txt", the main Toker loop cannot                        */
                  /*  tell that "*" is the beginning of a command, because the                        */
                  /*  EOC had not been inserted yet. Split the token.                                 */
                  /*  fix: It should be possible to do that in the main loop, but                     */
                  /*  maybe not worth the trouble                                                     */
                  if left(t_val.t, 1) == '@' then do
                        /*  "@echo" as in "if a eq b @echo %b"                                        */
                        Command = substr(t_val.t, 2)
                        call Toker_InsertToken t, 1
                        t_class.t = '4OS2'
                        t_type.t = '@'
                        t_prefix.t = T_Prefix(t +1)
                        t_val.t = '@'
                        t_line.t = T_Line(t +1)
                        t_col.t = T_Col(t +1)
                        t = t +1
                        t_prefix.t = ''
                        t_col.t = t_col.t +1
                        t_val.t = substr(t_val.t, 2)
                     end
                  if left(t_val.t, 1) == '*' then do
                        call Toker_InsertToken t, 1
                        t_class.t = '4OS2'
                        t_type.t = '*'
                        t_prefix.t = T_Prefix(t +1)
                        t_val.t = '*'
                        t_line.t = T_Line(t +1)
                        t_col.t = T_Col(t +1)
                        t = t +1
                        t_prefix.t = ''
                        t_col.t = t_col.t +1
                        t_val.t = substr(t_val.t, 2)
                     end
                  if MatchCmd(t_val.t, Tkr.0Keywords) then
                     t_type.t = 'W_KEY'
                  else
                     t_type.t = 'W_CMD'
               end   /*  if first word  */
         end   /*  if W_SYMBOL  */
      /* -------------------------------------------------------------------------------------------- */
   end   /*  while t < t_type.0  */

   if Opt.0Progress then call Progress
   return 0
   /* === End of Toker_Fixup ======================================================================== */

   Toker_KeyWords: procedure expose (Shared_vars) (all_tokens) Tkr. /*fold00*/
   /* =============================================================================================== */
   /*  this is a part of Toker_Fixup that identifies sub-keywords                                     */
   /* ----------------------------------------------------------------------------------------------- */
   /*  SUB-KEYWORD                                                                                    */
   /*                                                                                                 */
   /*  the loop counter is always on here, to protect against defective code that                     */
   /*  can cause it to loop endlessly                                                                 */
   /*                                                                                                 */
   /* =============================================================================================== */

   if Opt.0Progress then call Progress G.0meID SubID()

   t = 0
   do while t < t_type.0
      t = t +1
      call LoopCounter
      if t_type.t \== 'W_KEY' then iterate

      Tval = translate(t_val.t)
      i = t

      select   /*  value of t_val.i  */
         when Tval == 'WINDOW' then
            if wordpos(translate(T_Val(t +1)), 'MIN MAX RESTORE') > 0 then
               call T_Type t +1, 'W_KEY2'

         when Tval == 'ACTIVATE' then
            if wordpos(translate(T_Val(t +4)), 'MIN MAX RESTORE CLOSE') > 0 then
               call T_Type t +4, 'W_KEY2'

         when wordpos(Tval, 'ECHO LOADBTM LOG VERIFY KEYS') > 0 then
            if wordpos(translate(T_Val(t +1)), 'ON OFF LIST') > 0 then
               call T_Type t +1, 'W_KEY2'

         when Tval == 'IF' ,
            | Tval == 'IFF' then do
               do i = t to R_LastWord(t)
                  if wordpos(translate(t_val.i), translate(G.0Conditions)) > 0 then do
                        t_type.i = 'W_KEY2'
                        if translate(t_val.i) == 'DEFINED' then
                           /*  note R_NextWord(i), not i +1, because this is legal:                   */
                           /*        set aa=`bb`                                                      */
                           /*        set %aa=`value of variable bb`                                   */
                           call T_Type R_NextWord(i), 'W_VARIABLE'
                     end
               end i
               if Tval == 'IFF' then do
                     then = R_LastWord(t)
                     if translate(t_val.then) == 'THEN' then
                        t_type.then = 'W_KEY2'
                  end
            end

         when Tval == 'FOR' then do
               /*     find first "(";                                                                 */
               /*     word before it is IN;                                                           */
               /*     word before IN is variable name                                                 */
               /*     word after ")" is either command, "(", or "DO"                                  */
               do j = i to t_type.0
                  if t_val.j == '(' then leave
               end
               if translate(T_Val(j -1)) == 'IN' then do
                     call T_Type j -1, 'W_KEY2'
                     call T_Type j -2, 'W_VARIABLE'
                     do k = j to t_type.0
                        if t_val.k == ')' then leave
                     end
                     k = k +1
                     /*  token l must be a single command, a DO, or a ( that begins a command group   */
                     /*  optional DO here is not to be confused with the keyword DO                   */
                     if translate(t_val.k) == 'DO' then do
                           t_type.k = 'W_KEY2'
                           k = k +1
                        end

                     i = k
                  end   /*  if "IN"  */
               else do
                     call ShowError 'error parsing FOR statement' T_Val(j -1) 'should be IN'
                     Tkr.0ExitCode = Tkr.0ExitCode 'FILE'
                  end
            end   /*  when FOR  */

         when Tval == 'INKEY' then do
               /*  no sub-keywords                                                                    */
               nop
            end

         when Tval == 'INPUT' then do
               /*  no sub-keywords                                                                    */
               nop
            end

         when Tval == 'ON' then do
               if wordpos(translate(T_Val(t +1)), 'BREAK ERROR ERRORMSG') > 0 then
                  call T_Type t +1, 'W_KEY2'
            end

         when Tval == 'PROMPT' then do
               /*  no sub-keywords                                                                    */
               nop
            end

         when Tval == 'DO' then do
               select
                  when T_Val(t +2) == '=' then do
                        /*  "DO var = 1 to 5"                                                         */
                        call T_Type t +1, 'W_VARIABLE'
                        do s = t +2 to R_LastWord(t +2)
                           if translate(t_val.s) == 'TO' then
                              t_type.s = 'W_KEY2'
                        end s
                     end
                  when translate(T_Val(t +2)) == 'IN' then do
                        /*  "DO var IN (`a b c`)"                                                     */
                        call T_Type t +1, 'W_VARIABLE'
                        call T_Type t +2, 'W_KEY2'
                     end
                  when wordpos(translate(T_Val(t +1)), 'WHILE UNTIL FOREVER') > 0 then
                     /*  "DO FOREVER"                                                                 */
                     call T_Type t +1, 'W_KEY2'
                  otherwise
                     /*  "DO 5"                                                                       */
                     nop
               end   /*  select  */
            end   /*  when DO  */

         when wordpos(Tval, 'CLS COLOR DRAWHLINE DRAWVLINE SCRPUT VSCRPUT') > 0 then
            do i = t to R_LastWord(t)
               if t_type.i == 'W_SYMBOL' ,
                & wordpos(translate(t_val.i), 'BRIGHT BLINK ON BORDER') > 0 then
                  t_type.i = 'W_KEY2'
            end i

         when Tval == 'DRAWBOX' then
            do i = t to R_LastWord(t)
               if t_type.i == 'W_SYMBOL' ,
                & wordpos(translate(t_val.i), 'BRIGHT BLINK ON FILL ZOOM') > 0 then
                  t_type.i = 'W_KEY2'
            end i

         otherwise
            call Toker_do_expression

      end   /*  select t_val.t  */
      t = i

   end   /*  while t < t_type.0  */
   if Opt.0Progress then call Progress
   return 0
   /* === End of Toker_KeyWords ===================================================================== */

   Toker_do_expression: procedure expose (Shared_vars) (all_tokens) Tkr. i /*fold00*/
   /* =============================================================================================== */
   /*  this is part of Toker_Fixup/Toker_KeyWords                                                     */
   /* =============================================================================================== */
   do until t_type.i = 'EOC'
      i = i +1
      if Toker_AtGosub(i) then t_type.i = 'W_PROCEDURE'
   end
   return
   /* === End of Toker_do_expression ================================================================ */

   Toker_AtGosub: procedure expose (Shared_vars) (all_tokens) Tkr. /*fold00*/
   /* =============================================================================================== */
   /*  returns True if in the form "GOSUB xxx"                                                        */
   /*  this is part of Toker_Fixup/Toker_KeyWords                                                     */
   /* =============================================================================================== */
   arg t
   if t_type.t == 'W_SYMBOL' ,
    | t_type.t == 'W_CONSTANT' then
      if MatchCmd(T_Val(t -1), 'GOSUB GOTO') then
         return 1
   return 0
   /* === End of Toker_AtGosub ====================================================================== */

   Toker_InsertToken: procedure expose (Shared_vars) (all_tokens) Tkr. /*fold00*/
   /* =============================================================================================== */
   /*  This is similar to T_InsertToken but for the Toker.                                            */
   /*  It uses a different list of token attributes                                                   */
   /*  this inserts new element(s) in each of the arrays listed in Tkr.0token_vars. If I add a new    */
   /*  token array or change the name of one, nothing here needs to be rewritten                      */
   /* ----------------------------------------------------------------------------------------------- */
   Esigl = sigl
   t = arg(1)
   HowMany = arg(2)
   if HowMany == '' then HowMany = 1
   if \datatype(HowMany, 'w') then do
         call ShowError 'In Toker_InsertToken, invalid argument(s): "'t'"  "'HowMany'"'
         call Egress 'BUG_Function'
      end

   Tlist = Tkr.0token_vars
   do while Tlist \= ''
      parse var Tlist attrname Tlist
      !vName = value('AttrName')
      call ArrayInsert !vName, t, HowMany
   end   /*  of tlist  */
   return 0
   /* === End of Toker_InsertToken ================================================================== */

   Toker_addEOC: procedure expose (Shared_vars) (all_tokens) Tkr. /*fold00*/
   /*=================================================================================================*/
   /*  Usage:                                                                                         */
   /*        call Toker_addEOC t                                                                      */
   /*                                                                                                 */
   /*  where t is the insert position                                                                 */
   /*  a hidden EOC token will be inserted before token t                                             */
   /*                                                                                                 */
   /*=================================================================================================*/
   Esigl = sigl
   arg h
   select
      when h == t_type.0 +1 then do
            /*  this is easy, just add a new element at the end of each array                         */
            t_class.0 = h
            t_type.0 = h
            t_line.0 = h
            t_col.0 = h
            t_prefix.0 = h
            t_val.0 = h
         end
      when h <= t_type.0 then
         call Toker_InsertToken h
      otherwise
         call ShowError 'BUG: invalid argument to Toker_addEOC at line' Esigl
         call ShowError '      'sourceline(Esigl)
         call Egress 'bug'
   end   /*  select  */

   /*  either way the token was added, set it to a hidden EOC                                         */
   p = h -1
   t_class.h = '4OS2'
   t_type.h = 'EOC'
   t_line.h = t_line.p
   t_col.h = t_col.p + length(t_val.p)
   t_prefix.h = ''
   t_val.h = ''
   return 0
   /* === End of Toker_addEOC ======================================================================= */

   SaveTokensT: procedure expose (Shared_vars) (line_vars) (all_tokens) Tkr. /*fold00*/
   /* =============================================================================================== */
   /*  send a lot of information about the tokens to a file                                           */
   /* =============================================================================================== */
   if Opt.0Progress then call Progress G.0meID SubID()
   parse arg TokenFileN
   if TokenFileN == '' then do
         call ShowError 'BUG: in SaveTokensT, no file name given.  Called from line' Esigl
         call ShowError '      'sourceline(Esigl)
         call Egress 'BUG'
      end

   if FileExist(TokenFileN) then do
         suff = 2
         do while FileExist(TokenFileN||'_'||suff)
            suff = suff +1
         end
         NewN = TokenFileN||'_'||suff
         call Msg G.0meID 'token file' TokenFileN 'changed to' NewN
         TokenFileN = NewN
      end

   /*   header lines and tab positions                                                                */
   tab_Source = 22
   tab_sigline = 7
   tab_Token     = tab_sigline +9
   tab_line      = tab_Token   +10
   tab_column    = tab_line    +10
   tab_class     = tab_column  +8
   tab_type      = tab_class   +12
   tab_val       = tab_type    +22
   Header_1 = left(' ', tab_Source)'000000000'
   Header_2 = left(' ', tab_Source)'123456789'
   do i = 1 to 11
      c = i // 10
      Header_1 = Header_1||copies(c, 10)
      Header_2 = Header_2||' 123456789'
   end
   Header_1 = overlay(' Column # -->', Header_1, 1)
   Header_3 = ' '
   Header_3 = overlay('Token #', Header_3, tab_Token)
   Header_3 = overlay('Line #', Header_3, tab_line)
   Header_3 = overlay('Col #', Header_3, tab_column)
   Header_3 = overlay('Class', Header_3, tab_class)
   Header_3 = overlay('Type', Header_3, tab_type)
   Header_3 = overlay('{Prefix}{Value}', Header_3, tab_val)

   s_printed = 0
   do t = 1 to t_type.0
      Header_3 = overlay('Toker line:', Header_3, 1)
      Outline = copies(' ', 140)
      Outline = overlay(format(t, 7), Outline, tab_Token)
      Outline = overlay(format(t_line.t, 6), Outline, tab_line)
      Outline = overlay(format(t_col.t, 5), Outline, tab_column)
      Outline = overlay(t_class.t, Outline, tab_class)
      Outline = overlay(t_type.t, Outline, tab_type)
      Outline = overlay("{"t_prefix.t"}{"t_val.t"}", Outline, tab_val)
      st_ln = t_line.t
      if st_ln > s_printed then do
            s_printed = st_ln
            srcline = 'source line # 'right(st_ln, 5, '0')
            srcline = overlay('{'SrcIn.st_ln'}', srcline, tab_Source)
            call lineout TokenFileN, ''
            call lineout TokenFileN, Header_1
            call lineout TokenFileN, Header_2
            call lineout TokenFileN, srcline
            call lineout TokenFileN, Header_3
         end
      Outline = strip(Outline, 't')
      call lineout TokenFileN, Outline
   end t
   call lineout TokenFileN
   if Opt.0Progress then call Progress
   say G.0meID 'Token list was written to 'TokenFileN
   return 0
   /* === End of SaveTokensT ======================================================================== */

   LoopCounter: procedure expose (Shared_vars) loop_max loop_ctr /*fold00*/
   /* =============================================================================================== */
   /*  This is a debugging tool to prevent the formatter from being caught in                         */
   /*  an endless loop. To use it, just add "CALL LoopCounter" at the start of a loop.                */
   /*  If there have been more than loop_max loops, this will break out of the loop                   */
   /*  and show the values of any variables listed in G.0track_vars.                                  */
   /*                                                                                                 */
   /*  the calling routine does not need to set any variables, but it can set                         */
   /*  loop_ctr or loop_max to whatever is convenient. Make sure loop_max is                          */
   /*  high enough to allow all the legitimate passes through the loop.                               */
   /*  later: that should be done automatically now-- the default is based                            */
   /*  on the file size.                                                                              */
   /*                                                                                                 */
   /*  note that loop_max and loop_ctr are not global, they are local to                              */
   /*  the calling routine                                                                            */
   /* =============================================================================================== */
   Esigl = sigl
   if symbol('loop_max') == 'LIT' then loop_max = max(50000, stream(G.0InfileN, 'c', 'query size'))
   if symbol('loop_ctr') == 'LIT' then loop_ctr = 1
   else loop_ctr = loop_ctr +1

   if loop_ctr >= loop_max then do
         call ShowError 'Over maximum loop limit ('loop_max') in' G.0me 'at line 'Esigl
         call ShowError 'probably because of an endless loop in the' LastLabel(Esigl) 'procedure', ''
         call Mark G.0track_vars, ''
         call Egress 'BUG_LoopCounter'
      end
   return 0
   /* === End of LoopCounter ======================================================================== */

   /***************************************************************************************************/
   /*  External functions section                                                                     */
   /*  akm: copied here from my macro library                                                         */
   /***************************************************************************************************/

   FileExist: procedure /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        if FileExist(<filename>) then                                                            */
   /*                                                                                                 */
   /*  returns 1 (True) if <filename> exists and 0 (False) if it does not.                            */
   /*                                                                                                 */
   /*  <filename> must be a real file. This returns 0 for devices or for streams like STDOUT.         */
   /*  Granted, STDOUT does exist, but it more useful this way. I mostly use this when I want         */
   /*  to avoid clobbering an existing file,                                                          */
   /*                                                                                                 */
   /*  The only reason for this is that                                                               */
   /*        if FileExist(fileN) then                                                                 */
   /*  is easier to remember than                                                                     */
   /*         if stream(fileN, 'c', 'query exists') then                                              */
   /* =============================================================================================== */
   /*  Most of the complications are because OS/2 Rexx and Regina return different values             */
   /*  from stream() for stdout and for devices. Regina usually returns 0 for stdout, but             */
   /*  not always. I don't know why.                                                                  */
   parse arg fileN
   if fileN = '' then return 0
   if wordpos(translate(fileN), 'STDOUT STDERR') > 0 then return 0
   ret = stream(fileN, 'c', 'query exists')
   if ret == '' then return 0
   if left(translate(ret), 5) == '\DEV\' then return 0
   if left(translate(ret), 5) == '/DEV/' then return 0
   return 1
   /* === End of FileExist ========================================================================== */

   FileExt: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = FileExt(<filename>)                                                             */
   /*                                                                                                 */
   /*    returns the extension of a file name, like 4os2 "%@ext[]                                     */
   /* =============================================================================================== */
   parse arg Fname
   ret = ''
   Fname = filespec('N', Fname)
   dot = lastpos('.', Fname)
   if dot > 0 then ret = substr(Fname, dot +1)
   return ret
   /* === End of FileExt ============================================================================ */

   GetOpts: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     INTERPRET GetOpts(args [,valid])                                                            */
   /*                                                                                                 */
   /*  where 'args' is the command line to be parsed, and                                             */
   /*  'valid' is an optional space-separated list of valid switches                                  */
   /* ----------------------------------------------------------------------------------------------- */
   /*  switches must start with '-' or '/', for example:                                              */
   /*      /help                                                                                      */
   /*      -verbose                                                                                   */
   /*      -debug=3                                                                                   */
   /*      -filespec="*.txt *.html"                                                                   */
   /*                                                                                                 */
   /*  switches must not have spaces before or after the '='                                          */
   /*  a switch without a value is set to '1'                                                         */
   /*                                                                                                 */
   /*                                                                                                 */
   /*  Interpreting what GetOpts returns will create several variables that the                       */
   /*  caller can use--                                                                               */
   /*     Opt.              is a stem variable that holds whatever was specified by switches          */
   /*     Opt.0optlist      a space-separated list of options that were read                          */
   /*     Opt.0errors       number of errors; that is, when GetOpts() did not understand something    */
   /*                                                                                                 */
   /*                       elements of the Opt. compound variable all start with '0', to avoid       */
   /*                       name conflicts with simple variables                                      */
   /*                                                                                                 */
   /*     ArgV.             is an array that holds the non-switch arguments                           */
   /*                       (ArgV.0 is the name of the .cmd script, NOT the number of items           */
   /*                       in the array)                                                             */
   /*     ArgC              is the number of non-switch arguments                                     */
   /*                                                                                                 */
   /*  Example:                                                                                       */
   /*     <command> -dir=up -msg="that way" -v "From here"                                            */
   /*     will result in these variables:                                                             */
   /*        Opt.0optlist == opt.0dir opt.0msg                                                        */
   /*        Opt.0dir     == up                                                                       */
   /*        Opt.0msg     == that way                                                                 */
   /*        Opt.0v       == 1                                                                        */
   /*        ArgV.0       == <name of cmd file>                                                       */
   /*        ArgV.1       == From here                                                                */
   /*        ArgC         == 1                                                                        */
   /*                                                                                                 */
   /*     <command> -help                                                                             */
   /*     <command> -h                                                                                */
   /*     <command> -?                                                                                */
   /*     will all result in:                                                                         */
   /*        Opt.0help    == 1                                                                        */
   /*                                                                                                 */
   /*  in the case of errors, GetOpts() will cause the caller to show a message                       */
   /*  about it on stderr                                                                             */
   /* =============================================================================================== */
   /* Some important variables (inside GetOpts) are:                                                  */
   /*    Str.SetOpts       == commands the the caller will execute to set options                     */
   /*                          in the Opt. stem variable                                              */
   /*    Str.SetArgs       == commands the the caller will execute to put the arguments               */
   /*                          in the ArgV. and ArgC variables                                        */
   /*    Str.Errors        == commands the caller will execute to report errors                       */
   /*                          reading the command line                                               */
   /*    ReturnStr         == all of the above, combined and returned to caller                       */
   /*    flag.             == variables for GetOpt's internal use                                     */
   /*                          flag.error == number of errors trying to read command line             */
   /*    Str.SwitchList    == list of options (switches) used on command line                         */
   /*    ArgC              == number of non-switch arguments                                          */
   /*    ArgV.             == array of non-switch arguments                                           */
   /*    valid             == list of valid switches                                                  */
   /*                                                                                                 */
   /* =============================================================================================== */
   flag. = ''
   flag.errors = 0
   str. = ''
   ReturnStr = ''
   ArgC = 0
   ArgV. = ''
   parse arg args , valid
   numwords = WordsQ(args)
   do i = 1 to numwords
      thisArg = WordQ(args, i)
      /*  allow '/' as a switch character by changing it to '-'                                       */
      if left(thisArg, 1) == '/' then
         thisArg = '-'||strip(thisArg, 'L', '/')

      if left(thisArg, 1) == '-' then
         call ReadSwitch thisArg
      else do
            /*    not a switch, add to Arg list                                                       */
            ArgC = ArgC +1
            ArgV.ArgC = thisArg
         end
   end

   /* ----------------------------------------------------------------------------------------------- */
   /*    put it all together into a series of commands for the caller to interpret                    */
   if flag.errors > 0 then do
         str.SwitchList = strip(str.SwitchList' Opt.0Errors')
      end
   str.SetOpts = str.SetOpts' Opt.0Errors = 'flag.errors ';'
   str.SetOpts = str.SetOpts 'Opt.0optlist = "'str.SwitchList'" ;'

   str.SetArgs = 'ArgV. = "" ; parse source . . ArgV.0 ;  ArgC = "'ArgC'" ;'
   do i = 1 to ArgC
      varVal = strip(ArgV.i, , '"')
      varVal = strip(varVal, , "'")
      str.SetArgs = str.SetArgs' ArgV.'i' = 'Quoted(varVal)' ;'
   end

   ReturnStr = str.errors str.SetOpts str.SetArgs
   return ReturnStr
   /* === End of GetOpts ============================================================================ */

   ReadSwitch: procedure expose (Shared_vars) flag. str. valid /*fold00*/
   /* =============================================================================================== */
   /*  this is part of GetOpts()                                                                      */
   /* =============================================================================================== */
   parse arg thisArg
   switch = strip(thisArg, 'L', '-')
   optName = strip(WordQ(switch, 1, '='))
   select
      /* a few switches have fixed meanings                                                           */
      when translate(optName) == '?' then
         optName  = 'Opt.0Help'
      when translate(optName) == 'H' then
         optName  = 'Opt.0Help'
      when translate(optName) == 'HELP' then
         optName  = 'Opt.0Help'
      when valid \= '' ,
         & pos(' '||translate(optName)||' ', ' '||translate(valid)||' ') == 0 then do
            str.errors = ,
              str.errors||'call lineout "stderr", "Error: GetOpts does not understand switch `'optName'`" ;'
            flag.errors = flag.errors +1
         end
      otherwise
         optName = 'Opt.0'optName
   end   /*  select  */

   optVal = strip(WordQ(switch, 2, '='))
   /*  "-switch=" means Opt.0switch == '', not 1                                                      */
   if optVal == '' ,
    & right(switch, 1) \= '=' then
      optVal = '1'
   else do
         /*    try to be tolerant of unusual arguments; only strip quotes if they are balanced        */
         /*    and only one level of nested quotes                                                    */
         optVal = Unquoted(optVal)
      end

   /*    there should be no third word                                                                */
   if WordQ(switch, 3, '=') \== '' then do
         str.errors = ,
           str.errors||'call lineout "stderr", "Error: getopts does not understand `'thisArg'`" ;'
         flag.errors = flag.errors +1
      end

   if wordpos(optName, str.SetOpts) > 0 then do
         str.errors = ,
           str.errors||'call lineout "stderr", "Error: duplicate switch 'thisArg'" ;'
         flag.errors = flag.errors +1
      end
   else do
         str.SetOpts = str.SetOpts' 'optName' = 'Quoted(optVal)' ;'
         str.SwitchList = strip(str.SwitchList' 'optName)
      end
   return
   /* === End of ReadSwitch ========================================================================= */

   LeftStr: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        ret = LeftStr(string , length)                                                           */
   /*                                                                                                 */
   /*  a better Left() function that does not pad                                                     */
   /* =============================================================================================== */
   parse arg string , len , .

   if len > length(string) then
      len = (length(string))
   if len > 0 then
      return left(string, len)
   else
      return ''
   /* === End of LeftStr ============================================================================ */

   MidStr: procedure expose  expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = MidStr(<string> [, <start>, <length>])                                          */
   /*        string2 = MidStr(string1, 5)                                                             */
   /*        string2 = MidStr(string1, 5, 3)                                                          */
   /*                                                                                                 */
   /*  like SUBSTR(), it returns a substring from a string                                            */
   /*  unlike SUBSTR(),                                                                               */
   /*      it never pads the substring; you get only what you ask for                                 */
   /*      If <start> is past the end of the string, you get a null string.                           */
   /*      If <start> + <length> is longer than the string, you get the substring from                */
   /*         <start> to the end of the string, with no padding                                       */
   /*      If the string is '', you get ''                                                            */
   /* =============================================================================================== */
   parse arg instring , start , len , .
   if len == '' then
      return substr(instring, start)
   if start > length(instring) then
      return ''
   if start -1 + len > length(instring) then
      len = (length(instring) - start +1)
   return substr(instring, start, len)
   /* === End of MidStr ============================================================================= */

   RightStr: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*  ret = RightStr(string , length)                                                                */
   /*  a better Right() function that does not pad                                                    */
   /* =============================================================================================== */
   parse arg string , len , .

   if len > length(string) then
      len = (length(string))
   if len > 0 then
      return right(string, len)
   else
      return ''
   /* === End of RightStr =========================================================================== */

   Quoted: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = Quoted(string)                                                                  */
   /*                                                                                                 */
   /*    if possible, quotes a string with the kind of quotes it does not contain                     */
   /*        {He said "no"}   --> {'He said "no"'}                                                    */
   /*        {He said 'no'}   --> {"He said 'no'"}                                                    */
   /*    otherwise uses double quotes and lets the caller deal with the errors                        */
   /* =============================================================================================== */
   parse arg string
   select
      when pos('"', string) == 0 then
         newstring = '"'string'"'
      when pos("'", string) == 0 then
         newstring = "'"string"'"
      otherwise
         newstring = '"'string'"'
   end
   return newstring
   /* === End of Quoted ============================================================================= */

   Progress: procedure expose (Shared_vars) /*fold00*/
   ShowProgress:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        CALL ShowProgress <STRING> , <ROW> , <COL>                                               */
   /* ----------------------------------------------------------------------------------------------- */
   /*  writes STRING to the screen at ROW, COL and returns cursor to where it was                     */
   /*  STRING is padded or trimmed to screen width                                                    */
   /*  If ROW or COL are not given, defaults to current cursor position                               */
   /*  Call it from a loop to show progress. When done, call it with a null string                    */
   /*  to erase the progress display                                                                  */
   /*  returns 1 on success, 0 if bad parameter                                                       */
   /* =============================================================================================== */
   /*  to prevent scrolling, do not use the last column of the last row                               */

   parse value SysCurPos() with origRow origCol
   parse arg msg , row , col
   if row = '' then row = origRow
   if col = '' then col = origCol
   if \datatype(row, 'w') ,
    | \datatype(col, 'w') then return 0
   parse value SysTextScreenSize() with maxRow maxCol
   maxRow = maxRow -1 ; maxCol = maxCol -1

   if row = '' then row = origRow
   if col = '' then col = origCol
   if row > maxRow then row = maxRow
   if col > maxCol then col = maxCol
   if row < 0 then row = 0
   if col < 0 then col = 0

   msgLen = (maxCol - col)
   if row < maxRow then msgLen = msgLen +1
   msg = left(msg, msgLen)
   call SysCurPos row, col
   call charout 'SCREEN$', msg
   call SysCurPos origRow, origCol

   return 1
   /* === End of ShowProgress ======================================================================= */

   StripStr: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        ret = StripStr(<string> [, <where>] ] [, <characters>]                                   */
   /* ----------------------------------------------------------------------------------------------- */
   /*  removes leading or trailing characters from a string.                                          */
   /*  the arguments are the same as STRIP(), but the third can be more than one character.           */
   /*  Like Strip(), <characters> is case-sensitive.                                                  */
   /*  Unlike Strip(), <characters> defaults to removing Space and Tab  ('20'x||'09'x)                */
   /*  like Strip(), <where> must be 'L'leading  or 'T'railing  or 'B'oth, only the first             */
   /*        letter matters, it can be in upper or lower case, and it defaults to 'B'                 */
   /*                                                                                                 */
   /*  <characters> is a string of one or more characters. All of them will be stripped.              */
   /* =============================================================================================== */

   str = arg(1)
   if arg(2, 'e') then
      where = translate(LeftStr(arg(2), 1))
   else
      where = 'B'
   if arg(3, 'e') then
      what = arg(3)
   else
      what = ' '||'09'x
   if pos(where, 'BLT') = 0 then where = 'B'
   if where = 'B' ,
    | where = 'L' then do
         i = 0
         do while pos(LeftStr(str, 1), what)  > 0
            i = i +1
            do j = 1 to length(what)
               c = substr(what, j, 1)
               str = strip(str, 'L', c)
            end j
            if i > 6 then leave
         end
      end

   i = 0
   if where = 'B' ,
    | where = 'T' then do
         do while pos(RightStr(str, 1), what)  > 0
            i = i +1
            do j = 1 to length(what)
               c = substr(what, j, 1)
               str = strip(str, 'T', c)
            end j
            if i > 6 then leave
         end
      end
   return str
   /* === End of StripStr =========================================================================== */

   Unquoted: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = Unquoted(string)                                                                */
   /*                                                                                                 */
   /*    strips quotes from a string if and only if it is enclosed in a balanced pair                 */
   /*    of single or double quotes                                                                   */
   /* =============================================================================================== */
   parse arg string
   if length(string) < 2 then return string
   select
      when left(string, 1) == '"' ,
         & right(string, 1) == '"' then
         string = substr(string, 2, length(string) -2)
      when left(string, 1) == "'" ,
         & right(string, 1) == "'" then
         string = substr(string, 2, length(string) -2)
      otherwise
         nop
   end
   return string
   /* === End of Unquoted =========================================================================== */

   WordQ: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     foo = WordQ(<string>, <word number>, <separator characters>, <quote characters>)            */
   /*                                                                                                 */
   /*  The last two parameters are optional, so it will work                                          */
   /*  as a replacement for the Rexx Word() function.                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  A better Word() function--                                                                     */
   /*  It recognizes a quoted string as one word                                                      */
   /*  Optional parameters let you decide which characters                                            */
   /*  count as word separators and as quotes.                                                        */
   /*                                                                                                 */
   /*  Default word separators are <space> and <tab>                                                  */
   /*  Default quote characters are single, double, and                                               */
   /*  back quotes-- ' " `                                                                            */
   /* =============================================================================================== */

   parse arg str , thisnum , chrsW , chrsQ
   if chrsW == '' then chrsW = ' '||d2c(9)
   if chrsQ == '' then chrsQ = '"'||"'"||'`'
   i         = 1
   quote     = ''
   numwords  = 0
   Thisword  = ''
   w1        = 0
   w2        = 0

   /* find start of first word by trimming blanks and word characters                                 */
   do while i <= length(str) ,
          & pos(substr(str, i, 1), chrsW) > 0
      i = i +1
   end
   if i <= length(str) then numwords = numwords +1
   if numwords == thisnum then w1 = i

   do while (i <= length(str))
      if pos(substr(str, i, 1), chrsQ) \= 0 then do
            /* start of quoted string;
             nothing counts until we get to the matching quote (or end of string) */
            quote = substr(str, i, 1)
            tsti = i +1
            /*  bugfix for "-opt=''"                                                                  */
            do while tsti < length(str) ,
                   & substr(str, tsti, 1) \= quote
               tsti = tsti +1
            end
            if substr(str, tsti, 1) == quote then
               /* found the matching quote                                                            */
               i = tsti
            else
               /* ran out of string with no closing quote, so ignore this quote                       */
               nop
            i = i +1
         end   /*  quote  */

      if pos(substr(str, i, 1), chrsW) \= 0 ,
       & i <= length(str) then do
            /* a word separator, skip all blanks, find start of next word                             */
            /* unless the end of the word we want, then quit                                          */
            if w1 > 0 then do
                  w2 = i -1
                  leave
               end
            do while (pos(substr(str, i, 1), chrsW) \= 0) ,
                   & (i <= length(str))
               i = i +1
            end
            if i <= length(str) then numwords = numwords +1
            /* if this is the start of the word we want, note position in string                      */
            if numwords == thisnum then w1 = i
         end   /*  wordsep  */
      else
         /* a normal character                                                                        */
         i = i +1
   end   /*  do i  */
   if i > length(str) then i = length(str)
   if w1 > 0 then do
         /* we know w1, if we ran out of string, w2 is the end of string                              */
         if w2 == 0 then w2 = i
         /* found start and end of word, copy it                                                      */
         wlen = w2 - w1 +1
         Thisword = substr(str, w1, wlen)
      end
   return Thisword
   /* === End of WordQ ============================================================================== */

   WordsQ: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     foo = WordsQ(string, <separator characters>, <quote characters>)                            */
   /*                                                                                                 */
   /*  The last two parameters are optional, so it will work                                          */
   /*  as a drop-in replacement for the Rexx Word() function.                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  A better Words() function--                                                                    */
   /*  It recognizes a quoted string as one word                                                      */
   /*  Optional parameters let you decide which characters                                            */
   /*  count as word separators and as quotes.                                                        */
   /*                                                                                                 */
   /*  Default word separators are <space> and <tab>                                                  */
   /*  Default quote characters are single, double, and                                               */
   /*  back quotes-- ' " `                                                                            */
   /* =============================================================================================== */

   parse arg str , chrsW , chrsQ
   if chrsW == '' then chrsW = ' '||d2c(9)
   if chrsQ == '' then chrsQ = '"'||"'"||'`'
   i         = 1
   quote     = ''
   numwords  = 0

   /* find start of first word by trimming blanks and word characters                                 */
   do while i <= length(str) ,
          & pos(substr(str, i, 1), chrsW) > 0
      i = i +1
   end
   if i <= length(str) then numwords = numwords +1

   do while (i <= length(str))
      if pos(substr(str, i, 1), chrsQ) \= 0 then do
            /* start of quoted string;
             nothing counts until we get to the matching quote (or end of string) */
            quote = substr(str, i, 1)
            tsti = i +1
            /*  bugfix for "-opt=''"                                                                  */
            do while tsti < length(str) ,
                   & substr(str, tsti, 1) \= quote
               tsti = tsti +1
            end
            if substr(str, tsti, 1) == quote then
               /* found the matching quote                                                            */
               i = tsti
            else
               /* ran out of string with no closing quote, so ignore this quote                       */
               nop
            i = i +1
         end   /*  quote  */

      if pos(substr(str, i, 1), chrsW) \= 0 ,
       & i <= length(str) then do
            /* a word separator, skip all blanks, find start of next word                             */
            do while (pos(substr(str, i, 1), chrsW) \= 0) ,
                   & (i <= length(str))
               i = i +1
            end
            if i <= length(str) then numwords = numwords +1
         end
      else
         /* a normal character                                                                        */
         i = i +1
   end   /*  do i  */
   return numwords
   /* === End of WordsQ ============================================================================= */

   ShowGetOpts: procedure expose (Shared_vars) ArgC ArgV. /*fold00*/
   /* =============================================================================================== */
   /*    show what GetOpts() returned and how it was interpreted                                      */
   /* =============================================================================================== */
   say copies('=', 80)
   say ' getopts returned these arguments:'
   do i = 1 to ArgC
      say '   ' ArgV.i
   end
   say copies('=', 80)

   me = filespec('name', ArgV.0)
   /*    show the opt.* variables                                                                     */
   VarList = 'Opt.0optlist 'Opt.0optlist' ArgC'
   do i = 0 to ArgC
      VarList = VarList' ArgV.'i
   end
   do while VarList <> ''
      parse var VarList thisvar VarList
      call lineout 'stderr', '  'overlay('=  'value(thisvar), thisvar, 25)
   end
   return 0
   /* === End of ShowGetOpts ======================================================================== */

   CountStr: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        Count = CountStr(<needle>, <haystack>)                                                   */
   /*                                                                                                 */
   /*  counts how many times the string <needle> occurs in the string <haystack>, without overlap     */
   /*  if either <needle> or <haystack> is null, returns 0                                            */
   /*                                                                                                 */
   /*  this is an imitation of the CountStr() function from Object Rexx                               */
   /* =============================================================================================== */
   parse arg needle , haystack

   found = 0
   p = pos(needle, haystack)
   do while p > 0
      found = found +1
      p = pos(needle, haystack, p + length(needle))
   end
   return found
   /* === End of CountStr =========================================================================== */

   WordSort: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = WordSort(string [, 'C'] [,<start>)                                              */
   /*  Examples:                                                                                      */
   /*        string = WordSort(string)                                                                */
   /*           sorts words by first letter, case-insensitive                                         */
   /*        string = WordSort(string, 'c')                                                           */
   /*           sorts words by first letter, case-sensitive                                           */
   /*        string = WordSort(string, , 2)                                                           */
   /*           sorts words by second letter, case insensitive                                        */
   /* ----------------------------------------------------------------------------------------------- */
   /*  sorts a string of words alphabetically.                                                        */
   /*  By default the sort is NOT case-sensitive. Sorting is case-sensitive if the second             */
   /*  argument is 'C'                                                                                */
   /*                                                                                                 */
   /*  if no third argument is given, comparison starts with the first letter of each word            */
   /*                                                                                                 */
   /* =============================================================================================== */
   /*  adapted from Toby Thurston's checkfunc.cmd                                                     */
   /*  case-insensitive option and starting position option added by akm                              */
   /* =============================================================================================== */
   s = arg(1)
   case = translate(arg(2))
   o = arg(3)
   if o == '' then o = 1
   if case == 'C' then do                                                      /* case-sensitive sort */
         s = '00'x s                                   /* the leading 'null word' simplifies the sort */
         do j = 2 to words(s)
            i = j -1
            if substr(word(s, i), o) >> substr(word(s, j), o) then do
                  key = word(s, j)
                  s = delword(s, j, 1)
                  do until substr(word(s, i), o) <<= substr(key, o)
                     i = i -1
                  end
                  s = subword(s, 1, i) key subword(s, 1 + i)
               end
         end j
         return space(subword(s, 2))                         /* get rid of the null word at the front */
      end   /*  if 'C'  */
   else do                                                             /* case-insensitive is default */
         s = '00'x s                                   /* the leading 'null word' simplifies the sort */
         do j = 2 to words(s)
            i = j -1
            if translate(substr(word(s, i), o)) >> translate(substr(word(s, j), o)) then do
                  key = word(s, j)
                  s = delword(s, j, 1)
                  do until translate(substr(word(s, i), o)) <<= translate(substr(key, o))
                     i = i -1
                  end
                  s = subword(s, 1, i) key subword(s, 1 + i)
               end
         end j
         return space(subword(s, 2))                         /* get rid of the null word at the front */
      end   /*  else  */
   /* === End of WordSort =========================================================================== */

   Error: /*fold00*/
   /* =============================================================================================== */
   /*  signal on error                                                                                */
   /* ----------------------------------------------------------------------------------------------- */
   /*  this does not do much, but it is not used much anyway                                          */
   /* =============================================================================================== */
   signal off Error
   parse arg errmsg
   call lineout 'stderr', '[ErrorHandler] 'rc' at line 'sigl': "'errmsg'"'
   call lineout 'stderr', '       Error # 'rc' is "'errortext(rc)'"'
   call lineout 'stderr', '       for more help, type "helpmsg REX'rc'"'
   call Egress 'BUG_Error'
   /* === End of Error ============================================================================== */

   Syntax: /*fold00*/
   /* =============================================================================================== */
   /*  signal on syntax                                                                               */
   /* ----------------------------------------------------------------------------------------------- */
   /*  depends on variables: G.0me sigl                                                               */
   /*  depends on functions: LastLabel(), ShowError(), Egress()                                       */
   /* ----------------------------------------------------------------------------------------------- */
   /*  traps syntax errors and shows a better error message than Rexx does. In particular,            */
   /*  an incomplete DO/SELECT/IF is fairly common and the error message from Rexx is no              */
   /*  help at all.                                                                                   */
   /*                                                                                                 */
   /*  Some errors cannot be handled by an error handler, like REX06 "unmatched comment or quote"     */
   /*  (Which makes sense. If a defective script cannot be tokenised, then it cannot be run, so       */
   /*  the handler never exists. But the help message for REX06 implies that error 6 may not be       */
   /*  detected until later)                                                                          */
   /*                                                                                                 */
   /*  This will always be a work in progress                                                         */
   /* =============================================================================================== */
   SyntaxSigl = sigl
   Err_func = LastLabel(SyntaxSigl)
   call lineout 'stderr', ''
   select
      when rc = 14 then do
            /*  (incomplete DO/SELECT/IF) Rexx reports a sigl at the end of file, which is useless    */
            call ShowError '[SYNTAX Handler]'
            call ShowError '    Syntax error 'rc' which is "'errortext(rc)'"', ''
            if SyntaxSigl >= sourceline() then do
                  call ShowError '    Rexx says it is at line 'SyntaxSigl', which is the end of file.', ''
                  call ShowError '    To find where the error really is, use ', ''
                  call ShowError '        "RexxTool.cmd checkonly' G.0me'"', ''
                  call ShowError '    Or you could try putting "SIGNAL Runout_Handler" between routines'
               end
            else do
                  call ShowError '    at line 'SyntaxSigl, ''
                  call ShowError '       ==> {'sourceline(SyntaxSigl)'}'
               end
         end
      otherwise
         call ShowError '[SYNTAX Handler] Syntax error at line 'SyntaxSigl 'in label' Err_func
         call ShowError '==> {'sourceline(SyntaxSigl)'}'
         call ShowError 'Error # 'rc' is "'errortext(rc)'"', ''
         if condition('d') \= '' then call ShowError 'condition(D) is "'condition('d')'"'
   end
   call ShowError 'for more help, type "helpmsg REX'rc'"', ''
   if G.0track_vars \= '' then call Mark G.0track_vars, ''
   call Egress 'BUG_Syntax'
   /* === End of Syntax ============================================================================= */

   NoValue: /*fold00*/
   /* =============================================================================================== */
   /* =============================================================================================== */
   signal off NoValue
   NVsigl = sigl
   Err_func = LastLabel(NVsigl)
   call Msg ''
   call Msg G.0meID'[NOVALUE Handler] at line 'NVsigl' for variable "'condition('d')'"'
   call Msg '   ==> 'sourceline(NVsigl)
   call Msg 'under label "'Err_func'"'
   call Mark G.0track_vars, ''
   call Egress 'bug_NoValue'
   /* === End of NoValue ============================================================================ */


